;; -*-Mode: Elisp;-*-
;; $Id: make-docs.el 1.8 Wed, 31 Jul 1996 20:10:06 -0700 jmacd $
;; make-docs.el:
;;
;; Copyright (C) 1996, Regents of the University of California.
;; All Rights Reserved.
;;
;; This file is part of PRCS, the Project Revision Control System.
;;
;; Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>

;; Much of the code in this file comes from makeinfo.el and falls
;; under the GNU license.  It was modified to suit the purpose of
;; generating document strings.

;;; makeinfo.el --- run makeinfo conveniently

;; Copyright (C) 1991, 1993 Free Software Foundation, Inc.

;; Author: Robert J. Chassell
;; Maintainer: FSF

;; This file is part of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

(defvar makeinfo-run-command "makeinfo"
  "*Command used to run `makeinfo' subjob.
The name of the file is appended to this string, separated by a space.")

(defvar makeinfo-options '("--fill-column=70" "--no-headers" "--no-validate")
  "*String containing options for running `makeinfo'.
Do not include `--footnote-style' or `--paragraph-indent';
the proper way to specify those is with the Texinfo commands
`@footnotestyle` and `@paragraphindent'.")

(require 'texinfo)

(defvar makeinfo-compilation-process nil
  "Process that runs `makeinfo'.  Should start out nil.")

(defvar makeinfo-temp-file nil
  "Temporary file name used for text being sent as input to `makeinfo'.")

(defvar makeinfo-output-file-name nil
  "Info file name used for text output by `makeinfo'.")

;;; The `makeinfo' function definitions

(defun makeinfo-region (region-beginning region-end output-buffer)
  "Make Info file from region of current Texinfo file, and switch to it.

This command does not offer the `next-error' feature since it would
apply to a temporary file, not the original; use the `makeinfo-buffer'
command to gain use of `next-error'."

  (interactive "r")
  (let (filename-or-header
        filename-or-header-beginning
        filename-or-header-end)
    ;; Cannot use `let' for makeinfo-temp-file or
    ;; makeinfo-output-file-name since `makeinfo-compilation-sentinel'
    ;; needs them.

    (setq makeinfo-temp-file
          (concat
           (make-temp-name
            (substring (buffer-file-name)
                       0
                       (or (string-match "\\.tex" (buffer-file-name))
                           (length (buffer-file-name)))))
           ".texinfo"))

    (save-excursion
      (save-restriction
        (widen)
        (goto-char (point-min))
        (let ((search-end (save-excursion (forward-line 100) (point))))
          ;; Find and record the Info filename,
          ;; or else explain that a filename is needed.
          (if (re-search-forward
               "^@setfilename[ \t]+\\([^ \t\n]+\\)[ \t]*"
               search-end t)
              (setq makeinfo-output-file-name
                    (buffer-substring (match-beginning 1) (match-end 1)))
            (error
             "The texinfo file needs a line saying: @setfilename <name>"))

          ;; Find header and specify its beginning and end.
          (goto-char (point-min))
          (if (and
               (prog1
                   (search-forward tex-start-of-header search-end t)
                 (beginning-of-line)
                 ;; Mark beginning of header.
                 (setq filename-or-header-beginning (point)))
               (prog1
                   (search-forward tex-end-of-header nil t)
                 (beginning-of-line)
                 ;; Mark end of header
                 (setq filename-or-header-end (point))))

              ;; Insert the header into the temporary file.
              (write-region
               (min filename-or-header-beginning region-beginning)
               filename-or-header-end
               makeinfo-temp-file nil nil)

            ;; Else no header; insert @filename line into temporary file.
            (goto-char (point-min))
            (search-forward "@setfilename" search-end t)
            (beginning-of-line)
            (setq filename-or-header-beginning (point))
            (forward-line 1)
            (setq filename-or-header-end (point))
            (write-region
             (min filename-or-header-beginning region-beginning)
             filename-or-header-end
             makeinfo-temp-file nil nil))

          ;; Insert the region into the file.
          (write-region
           (max region-beginning filename-or-header-end)
           region-end
           makeinfo-temp-file t nil)

	  (if (not (eq (apply (function call-process)
			     makeinfo-run-command nil output-buffer nil
			     (append makeinfo-options (list makeinfo-temp-file)))
		       0))
	      (error "makeinfo failed--you need a bug-free(ha!) version of makeinfo.  This particular bug can be fixed with a patch obtainable from jmacd@cs.berkeley.edu.")
	    (if (and makeinfo-temp-file (file-exists-p makeinfo-temp-file))
		(delete-file makeinfo-temp-file))))))))

;; All code below this comment was written by Josh MacDonald and is
;; part of PRCS.

(defun make-new-file(file-name source)
  (let ((buffer (get-buffer-create (generate-new-buffer-name file-name))))
    (set-buffer buffer)
    (insert
     "/* -*-Mode: C;-*-
 *
 * Copyright (C) ")
    (insert (substring (current-time-string) -4))
    (insert ", Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * This file was automatically generated by Emacs and the source
 * file ")
    (insert (expand-file-name source))
    (insert " at ")
    (insert (current-time-string))
    (insert ": DO NOT EDIT.
 */

")
    buffer))

(defun make-docs(file output output-h)
  "Takes the name of a texinfo file marked with @c BEGIN DOCSTRING name
and @c END DOCSTRING and inserts the region into a c file named output
which contains the region after applying makeinfo"
  (let ((source-buffer (find-file-noselect file))
	(doc-buffer (make-new-file output file))
	(doc-h-buffer (make-new-file output-h file)))
    (if (not source-buffer)
	(error "Can't open file %s" file)
      (set-buffer source-buffer)
      (goto-char (point-min))
      (while (re-search-forward "@c BEGIN DOCSTRING \\(\\(\\w\\|[$_]\\)+\\)\n" nil t)
	(let ((source-start (point)))
	  (set-buffer doc-buffer)
	  (insert "\n\nchar const ")
	  (insert-buffer-substring source-buffer (match-beginning 1) (match-end 1))
	  (insert "[] = \n")
	  (set-buffer doc-h-buffer)
	  (insert "extern char const ")
	  (insert-buffer-substring source-buffer (match-beginning 1) (match-end 1))
	  (insert "[];\n")
	  (set-buffer doc-buffer)
	  (let ((last-doc-beginning (point)))
	    (set-buffer source-buffer)
	    (if (not (search-forward "@c END DOCSTRING"))
		(error "unbalanced docstring markers")
	      (beginning-of-line)
	      (makeinfo-region source-start (point) doc-buffer)
	      (set-buffer doc-buffer)
	      (c-escape-stuff last-doc-beginning (point))
	      (insert ";\n")
	      (set-buffer source-buffer)))))
      (set-buffer doc-buffer)
      (write-file output)
      (set-buffer doc-h-buffer)
      (write-file output-h)
      (kill-buffer doc-buffer))))

(defun c-escape-stuff(beg end)
    (save-excursion
      (save-restriction
      (narrow-to-region beg end)
      (goto-char (point-min))
      (perform-replace "\"" "\\\"" nil nil nil)
      (goto-char (point-min))
      (while (< (point) (- (point-max) 1))
	(insert "\"")
	(end-of-line)
	(insert "\\n\"")
	(forward-char)))))

(defun make-docs-noargs()
  (setq make-backup-files nil)
  (make-docs "../doc/prcs.texi" "docs.c" "include/docs.h"))
