/* -*-Mode: C++;-*-
 * $Id: checkin.cc 1.22.1.2.1.9.1.23.1.1.1.6.1.10 Tue, 03 Sep 1996 20:09:11 -0700 jmacd $
 * checkin.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include "prcs.h"

#include "sexp.h"
#include "setkeys.h"
#include "dstring.h"
#include "fileent.h"
#include "projdesc.h"
#include "repository.h"
#include "vc.h"
#include "checkout.h"
#include "checkin.h"
#include "misc.h"
#include "memseg.h"
#include "rebuild.h"
#include "system.h"
#include "quick.h"

static char const prcs_version_format[] =
"PRCS major version: %s\n"
"PRCS minor version: %s\n"
"PRCS parent major version: %s\n"
"PRCS parent minor version: %s\n";

static char const descends_from_format[] =
"PRCS descends from major: %s\n"
"PRCS descends from minor: %s\n";

static char const rlog_header[] = "checked in by PRCS version "
   STRINGIFY(PRCS_MAJOR_VERSION_NUMBER) " "
   STRINGIFY(PRCS_MINOR_VERSION_NUMBER) " "
   STRINGIFY(PRCS_MICRO_VERSION_NUMBER) "\n";

static PrVoidError write_new_prjfile(ProjectDescriptor*);
static void update_attributes(ProjectDescriptor*);
static PrVoidError checkin_each_file(ProjectDescriptor*);
static PrVoidError maybe_query_user(ProjectDescriptor*);
static PrVoidError slow_eliminate_prepare(ProjectDescriptor*);
static PrVoidError checkin_cleanup_handler(void* data, bool);

PrPrcsExitStatusError checkin_command()
{
    ProjectDescriptor *project;

    Return_if_fail(project << read_project_file(cmd_root_project_full_name,
						cmd_root_project_file_path));

    install_cleanup_handler(checkin_cleanup_handler, project);

    Return_if_fail(project->init_repository_entry(cmd_root_project_name, true, true));

    /* remove files which were not named on the command line */
    eliminate_unnamed_files(project);

    /* warn the user about filenames on the command line which didn't match anything */
    Return_if_fail(warn_unused_files(true));

    /* select the new version and possibly warn the user of unwise checkins */
    Return_if_fail(project->resolve_checkin_version(cmd_version_specifier_major));

    /* check that no unselected files have empty descriptors and that all selected
     * files exist */
    Return_if_fail(check_project(project));

    Return_if_fail(eliminate_working_files(project, QueryUserRemoveFromCommandLine));

    /* if the auxiliary database is present, read it and eliminate files which
     * have not been touched since the last checkout or rekey */
    project->read_quick_elim();
    project->quick_elim_unmodified();

    /* now eliminate files which have not changed by reading and unkeying files,
     * comparing the unkeyed length, then computing their checksum and comparing
     * the unkeyed checksum.  with the unkeyed file in memory, prepare the repository
     * file (possibly uncompress), and write it out into the temporary location
     * in the repository. */
    Return_if_fail(slow_eliminate_prepare(project));

    prcsinfo << "Checking in project " << squote(cmd_root_project_name)
	     << " version " << project->project_version_major() << "."
	     << project->project_version_minor() << dotendl;

    bool any_modified = false;

    foreach_fileent(fe_ptr, project) {
	FileEntry *fe = *fe_ptr;

	if(fe->active_modified())
	    any_modified = true;
    }

    if(!any_modified) {
	prcsquery << "No modified files were found.  "
		  << force("Aborting checkin")
		  << report("Abort checkin")
		  << option('n', "Continue, with no modified files in this version")
		  << deffail('y')
		  << query("Abort");

	Return_if_fail(prcsquery.result());
    }

    /* check if the user wants to enter a log. */
    Return_if_fail(maybe_query_user(project));

    if(option_long_format) {
	foreach_fileent(fe_ptr, project) {
	    FileEntry *fe = *fe_ptr;

	    if(!fe->active_modified())
		continue;

	    prcsoutput << "New version of file " << squote(fe->working_path()) << prcsendl;
	}
    }

    if(!option_report_actions) {

	Return_if_fail(checkin_each_file(project));
	/* if failure: the new version isn't completely checked in, all
	 * this means is that extra RCS versions have been inserted but
	 * that's not a big deal. */

	foreach_fileent(fe_ptr, project) {
	    FileEntry* fe = *fe_ptr;

	    if(fe->file_type() == RealFile && fe->on_command_line())
		project->quick_elim_update_fe(fe);
	}

	project->quick_elim_update();

	update_attributes(project);

	Return_if_fail(write_new_prjfile(project));
	/* if failure: the new version isn't completely checked in, all
	 * this means is that extra RCS versions have been inserted but
	 * that's not a big deal. */
    }

    Return_if_fail(project->update_repository());

    return ExitSuccess;
}

/* update_attributes --
 *
 *     now the new project file has to be built.  the project file's
 *     attribues are modified for the new version, prior to writing
 *     the new project file.  */
void update_attributes(ProjectDescriptor* project)
{
    foreach_fileent(fe_ptr, project)
        (*fe_ptr)->update_project_file();

    /* All version names and numbers are updated in
     * ProjectDescriptor::resolve_checkin_version */

    /* the Checkin-Time attribute */
    project->checkin_time()->sprintf("\"%s\"", get_utc_time());

    /* the Logs */
    project->version_log()->assign(*project->new_version_log());
    project->new_version_log()->assign("\"\"");

    /* the Checkin-Login attribute */
    project->checkin_login()->assign(get_login());
}

/* check_project --
 *
 *     this checks lots of things about the project descrition.
 *     basically it checks consistency as much as is possible (I
 *     think).  A call to checkDescriptor() for each FileEntry checks
 *     that the file exists and is readable, that the name is "proper"
 *     (no ".."'s), warns when symlinks point to files that don't
 *     exist, and that the file type is right.  then this function
 *     checks that there are no duplicate inodes in the description,
 *     that the project file is not put under version control (this is
 *     implicit).  then it warns when hard links point to files that
 *     are not under version control.
 *
 *     init tells the procedure to actually initialize each file entry
 *     with a call to initialize_descriptor and register new RCS
 *     files.
 *
 *     for all files with non-empty file descriptors, check that the
 *     file repository exists and is readable.
 *
 *     for all files with empty file descriptors, determine a unique
 *     name for the repository RCSfile, and replace the empty filedesc
 *     wth it.  */
PrVoidError check_project(ProjectDescriptor *project)
{
    bool oncmd;
    FileType type;
    const char* name;

    foreach_fileent(fe_ptr, project) {
	FileEntry *fe = *fe_ptr;

	oncmd = fe->on_command_line();
        type = fe->file_type();
        name = *fe->working_path();

        if(!oncmd && (type == RealFile || type == SymLink) && !fe->descriptor_name()) {
	    /* if the user is doing a partial operation, the file needs
	     * to have a non-empty internal descriptor, the descriptor
	     * is currently, empty, and init has asked to register new
	     * files for empty descriptors, complain. */
	    static BangFlag bang;

	    prcsquery << "The file " << squote(fe->working_path())
		      << " has an empty descriptor, and was excluded on the command line.  "
		      << force("Continuing")
		      << report("Continue")
		      << allow_bang(bang)
		      << optfail('n')
		      << defopt('y', "Select this file, as if it were on the command line")
		      << query("Select this file");

	    Return_if_fail(prcsquery.result());

	    oncmd = true;
	    fe->set_on_command_line(true);
        } else if(type == RealFile && fe->descriptor_name()) {
	    RcsVersionData* data;

	    /* Check to see that the version exists, even if not on the
	     * command line */

	    Return_if_fail(data << project->repository_entry()->
			   lookup_rcs_file_data(*fe->descriptor_name(),
						*fe->descriptor_version_number()));

	}
    }

    return NoError;
}

/* this updates the project file and unset_keys it into the repository with
 * a temp name, the repostiory name less the ",v".  then runs checkin. */
static PrVoidError write_new_prjfile(ProjectDescriptor *P)
{
#ifndef NDEBUG
    ProjectDescriptor* copy;

    Return_if_fail(P->write_project_file(temp_file_1));

    If_fail(copy << read_project_file(P->project_full_name(), temp_file_1))
       throw prcserror << "Internal error: produced new project file with "
	"invalid syntax.  Please mail a copy of this project file to "
		       << maintainer << dotendl;

    delete copy;
#endif

    Return_if_fail(P->write_project_file(cmd_root_project_file_path));

    Return_if_fail(P->checkin_project_file());

    return NoError;
}

static PrVoidError checkin_each_file(ProjectDescriptor *project)
{
    /* inconsistencies have been checked the Files attribute has been
     * modified already, excepting the new RCS version numbers */

    foreach_fileent(fe_ptr, project) {
	FileEntry *fe = *fe_ptr;

	if(!fe->active_modified())
	    continue;

	Return_if_fail(VC_checkin(*fe->temp_file_path(),
				  fe->descriptor_version_number(),
				  *fe->full_descriptor_name(),
				  NULL));
    }

    project->repository_entry()->Rep_set_umask();

    Return_if_fail(VC_checkin(NULL, NULL, NULL, rlog_header));

    project->repository_entry()->Rep_unset_umask();

    return NoError;
}

/*
 * eliminate_unnamed_files --
 *
 *     this marks files that were not given on the command line so
 *     that they can be ignored in certain stages of the checkin (and
 *     other) commands.  if no files were given, it leaves them all
 *     marked, otherwise, it unmarks them all and marks all files (and
 *     recursively marks directories).  */
void eliminate_unnamed_files(ProjectDescriptor* project)
{
    if(cmd_filenames_count < 1) {
        cmd_prj_given_as_file = true;
        return;
    }

    for(int i = 0; i < cmd_filenames_count; i += 1) {
        /* special case, the project file */
        if(strcmp(cmd_root_project_file_path, cmd_corrected_filenames_given[i]) == 0) {
            cmd_filenames_found[i] = true;
            cmd_prj_given_as_file = true;
        }
    }

    foreach_fileent(fe_ptr, project) {
	FileEntry *fe = *fe_ptr;

        fe->set_on_command_line(false);

        for(int j = 0; j < cmd_filenames_count; j += 1) {
            const char *f, *w;

            w = *fe->working_path();
            f = cmd_corrected_filenames_given[j];

            while(*f && *w && *f == *w) {
                while(f[1] == '/') { f += 1; }
                while(w[1] == '/') { w += 1; }
                f += 1;
                w += 1;
            }

            if (*f == 0 && (w[-1] == '/' || *w == 0 || f[-1] == '/')) {
                cmd_filenames_found[j] = true;
                fe->set_on_command_line(true);
            }
        }
    }
}


/*
 * warn_unused_files --
 *
 *     this optionally prompts the user to continue and warns if any files
 *     were not matched by a call to eliminate_unnamed_files.
 *
 *     if prompt_abort is true, returns an error if the user does not
 *     confirm ignoring unmatched filenames.
 *
 *     if prompt_abort is false, it will return an error if files are unmatched.
 */
PrVoidError warn_unused_files(bool prompt_abort)
{
    bool any = false, ret = true, need_prompt = false;

    if(cmd_filenames_count < 1)
        return NoError;

    for(int j = 0; j < cmd_filenames_count; j += 1) {
        any |= cmd_filenames_found[j];
        if(!cmd_filenames_found[j]) {
            if(prompt_abort) {
		need_prompt = true;
                prcswarning << "File or directory " << squote(cmd_filenames_given[j])
			    << " on command line does not match any working files" << dotendl;
            } else {
                ret = false;
                prcswarning << "File or directory " << squote(cmd_filenames_given[j])
			    << " on command line did not match any working files" << dotendl;
            }
        }
    }

    if(!any)
	throw prcserror << "No files on command line match working files" << dotendl;

    if(need_prompt) {
	prcsquery << "Files on the command line did not match files in the project.  "
		  << force("Continuing")
		  << report("Continue")
		  << optfail('n')
		  << defopt('y', "Continue, ignoring these files")
		  << query("Continue");

	Return_if_fail(prcsquery.result());
    }

    if(ret)
	return NoError;
    else
	return PrVoidError(NonFatalError);
}

static PrVoidError maybe_query_user(ProjectDescriptor* P)
{
    int tty, c, lc, llc;

    if(P->new_version_log()->length() != 2 || /* 2 is the two quote marks */
       !getenv("PRCS_LOGQUERY") || /* either it has been edited or PRCS_LOGQUERY is not set */
       option_report_actions)
        return NoError;

    tty = isatty(STDIN_FILENO);

    if(tty) {
	if (option_force_resolution)
	    return NoError;

	prcsquery << "Empty New-Version-Log.  "
		  << option('e', "Enter version log now")
		  << defopt('p', "Proceed with an empty version log")
		  << query("Enter a log, or proceed");

	char c;

	Return_if_fail(c << prcsquery.result());

	if(c == 'p')
	    return NoError;

        prcsoutput << "Enter description terminated by a single '.' or EOF" << dotendl;

	fprintf(stdout, ">> ");
    }

    re_query_message = ">> ";
    re_query_len = 3;

    P->new_version_log()->truncate(1);

    lc = '\0';
    llc = '\0';

    while(true) {
	If_fail(c << Err_fgetc(stdin))
	    throw prcserror << "Read failure on stdin" << perror;

	if (c == EOF)
	    break;

        if(tty && c == '\n') {
            if(lc == '.' && llc == '\n') {
                P->new_version_log()->truncate(P->new_version_log()->length() - 1);
                break;
            }
            fprintf(stdout, ">> ");
        }
        llc = lc;
        lc = c;
        if(c == '\\' || c == '\"')
            P->new_version_log()->append('\\');
        P->new_version_log()->append(c);
    }

    if(tty) fprintf(stdout, "\nDone.\n");

    if(P->new_version_log()->length() > 1)
        P->new_version_log()->truncate(P->new_version_log()->length() - 1);
    P->new_version_log()->append('\"');

    /* I don't think its possible that PRCS will ever need to query the
     * user after this, but I might as well reopen it. */
    if(freopen(ctermid(NULL), "r", stdin) == NULL)
	throw prcserror << "Couldn't reopen the standard input" << perror;

    return NoError;
}

PrVoidError ProjectDescriptor::resolve_checkin_version(const char* maj)
{
    /* since checkin ignores minor versions, we can assume minor is "@" */
    const char *new_major_version, *old_major_version;
    int new_minor_version, old_minor_version;
    bool changing_major;
    ProjectVersionData *project_data, *most_distant_ancestor_data;

    if(descends_from_major()) {
        descends_from_name()->assign("-*-");
	descends_from_major()->assign("-*-");
	descends_from_minor()->assign("-*-");
    }

    Return_if_fail(adjust_merge_parents());

    if(strcmp(*project_version_minor(), "0") == 0) {
	old_minor_version = 0;
	old_major_version = intern(*project_version_major());
    } else {
       /* Don't beleive what's in the project file, make sure its in the
	* repository */
	project_data = repository_entry()->lookup_version(*project_version_major(),
							  *project_version_minor());

	if(!project_data)
	    throw prcserror << "Illegal version, " << project_version_major()
			     << '.' << project_version_minor()
			     << ", in project file, it may have been deleted" << dotendl;

	if(project_data->deleted())
	    throw prcserror << "Version " << project_version_major()
			     << '.' << project_version_minor()
			     << " has been deleted, this project file is invalid"
			     << dotendl;

	old_major_version = project_data->prcs_major();
	old_minor_version = project_data->prcs_minor_int();
    }

    /* First compute the new major version */

    if(!maj[0]) {
	new_major_version = *project_version_major();
    } else if(strcmp(maj, "@") == 0) {
	int highest_major;

	if(repository_entry()->version_count() == 0)
	    throw prcserror << "No versions in the repository, can't "
		"resolve major version '@'" << dotendl;

	highest_major = repository_entry()->highest_major_version();

	if(highest_major >= 0) {
	    project_version_major()->sprintf("%d", highest_major);
	    new_major_version = *project_version_major();
	} else
	    throw prcserror << "No numeric major versions in the repository, "
		"can't resolve major version '@'" << dotendl;
    } else {
	project_version_major()->assign(maj);
	new_major_version = maj;
    }

    changing_major = strcmp(old_major_version, new_major_version) != 0;

    new_minor_version = repository_entry()->highest_minor_version(new_major_version) + 1;
    project_version_minor()->sprintf("%d", new_minor_version);

    /* project_version_minor and project_version_major have been set.
     * Now check to see that the new version is safe, otherwise
     * query/warn the user and set the parent_versions. */

    parent_version_name()->assign(project_name());
    parent_version_major()->assign(old_major_version);
    parent_version_minor()->sprintf("%d", old_minor_version);

    bool skip_check = false;

    if(changing_major) {
	if(!repository_entry()->major_version_exists(new_major_version)) {
	    /* This major does not already exist, creating a new branch */
	    if(repository_entry()->version_count() != 0) {
		/* Other majors do exist, confirm */
		prcsquery << "No previous major version named "
			  << squote(new_major_version) << ".  "
			  << force("Creating")
			  << report("Create")
			  << optfail('n')
			  << defopt('y', "Create the new branch and continue")
			  << query("Create");

		Return_if_fail(prcsquery.result());
	    } else if(option_report_actions) {
		prcsinfo << "Initialize repository entry for new project"
			 << new_major_version << dotendl;
	    } /* else, the user is creating an initial version, so don't query */

	    skip_check = true;

	} else {
	    /* changing major version names, check to see that the
	     * common version of the new version which hasn't been
	     * incremented and the old version is the new version,
	     * otherwise either the head of the branch being checked
	     * into has no common version with the new version or the
	     * branch being checked into has had other versions
	     * checked in between it and the common version.  In
	     * either of these cases a merge is reccomended */
	    most_distant_ancestor_data = repository_entry()->
		most_distant_ancestor(new_major_version,
				      new_minor_version - 1,
				      old_major_version,
				      old_minor_version);

	    set_descends_from(old_major_version, old_minor_version);

	    if(most_distant_ancestor_data) {
	        prcsinfo << "Changing major version name, most distant ancestor with same major is "
			 << most_distant_ancestor_data->prcs_major() << '.'
			 << most_distant_ancestor_data->prcs_minor() << dotendl;

		old_major_version = most_distant_ancestor_data->prcs_major();
		old_minor_version = most_distant_ancestor_data->prcs_minor_int();

	    } else {
	        prcsinfo << "Changing major version name, no ancestor on new branch" << dotendl;

		old_major_version = new_major_version;
		old_minor_version = 0;
	    }

	    parent_version_major()->assign(old_major_version);
	    parent_version_minor()->sprintf("%d", old_minor_version);

	}
    }

    if(!skip_check) {
	for(int i = new_minor_version - 1; i > old_minor_version; i -= 1) {
	    if(i == 0)
		continue;

	    char buf[10];

	    sprintf(buf, "%d", i);

	    project_data = repository_entry()->lookup_version(new_major_version, buf);

	    if(!project_data)
		throw prcserror << "Missing version, " << new_major_version
				 << '.' << i
				 << ", referenced in repository, your repository is seriously damanged" << dotendl;

	    if(project_data->deleted())
		continue;

	    /* skipped a minor version, warn user, tell them to consider
	     * merging */
	    prcsquery << "A newer minor version " << i
		      << " on this branch has been checked in by " << project_data->author()
		      << ", the working files here descend from minor version " << old_minor_version << ".  "
		      << force("Continuing")
		      << report("You are advised to merge with the latest version or create a new branch")
		      << optfail('n')
		      << defopt('y', "Continue checking in this version")
		      << query("You are advised to merge with the latest version or create a new branch, continue");

	    Return_if_fail(prcsquery.result());

	    break;
	}
    }

    return NoError;
}

PrVoidError ProjectDescriptor::checkin_project_file()
{
    Dstring version_log;
    ProjectVersionData new_project_data;

    new_project_data.prcs_major(project_version_major()->cast());
    new_project_data.prcs_minor(project_version_minor()->cast());
    new_project_data.parent_major(parent_version_major()->cast());
    new_project_data.parent_minor(parent_version_minor()->cast());

    if(descends_from_major() && strcmp("-*-", *descends_from_major()) != 0) {
	new_project_data.descends_from_major(descends_from_major()->cast());
	new_project_data.descends_from_minor(descends_from_minor()->cast());
    }

    format_version_log(&new_project_data, version_log);

    repository_entry()->Rep_set_umask();

    If_fail(VC_checkin(project_file_path(),
		       &_new_rcs_version,
		       repository_entry()->Rep_name_of_version_file(),
		       NULL))
	throw prcserror << "Failure checking in project file" << dotendl;

    If_fail(VC_checkin(NULL, NULL, NULL, version_log))
	throw prcserror << "Failure checking in project file" << dotendl;

    repository_entry()->Rep_unset_umask();

    return NoError;
}

void format_version_log(ProjectVersionData* project_data, Dstring& log)
{
    log.assign(rlog_header);
    log.sprintfa(prcs_version_format,
		 project_data->prcs_major(),
		 project_data->prcs_minor(),
		 project_data->parent_major(),
		 project_data->parent_minor());

    if(project_data->has_descends_from()) {
	log.sprintfa(descends_from_format,
		     project_data->descends_from_major(),
		     project_data->descends_from_minor());
    }

    if(project_data->deleted())
	log.append("PRCS version deleted\n");
}

void ProjectDescriptor::quick_elim_unmodified()
{
    if(!_quick_elim)
	return;

    foreach_fileent(fe_ptr, this) {
	FileEntry* fe = *fe_ptr;

	if(!fe->active_modified())
	    continue;

	if(_quick_elim->unchanged(fe)) {
	    DEBUG("Quick elim succeeds for file " << squote(fe->working_path()));
	    if(option_long_format) {
		prcsoutput << "File " << squote(fe->working_path())
			   << " is unmodified" << dotendl;
	    }

	    fe->set_modified(false);
	}
    }
}

static PrVoidError slow_eliminate_prepare(ProjectDescriptor* project)
{
    foreach_fileent(fe_ptr, project) {
	FileEntry *fe = *fe_ptr;
	static MemorySegment file_buffer;
	static MemorySegment setkey_buffer;
	const char* output_buffer;
	int output_len;
	bool setkeys;
	char* checksum;
	RcsVersionData* old_data;

	if(!fe->active_modified())
	    continue;

	If_fail(file_buffer.map_file(*fe->working_path()))
	    throw prcserror << "Error reading file " << squote(fe->working_path())
			     << perror;

	if(fe->keyword_sub()) {
	    If_fail( setkeys << setkeys_inoutbuf(file_buffer.segment(),
						 file_buffer.length(),
						 &setkey_buffer,
						 Unsetkeys) )
		throw prcserror << "Keyword replacement failed on file "
				 << squote(fe->working_path()) << perror;

	    output_buffer = setkey_buffer.segment();
	    output_len = setkey_buffer.length();
	} else {
	    output_buffer = file_buffer.segment();
	    output_len = file_buffer.length();
	}

	checksum = md5_buffer(output_buffer, output_len);

	fe->set_checksum(checksum);
	fe->set_keyed_length(output_len);

	if(fe->descriptor_name()) { /* if it doesn't have an empty descriptor */
	    Return_if_fail(old_data << project->repository_entry() ->
			   lookup_rcs_file_data(*fe->descriptor_name(),
						*fe->descriptor_version_number()));

	    if(output_len == old_data->length()) {

		if(strncmp(checksum, old_data->unkeyed_checksum(), 16) == 0) {

		    DEBUG("Slow elim succeeds for file " << squote(fe->working_path()));

		    if(option_long_format) {
			prcsoutput << "File " << squote(fe->working_path())
				   << " is unmodified" << dotendl;
		    }

		    fe->set_modified(false);
		    continue;
		}
 	    }
	}

	Return_if_fail(fe->initialize_descriptor(project->project_name(),
						 project->repository_entry(),
						 (option_report_actions ? false : true)));

	if(!option_report_actions) {
	    unlink(*fe->temp_file_path());

	    const char* fullpath;

	    if(fe->keyword_sub() && setkeys) {
		FILE* temp_out;

		fullpath = make_temp_file("");

		If_fail(temp_out << Err_fopen(fullpath, "w"))
		    throw prcserror << "Failed opening temp file for writing "
				    << squote(fullpath) << perror;

		if(fwrite(output_buffer, output_len, 1, temp_out) != 1)
		    throw prcserror << "Failed writing temp file "
				    << squote(fullpath) << perror;

		If_fail(Err_fclose(temp_out))
		    throw prcserror << "Failed writing temp file "
				    << squote(fullpath) << perror;
	    } else {
		Return_if_fail(fullpath << name_in_cwd(*fe->working_path()));

	    }

	    if (symlink(fullpath, fe->temp_file_path()->cast()) < 0) {
		throw prcserror << "Error creating symbolic link "
				<< squote(fe->temp_file_path()) << perror;
	    }
	}
    }

    project->repository_entry()->Rep_set_umask();

    /* This frees the registration table. */
    project->repository_entry()->Rep_new_filename(NULL, true, NULL);

    if(!option_report_actions)
	Return_if_fail(VC_register(NULL));

    project->repository_entry()->Rep_unset_umask();

    return NoError;
}

PrVoidError ProjectDescriptor::update_repository()
{
    foreach_fileent(fe_ptr, this) {
        (*fe_ptr)->update_repository(repository_entry());
    }

    ProjectVersionData ver;
    struct stat buf;

    If_fail(Err_stat(project_file_path(), &buf))
	throw prcserror << "Stat failed on newly written project file" << project_file_path() << perror;

    ver.date(get_utc_time_t());
    ver.author(get_login());
    ver.length(buf.st_size);
    ver.rcs_version(_new_rcs_version);
    /* no checksum */
    ver.prcs_major(*project_version_major());
    ver.prcs_minor(*project_version_minor());
    ver.parent_major(*parent_version_major());
    ver.parent_minor(*parent_version_minor());

    if(descends_from_major()) {
	ver.descends_from_major(*descends_from_major());
	ver.descends_from_minor(*descends_from_minor());
    }

    repository_entry()->add_project_data(&ver);

    Return_if_fail(repository_entry()->close_repository());

    return NoError;
}

void FileEntry::update_repository(RepEntry* rep_entry)
{
    if(active_modified()) {
	RcsVersionData ver;
	ver.date(get_utc_time_t());
	ver.author(get_login());
	ver.length(keyed_length());
	ver.unkeyed_checksum(checksum());
	ver.rcs_version(*descriptor_version_number());

	rep_entry->add_rcs_file_data(*descriptor_name(), &ver);
    }
}

static PrVoidError checkin_cleanup_handler(void* data, bool)
{
    ProjectDescriptor* project = (ProjectDescriptor* ) data;

    foreach_fileent(fe_ptr, project) {
	FileEntry* fe = *fe_ptr;

	if(!fe->on_command_line())
	    continue;

	if(fe->temp_file_path() != NULL)
	    unlink(*fe->temp_file_path());
    }

    return NoError;
}

static void kill_one_subdir(const char* string, Dstring& fill)
{
    int len = strip_leading_path(string) - string;

    if(len > 0)
	len -= 1;

    fill.assign(string, len);
}

static PrBoolError try_elim_subdir(FileEntry* fe, MissingFileAction action)
    /* returns whether to skip this file.  It is given that stat() or
     * lstat() failed on this file */
{
    static DstringPtrArray* eliminated_subdirs = NULL;
    struct stat buf;
    const char* name = *fe->working_path();

    if(!eliminated_subdirs)

	eliminated_subdirs = new DstringPtrArray;

    if(!strchr(name, '/'))
	return false;

    Dstring this_subdir;
    Dstring last_subdir;

    kill_one_subdir(name, this_subdir);

    foreach(elim_ptr, eliminated_subdirs, DstringPtrArray::ArrayIterator) {
	Dstring* elim = *elim_ptr;

	if(strncmp(elim->cast(), name, elim->length()) == 0) {
	    switch (action) {
	    case QueryUserRemoveFromCommandLine:
		fe->set_on_command_line(false);
		break;
	    case NoQueryUserRemoveFromCommandLine:
		fe->set_on_command_line(false);
		break;
	    case SetUnmodifiedWarnUser:
		fe->set_modified(false);
		break;
	    }

	    return true;
	}
    }

    if(stat(this_subdir, &buf) >= 0 || errno != ENOENT)
	return false;

    while(this_subdir.length() > 0) {
	last_subdir.assign(this_subdir);

	kill_one_subdir(this_subdir, this_subdir);

	if(stat(this_subdir, &buf) >= 0) {
	    static BangFlag bang;

	    switch (action) {
	    case QueryUserRemoveFromCommandLine:
		prcsquery << "The directory " << squote(last_subdir)
			  << " does not exist.  "
			  << force("Ignoring all entries")
			  << report("Ignore all entries")
			  << optfail('n')
			  << defopt('y', "Ignore all entries of this directory")
			  << query("Ignore it and descendants");

		Return_if_fail(prcsquery.result());

		fe->set_on_command_line(false);
		break;
	    case NoQueryUserRemoveFromCommandLine:
		prcswarning << "The directory " << squote(last_subdir)
			    << " does not exist.  Ignoring all entries" << dotendl;

		fe->set_on_command_line(false);
		break;
	    case SetUnmodifiedWarnUser:
		prcswarning << "The directory " << squote(last_subdir)
			    << " does not exist.  Assuming all entries are unmodified" << dotendl;

		fe->set_modified(false);
		break;
	    }

	    eliminated_subdirs->append(new Dstring(last_subdir));

	    return true;

	} else if(errno != ENOENT) {
	    return false;
	}
    }

    return false;
}

PrVoidError eliminate_working_files(ProjectDescriptor* project, MissingFileAction action)
{
    foreach_fileent(fe_ptr, project) {
	FileEntry* fe = *fe_ptr;

        if(fe->on_command_line()) {
	    static BangFlag bang;
	    bool file_present;
	    bool error = false;

	    If_fail(file_present << fe->check_working_file())
		error = true;

	    if(!error && file_present)
		continue;

	    if(!error) {
		bool ignore_subdir;

		Return_if_fail(ignore_subdir << try_elim_subdir(fe, action));

		if(ignore_subdir)
		    continue;
	    }

	    switch (action) {
	    case QueryUserRemoveFromCommandLine:
		prcsquery << "File " << squote(fe->working_path()) << " is unavailable.  "
			  << force("Continuing")
			  << report("Continue")
			  << allow_bang(bang)
			  << defopt('y', "Ignore this file, as if it were not on the command line")
			  << optfail('n')
			  << query("Ignore this file");

		Return_if_fail(prcsquery.result());
		fe->set_on_command_line(false);
		break;
 	    case NoQueryUserRemoveFromCommandLine:
		prcswarning << "File " << squote(fe->working_path())
			    << " is unavailable.  Continuing" << dotendl;
		fe->set_on_command_line(false);
		break;
	    case SetUnmodifiedWarnUser:
		prcswarning << "File " << squote(fe->working_path())
			    << " is unavailable.  Assuming unmodified" << dotendl;
		fe->set_modified(false);
		break;
	    }
        }
    }

    return NoError;
}
