/* -*-Mode: C++;-*-
 * $Id: diff.cc 1.20.1.19.1.3.1.5 Mon, 02 Sep 1996 20:38:35 -0700 jmacd $
 * diff.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include "prcs.h"
#include "hash.h"
#include "misc.h"
#include "repository.h"
#include "projdesc.h"
#include "checkin.h"
#include "checkout.h"
#include "vc.h"
#include "dstring.h"
#include "fileent.h"
#include "diff.h"
#include "sexp.h"
#include "syscmd.h"
#include "setkeys.h"
#include "system.h"
#include "memseg.h"
#include "rebuild.h"


static PrPrcsExitStatusError diff_working_files();
static PrPrcsExitStatusError diff_named_versions();
static PrBoolError diff_similar_files(ProjectDescriptor* toP,
				      FileTable* fromfiles,
				      ProjectDescriptor* fromP);
static PrBoolError diff_missing_files(ProjectDescriptor* toP,
				      FileTable* tofiles,
				      ProjectDescriptor* fromP,
				      FileTable* fromfiles);
static PrBoolError diff_symlink_pair(FileEntry*, FileEntry*);
static PrBoolError diff_new_file(FileEntry*);
static PrVoidError prepare_diff_workingfile(FileEntry *fe);
static PrIntError try_optimizations(FileEntry *tofe, ProjectDescriptor *toP,
				    FileEntry *fromfe, ProjectDescriptor *fromP);
static PrVoidError prepare_diff_versionfile(FileEntry *fe,
					    ProjectDescriptor*,
					    bool);
static PrBoolError diff_project_files();
static PrBoolError diff_two_files(const char* label1, const char* label2,
				  const char* file1, const char* file2,
				  const char* index);
static PrVoidError get_info(FileEntry* fe, RepEntry* rep_entry);

static const char *from_format, *to_format;

static bool use_working_to_file = true;
static bool diffing_new_files = false;
/* Determined by scanning the diff options, given by -q or --brief,
 * and tells PRCS to fake the diff when it can. */
static bool diff_option_brief = false;

#define temp_file_from temp_file_1
#define temp_file_to temp_file_2

PrPrcsExitStatusError diff_command()
{
    kill_prefix(prcsoutput);

    for(int i = 0; i < cmd_diff_options_count; i += 1)
	if(strcmp("-q", cmd_diff_options_given[i]) == 0 ||
	   strcmp("--brief", cmd_diff_options_given[i]) == 0) {
	    diff_option_brief = true;
	}

    if(cmd_alt_version_specifier_major == NULL)
	return diff_working_files();
    else
	return diff_named_versions();
}

/* reports the differences from VERSION to WORKING files */
static PrPrcsExitStatusError diff_working_files()
{
    FileTable workfiles(pathname_hash, pathname_equal),
	      revfiles(pathname_hash, pathname_equal);
    ProjectDescriptor *From, *To;
    ProjectVersionData *from_version;
    bool diffs = false, diff = false;
    Dstring from_format_ds, to_format_ds;

    use_working_to_file = true;

    if(!fs_file_readable(cmd_root_project_file_path)) {
	throw prcserror << "Can't open file " << squote(cmd_root_project_file_path) << perror;
    }

    /* Set up the project file diff */
    const char* abs_path;

    Return_if_fail(abs_path << absolute_path(cmd_root_project_file_path));

    If_fail(Err_symlink(abs_path, temp_file_to))
	throw prcserror << "Failed creating symlink " << squote(temp_file_to) << perror;

    Return_if_fail(To << read_project_file(cmd_root_project_full_name, temp_file_to));

    Return_if_fail(To->init_repository_entry(cmd_root_project_name, false, false));

    eliminate_unnamed_files(To);

    Return_if_fail(from_version << resolve_version(cmd_version_specifier_major,
						   cmd_version_specifier_minor,
						   cmd_root_project_full_name,
						   cmd_root_project_file_path,
						   To,
						   To->repository_entry()));

    if(from_version->prcs_minor_int() == 0) {
	Return_if_fail(From << checkout_create_empty_prj_file(temp_file_from,
							      cmd_root_project_full_name,
							      from_version->prcs_major()));
    } else {
	Return_if_fail(From << To->repository_entry() ->
		       checkout_create_prj_file(temp_file_from,
						cmd_root_project_full_name,
						from_version->rcs_version()));
    }

    eliminate_unnamed_files(From);

    From->repository_entry(To->repository_entry());

    from_format_ds.sprintf("version %s.%s",
			   From->project_version_major()->cast(),
			   From->project_version_minor()->cast());
    to_format_ds.sprintf("working files of version %s.%s",
			 To->project_version_major()->cast(),
			 To->project_version_minor()->cast());

    from_format = from_format_ds.cast();
    to_format = to_format_ds.cast();

    prcsinfo << "Producing diffs from " << from_format << " to " << to_format << dotendl;

    populate_table(To, &workfiles);
    populate_table(From, &revfiles);

    Return_if_fail(eliminate_working_files(To, NoQueryUserRemoveFromCommandLine));

    To->read_quick_elim();

    To->quick_elim_unmodified();

    if(cmd_prj_given_as_file)
	Return_if_fail(diff << diff_project_files());

    diffs |= diff;

    Return_if_fail(diff << diff_similar_files(To, &revfiles, From));

    diffs |= diff;

    Return_if_fail(diff << diff_missing_files(To, &workfiles, From, &revfiles));

    diffs |= diff;

    Return_if_fail(warn_unused_files(false));

    if(diffs)
	return ExitDiffs;
    else
	return ExitNoDiffs;
}

/* reports the differences from VERSION1 to VERSION2 */
static PrPrcsExitStatusError diff_named_versions()
{
    FileTable workfiles(pathname_hash, pathname_equal),
	      revfiles(pathname_hash, pathname_equal);
    Dstring from_format_ds, to_format_ds;
    ProjectDescriptor *From, *To;
    ProjectVersionData *to_version, *from_version;
    RepEntry *rep_entry;
    bool diffs = false, diff = false;

    use_working_to_file = false;

    Return_if_fail(rep_entry << Rep_init_repository_entry(cmd_root_project_name,
							  false, false, true));

    /* Get the from version */
    Return_if_fail(from_version << resolve_version(cmd_version_specifier_major,
						   cmd_version_specifier_minor,
						   cmd_root_project_full_name,
						   cmd_root_project_file_path,
						   NULL,
						   rep_entry));

    if(from_version->prcs_minor_int() == 0) {
	Return_if_fail(From << checkout_create_empty_prj_file(temp_file_from,
							      cmd_root_project_full_name,
							      from_version->prcs_major()));
    } else {
	Return_if_fail(From << rep_entry ->
		       checkout_create_prj_file(temp_file_from,
						cmd_root_project_full_name,
						from_version->rcs_version()));
    }

    eliminate_unnamed_files(From);

    /* Get the to version */
    Return_if_fail(to_version << resolve_version(cmd_alt_version_specifier_major,
						 cmd_alt_version_specifier_minor,
						 cmd_root_project_full_name,
						 cmd_root_project_file_path,
						 NULL,
						 rep_entry));

    if(to_version->prcs_minor_int() == 0) {
	Return_if_fail(To << checkout_create_empty_prj_file(temp_file_to,
							    cmd_root_project_full_name,
							    to_version->prcs_major()));
    } else {
	Return_if_fail(To << rep_entry ->
		       checkout_create_prj_file(temp_file_to,
						cmd_root_project_full_name,
						to_version->rcs_version()));
    }

    eliminate_unnamed_files(To);

    To->repository_entry(rep_entry);
    From->repository_entry(rep_entry);

    to_format_ds.sprintf("version %s.%s",
			 to_version->prcs_major(),
			 to_version->prcs_minor());
    from_format_ds.sprintf("version %s.%s",
			   from_version->prcs_major(),
			   from_version->prcs_minor());

    from_format = from_format_ds.cast();
    to_format = to_format_ds.cast();

    populate_table(To, &workfiles);
    populate_table(From, &revfiles);

    /* Diff */

    prcsinfo << "Producing diffs from " << from_format << " to " << to_format << dotendl;

    if(cmd_prj_given_as_file)
	Return_if_fail(diff << diff_project_files());

    diffs |= diff;

    Return_if_fail(diff << diff_similar_files(To, &revfiles, From));

    diffs |= diff;

    Return_if_fail(diff << diff_missing_files(To, &workfiles, From, &revfiles));

    diffs |= diff;

    Return_if_fail(warn_unused_files(false));

    if(diffs)
	return ExitDiffs;
    else
	return ExitNoDiffs;
}

void populate_table(ProjectDescriptor* P, FileTable* T)
{
    foreach_fileent(fe_ptr, P) {
	const char* key;

	if (option_file_family) {
	    if (!(*fe_ptr)->descriptor_name()) {
		/* Its an empty file descriptor. */
		continue;
	    }

	    key = (*fe_ptr)->descriptor_name()->cast();
	} else {
	    key = (*fe_ptr)->working_path()->cast();
	}

	T->insert(key, (*fe_ptr));
    }

}

FileEntry* lookup_file(FileEntry* fe, FileTable* table)
{
    FileEntry **fe_ret;

    if (option_file_family) {
	if(fe->descriptor_name()) {
	    fe_ret = table->lookup(fe->descriptor_name()->cast());

	    if(fe_ret) {
		return *fe_ret;
	    } else {
		return NULL;
	    }
	} else {
	    return NULL;
	}
    } else {
	fe_ret = table->lookup(fe->working_path()->cast());

	if(fe_ret) {
	    return *fe_ret;
	} else {
	    return NULL;
	}
    }
}

static PrVoidError get_info(FileEntry* fe, RepEntry* rep_entry)
{
    if(use_working_to_file)
	return fe->get_file_sys_info();
    else
	return fe->get_repository_info(rep_entry);
}

/*
 * prepare_diff_workingfile --
 *
 *     places either a copy of a keywords unset working file or a symlink
 *     to a working file into temp_file_to of the given file entry.
 */
static PrVoidError prepare_diff_workingfile(FileEntry *fe)
{
    If_fail(Err_unlink(temp_file_to)) {
	throw prcserror << "Unlink " << squote(temp_file_to) << " failed" << perror;
    }

    if (fe->file_type() == RealFile) {

	Return_if_fail(fe->get_file_sys_info());

	if (fe->keyword_sub() && !option_diff_keywords) {
	    Return_if_fail (setkeys(*fe->working_path(),
				    temp_file_to,
				    Unsetkeys));
	} else {
	    const char* abs_path;

	    Return_if_fail(abs_path << absolute_path(*fe->working_path()));

	    If_fail (Err_symlink(abs_path, temp_file_to)) {
		throw prcserror << "Error creating symbolic link"
				<< squote(temp_file_to) << perror;
	    }
	}
    }

    return NoError;
}

static PrIntError try_optimizations(FileEntry *tofe, ProjectDescriptor *toP,
				    FileEntry *fromfe, ProjectDescriptor *fromP)
{
    static MemorySegment setkey_buf;
    RcsVersionData *to_version_data, *from_version_data;
    bool need_to_replace_to_keywords = option_diff_keywords && tofe->keyword_sub();
    bool need_to_replace_from_keywords = option_diff_keywords && fromfe->keyword_sub();
    bool same_descriptor;

    if(!tofe->descriptor_name()) /* empty descriptor for working file */
	return 0;

    same_descriptor = (strcmp(*tofe->descriptor_name(),
			      *fromfe->descriptor_name()) == 0 &&
		       strcmp(*tofe->descriptor_version_number(),
			      *fromfe->descriptor_version_number()) == 0);

#define DIFFS   3 /* first bit says that optimizations succeeded and to skip this file */
#define NODIFFS 1 /* second bit says whether there were differences */

    Return_if_fail(to_version_data << toP->repository_entry() ->
		   lookup_rcs_file_data(*tofe->descriptor_name(),
					*tofe->descriptor_version_number()));

    Return_if_fail(from_version_data << fromP->repository_entry() ->
		   lookup_rcs_file_data(*fromfe->descriptor_name(),
					*fromfe->descriptor_version_number()));

    /* If brief diff, compare checksums if possible */
    if(diff_option_brief && !use_working_to_file &&
       !need_to_replace_to_keywords && !need_to_replace_from_keywords) {

	if(strncmp(to_version_data->unkeyed_checksum(),
		   from_version_data->unkeyed_checksum(), 16) == 0) {
	    return NODIFFS;
	} else {
	    prcsoutput << "The file " << diff_tuple(fromfe, tofe) << " differs" << prcsendl;
	    return DIFFS;
	}
    }

    if(use_working_to_file && same_descriptor && !need_to_replace_to_keywords) {
	/* If the working file is unmodifed by quick_elim */
	if(!tofe->modified())
	    return NODIFFS;
    }

    if(use_working_to_file && !need_to_replace_to_keywords &&
       !need_to_replace_from_keywords) {

	if(tofe->keyword_sub()) {
	    Return_if_fail(setkeys_outbuf(*tofe->working_path(), &setkey_buf, Unsetkeys));
	} else {
	    If_fail(setkey_buf.map_file(*tofe->working_path()))
		throw prcserror << "Read failure on " << squote(tofe->working_path())
				<< perror;
	}

	if(setkey_buf.length() == from_version_data->length()) {
	    const char* checksum;

	    checksum = md5_buffer(setkey_buf.segment(), setkey_buf.length());

	    if(strncmp(checksum, from_version_data->unkeyed_checksum(), 16) == 0)
		return NODIFFS;
	}

	/* Checksums differ, if brief diff */
	if(diff_option_brief) {
	    prcsoutput << "The file " << diff_tuple(fromfe, tofe) << " differs" << prcsendl;
	    return DIFFS;
	}
    }

    return 0;
}

/*
 * prepare_diff_versionfile --
 *
 *     writes the file onto either temp_file_from or temp_file_to and
 *     maybe replaces keywords
 */
static PrVoidError prepare_diff_versionfile(FileEntry *fe,
					    ProjectDescriptor *P,
					    bool use_temp_file_from)
{
    const char* filename;

    if(use_temp_file_from)
	filename = temp_file_from;
    else
	filename = temp_file_to;

    Return_if_fail(fe->initialize_descriptor(P->project_name(), P->repository_entry(), false));

    Return_if_fail(fe->get_repository_info(P->repository_entry()));

    if (fe->keyword_sub() && option_diff_keywords) {
	FILE* cofile;

	Return_if_fail(cofile << VC_checkout_stream(*fe->descriptor_version_number(),
						    *fe->full_descriptor_name()));

	make_keyword_pair_list(fe, P);

	Return_if_fail(setkeys_infile(fileno(cofile), filename, Setkeys));

	Return_if_fail(VC_close_checkout_stream(cofile,
						*fe->descriptor_version_number(),
						*fe->full_descriptor_name()));

    } else {
	Return_if_fail(VC_checkout_file(filename,
					*fe->descriptor_version_number(),
					*fe->full_descriptor_name()));
    }

    return NoError;
}

static PrBoolError diff_similar_files(ProjectDescriptor* toP,
				      FileTable* fromfiles,
				      ProjectDescriptor* fromP)
{
    FileEntry *tofe, *fromfe;
    bool alldiffs = false, onediff = false;

    foreach_fileent(fe_ptr, toP) {
	tofe = *fe_ptr;

	if(!tofe->on_command_line())
	    continue;

	fromfe = lookup_file(tofe, fromfiles);

	if(fromfe) {

	    if(fromfe->file_type() == tofe->file_type()) {
		if(fromfe->file_type() == RealFile) {
		    int no_diffs;

		    /* If they are the same by checksum */
		    Return_if_fail(no_diffs << try_optimizations(tofe, toP, fromfe, fromP));

		    if(no_diffs & 1) {
			if(no_diffs & 2) {
			    alldiffs = true;
			    DEBUG("Opt succeeds for " << diff_tuple(tofe, fromfe) << ", diffs");
			} else {
			    DEBUG("Opt succeeds for " << diff_tuple(tofe, fromfe) << ", no diffs");
			}

			continue;
		    }

		    if(use_working_to_file) {
			Return_if_fail(prepare_diff_workingfile(tofe));
		    } else {
			Return_if_fail(prepare_diff_versionfile(tofe, toP, false));
		    }

		    Return_if_fail(prepare_diff_versionfile(fromfe, fromP, true));

		    Return_if_fail(onediff << diff_pair(fromfe,
							tofe,
							temp_file_from,
							temp_file_to,
							from_format,
							to_format));

		} else if(fromfe->file_type() == SymLink) {
		    Return_if_fail(tofe->initialize_descriptor(cmd_root_project_name,
							       toP->repository_entry(),
							       false));

		    Return_if_fail(fromfe->initialize_descriptor(cmd_root_project_name,
								 fromP->repository_entry(),
								 false));

		    Return_if_fail(onediff << diff_symlink_pair(tofe, fromfe));
		}
		/* directories ignored */
	    } else {
		prcsoutput << "Changes type from " << format_type(fromfe->file_type())
			   << " to " << format_type(tofe->file_type())
			   << ": " << diff_tuple(fromfe, tofe) << prcsendl;

		onediff = true;
	    }

	    alldiffs |= onediff;
	}
    }

    return alldiffs;
}

static PrBoolError diff_missing_files(ProjectDescriptor* toP,
				      FileTable* tofiles,
				      ProjectDescriptor* fromP,
				      FileTable* fromfiles)
{
    bool alldiffs = false;
    FileEntry *fe;

    diffing_new_files = true;

    foreach_fileent(fe_ptr, toP) {
	fe = *fe_ptr;

	if(!fe->active_modified())
	    continue;

	if(!lookup_file(fe, fromfiles)) {
	    Return_if_fail(fe->initialize_descriptor(cmd_root_project_name,
						     toP->repository_entry(),
						     false));

	    Return_if_fail(get_info(fe, toP->repository_entry()));

	    if(option_diff_new_file && fe->file_type() == RealFile) {
		if(use_working_to_file)
		    Return_if_fail(prepare_diff_workingfile(fe));
		else
		    Return_if_fail(prepare_diff_versionfile(fe, toP, false));

		diff_new_file(fe);
	    } else {
		prcsoutput << "Only in " << to_format << ": " << fe->working_path() << prcsendl;
	    }
	    alldiffs = true;
	}
    }

    foreach_fileent(fe_ptr2, fromP) {
	fe = *fe_ptr2;

	if(!fe->active_modified())
	    continue;

	if(!lookup_file(fe, tofiles)) {
	    prcsoutput << "Only in " << from_format << ": " << fe->working_path() << prcsendl;
	    alldiffs = true;
	}
    }

    diffing_new_files = false;

    return alldiffs;
}

static PrBoolError diff_symlink_pair(FileEntry* to, FileEntry* from)
{
    const char *link1, *link2;

    link1 = *from->link_name();
    link2 = *to->link_name();

    if(!pathname_equal(link1, link2)) {
	prcsoutput << "Symlink " << diff_tuple(from, to)
	     << " changes from " << squote(link1)
	     << "' to " << squote(link2) << prcsendl;
	return true;
    } else {
	return false;
    }
}

static PrBoolError diff_new_file(FileEntry *to)
{
    Dstring label1, label2;
    Dstring filedes;

    if(!to->descriptor_name()) {
	filedes.assign("()");
    } else {
	filedes.sprintf("(%s %s %s)",
			to->working_path()->cast(),
			to->internal_descriptor()->cast(),
			to->file_mode()->cast());
    }

    label1.sprintf("-L%s.orig %s %s ()",
		   to->working_path()->cast(),
		   get_utc_time(),
		   get_login());

    label2.sprintf("-L%s      %s %s %s",
		   to->working_path()->cast(),
		   to->last_mod_date(),
		   to->last_mod_user(),
		   filedes.cast());

    return diff_two_files(label1, label2, "/dev/null",
			  temp_file_to, *to->working_path());
}

void format_diff_label(FileEntry* fe, const char* name,
		       const char* version, Dstring* ds)
{
    Dstring desc;

    const char* date, *user;

    if(fe == NULL) {
	date = get_utc_time();
	user = get_login();
    } else {
	date = fe->last_mod_date();
	user = fe->last_mod_user();
    }

    if(!fe->descriptor_name()) {
	desc.assign("()");
    } else {
	desc.sprintf("(%s %s %s)",
		     fe->working_path()->cast(),
		     fe->internal_descriptor()->cast(),
		     fe->file_mode()->cast());
    }

    ds->sprintf("-L%s %s %s %s %s", name, date, user, desc.cast(), version);
}

PrBoolError diff_pair(FileEntry* from,
		      FileEntry* to,
		      const char* from_file,
		      const char* to_file,
		      const char* to_version,
		      const char* from_version)
{
    Dstring name1, name2, label1, label2;

    name1.sprintf("%s.orig", from->working_path()->cast());
    name2.sprintf("%s     ", to->working_path()->cast());

    format_diff_label(from, name1, from_version, &label1);
    format_diff_label(to, name2, to_version, &label2);

    return diff_two_files(label1, label2, from_file,
			  to_file, *to->working_path());
}

static PrBoolError diff_two_files(const char* label1, const char* label2,
				  const char* file1, const char* file2,
				  const char* index)
{
    ArgList *args;
    char buf[512];
    FILE* output;
    bool n_index_written(true);
    int c, ret;

    Return_if_fail(args << gdiff_command.new_arg_list());

    for(int i = 0; i < cmd_diff_options_count; i += 1)
	args->append(cmd_diff_options_given[i]);

    args->append("-a");
    args->append(label1);
    args->append(file1);
    args->append(label2);
    args->append(file2);

    Return_if_fail(gdiff_command.open(true, false));

    output = gdiff_command.standard_out();

    while((c = fread(buf, 1, 512, output)) > 0 ) {
	if(n_index_written) {
	    /* this sorta depends on diffs output, a leading Files
	     * means its in --brief mode, lets filter the filenames */
	    if(strncmp(buf, "Files ", 6) == 0) {
		if(diffing_new_files)
		    prcsoutput << "Only in " << to_format << ": " << index << prcsendl;
		else
		    prcsoutput << "The file " << squote(index) << " differs" << prcsendl;
		break;
	    }
	    prcsoutput << "Index: " << index << prcsendl;
	    n_index_written = false;
	}
	fwrite(buf, 1, c, stdout);
    }

    if(ferror(output))
	throw prcserror << "Error reading from diff output pipe" << perror;

    Return_if_fail(ret << gdiff_command.close());

    if(ret > 1)
	throw prcserror << "Gdiff exited abnormally on files "
			 << squote(file1) << " and " << squote(file2) << dotendl;

    return (bool)ret;
}

static PrBoolError diff_project_files()
{
    Dstring label1, label2;

    label1.sprintf("-L%s.prj.orig %s", cmd_root_project_name, from_format);
    label2.sprintf("-L%s.prj      %s", cmd_root_project_name, to_format);

    return diff_two_files(label1, label2, temp_file_from,
			  temp_file_to, cmd_root_project_file_path);
}
