/* -*-Mode: C++;-*-
 * $Id: dynarray.cc 1.2.1.1.1.5.1.8 Tue, 20 Aug 1996 17:09:46 -0700 jmacd $
 * dynarray.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include <std.h>
#include <stdarg.h>
#include "config.h"
#include "dynarray.h"
#include "dstring.h"
#include "utils.h"

#define generic template<class T, int DefaultSize, bool ZeroTerm>
#define selftype Dynarray<T, DefaultSize, ZeroTerm>
#define member selftype::

#define MAYBECNEWVEC(zeroit, type, elts) ((zeroit) ? NEWVEC0(type, elts) : NEWVEC(type, elts))
#define MAYBECEXPANDVEC(zeroit, T, V, N, I) (T*)((zeroit) ? EXPANDVEC0(V, N, I) : EXPANDVEC(V, N, I))

generic member Dynarray()
    :vec(MAYBECNEWVEC(ZeroTerm, T, DefaultSize)),
     filled(0), alloc(DefaultSize)
{}

generic member Dynarray(const T *s, int N0)
{
    unsigned int i(1), N(N0);

    ASSERT(N0 < 0 || ZeroTerm, "Can't end a non-zero-terminated vector");

    if(N0 < 0) {
	const T* t = s;
	N = 0;
	while(*t != 0) {
	    t += 1;
	    N += 1;
	}
    }

    while(i <= N) { i = i << 1; }

    vec = MAYBECNEWVEC(ZeroTerm, T, i);
    filled = N;
    alloc = i;

    for(i = 0; i < N; i += 1)
	vec[i] = s[i];
}

generic member Dynarray(int N0)
{
    unsigned int i = 1, N = N0;

    while(i <= N) { i = i << 1; }

    vec = MAYBECNEWVEC(ZeroTerm, T, i);
    filled = 0;
    alloc = i;
}

generic member Dynarray(const T c, int N0)
{
    unsigned int i(1), N(N0);

    ASSERT(N0 < 0 || ZeroTerm, "Can't end a non-zero-terminated vector");

    while(i <= N) { i = i << 1; }

    vec = MAYBECNEWVEC(ZeroTerm, T, i);
    filled = N;
    alloc = i;

    for(i = 0; i < N; i += 1)
	vec[i] = c;
}

generic member Dynarray(const selftype& s)
{
    unsigned int i(1), N(s.length());
    while(i <= N) { i = i << 1; }

    vec = MAYBECNEWVEC(ZeroTerm, T, i);
    filled = N;
    alloc = i;

    for(i = 0; i < N; i += 1)
	vec[i] = s[i];
}

generic member ~Dynarray()
{
    free(vec);
}

generic void member maybe_expand(int elts)
{
    if (filled + elts >= (ZeroTerm ? alloc : alloc + 1)) {
	unsigned int newsize(filled + elts + (ZeroTerm ? 1 : 0));
	unsigned int goalsize(alloc);
	while(goalsize < newsize) { goalsize = goalsize << 1; }
	vec = MAYBECEXPANDVEC(ZeroTerm, T, vec, alloc, goalsize - alloc);
	alloc = goalsize;
    }
}

generic void member expand(int elts, const T& val)
{
    if(filled < elts) {
	maybe_expand(elts - filled);
	while(filled < elts) {
	    vec[filled] = val;
	    filled += 1;
	}
    }
}

generic selftype& member append(const T c)
{
    maybe_expand(1);
    vec[filled] = c;
    filled +=1 ;
    return *this;
}

generic selftype& member append(const T *s, int N0)
{
    const T* c(s);
    unsigned int i = 0, N(N0);

    ASSERT(s != NULL, "tried to dereference null pointer");

    ASSERT(N0 < 0 || ZeroTerm, "Can't end a non-zero-terminated vector");

    if(ZeroTerm && N0 < 0) {
	while(*c != 0) {
	    append(*c);
	    c += 1;
	}
    } else {
	while( i < N ) {
	    append(*c);
	    c += 1;
	    i += 1;
	}
    }

    return *this;
}

generic selftype& member append(const selftype& s)
{
    int i(0), N(s.length());

    while( i < N ) {
	append(s[i]);
	i += 1;
    }

    return *this;
}

generic selftype& member prepend(const T c)
{
    maybe_expand(1);
    memmove(vec + 1, vec, sizeof(c) * filled);
    vec[0] = c;
    filled += 1;
    return *this;
}

generic member operator const T*() const
{
    ASSERT(this != NULL, "no null objects please");
    if(ZeroTerm) {
	ASSERT(vec[filled] == 0, "not zero-terminated");
    }
    return vec;
}

generic const T* member cast() const
{
    return operator const T*();
}

generic selftype& member truncate(unsigned int N)
{
    filled = N;
    if(ZeroTerm) {
	memset(vec + N, 0, sizeof(T) * (alloc - N));
    }
    return *this;
}

generic int member length() const
{
    return filled;
}

generic T& member index(int N) const
{
    ASSERT(N < filled, "indexed past end of array");
    return vec[N];
}

generic T& member index(int N, const T& elt)
{
    ASSERT(N < filled, "indexed past end of array");
    vec[N] = elt;
    return vec[N];
}

generic selftype& member assign(const T* s, int N)
{
    if (N < 0) {
	N = 0;
	const T* s_c = s;
	while (*s_c++) { N += 1; }
    }

    if (N > filled)
	maybe_expand(N - filled);

    memmove(vec, s, N * sizeof(*s));

    if (ZeroTerm) {
	int old_filled = filled;
	filled = N;
	if (filled < old_filled)
	    memset(vec + filled, 0, sizeof(*s) * (old_filled - filled));
    } else {
	filled = N;
    }

    return *this;
}

generic selftype& member assign(const selftype& s)
{
    return assign(s.cast(), s.length());
}

generic void member sort(int (*compare)(const void*, const void*))
{
    qsort(vec, filled, sizeof(T), compare);
}

#undef member
#undef generic
#undef selftype

#if EXPLICIT_TEMPLATES
#include "dynarray.tl"
#endif
