/* -*-Mode: C++;-*-
 * $Id: fileent.cc 1.9.1.1.1.7.1.13.1.5 Thu, 29 Aug 1996 11:37:17 -0700 jmacd $
 * fileent.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


extern "C" {
#include <pwd.h>
#include <utime.h>
}

#include "prcs.h"

#include "diff.h"
#include "fileent.h"
#include "sexp.h"
#include "repository.h"
#include "projdesc.h"
#include "misc.h"
#include "system.h"
#include "vc.h"

FileEntry::FileEntry()
{
    _mtime = -1;
    _on_command_line = true;
    _modified = true;
    _keywords = true;
    _file_type = RealFile;
    _link_name = NULL;
    _user = NULL;
    _date = NULL;
    _deleted = false;
    _tags = NULL;

    _rcs_version_string = NULL;
    _internal_descriptor = NULL;
    _file_mode = NULL;
    _descriptor_name = NULL;

    _temp_path = new Dstring;
    _full_rep_path = new Dstring;

    _initialized = false;
}

FileEntry::~FileEntry()
{
    delete _temp_path;
    delete _full_rep_path;
    delete _work_path;

    if(_link_name) delete _link_name;
    if(_rcs_version_string) delete _rcs_version_string;
    if(_file_mode) delete _file_mode;
    if(_descriptor_name) delete _descriptor_name;
    if(_internal_descriptor) delete _internal_descriptor;
    if(_tags) delete _tags;
}

void FileEntry::append_tag(const char* tag)
{
    if (option_record_tags) {
	if (!_tags) {
	    _tags = new Dstring(tag);
	} else {
	    _tags->append(' ');
	    _tags->append(tag);
	}
    }
}

PrVoidError FileEntry::initialize_descriptor(const char* projname, RepEntry* rep, bool regist)
{
    if (_initialized)
	return NoError;
    else
	_initialized = true;

    if(file_type() == Directory)
	return NoError;

    if(file_type() == SymLink) {
	if(_link_name == NULL) {
	    if (_file_descriptor->length() == 1 && !_file_descriptor->car()->is_pair()) {

		/* memory allocated */
		_link_name = correct_path(*_file_descriptor->car()->key());

		return NoError;

	    } else {
		throw prcserror << "No link name for working path "
				 << squote(_work_path) << dotendl;
	    }
	} else {
	    nullify_descriptor();

	    if (_link_name != NULL) {
		Dstring *fd(_file_descriptor->car()->key());
		Dstring *prot_path = protect_path(*_link_name);

		fd->assign(*prot_path);
	    }

	    return NoError;
	}
    }

    if (_file_descriptor->is_empty()) {

	const char *filename(strip_leading_path(*working_path()));

	_descriptor_name = new Dstring;

	rep->Rep_new_filename(filename, regist, _descriptor_name);
	rep->Rep_name_in_entry(*_descriptor_name, _full_rep_path);

	_temp_path->assign(*_full_rep_path);
	_temp_path->truncate(_full_rep_path->length() - 2); /* hack off the ,v */

	Dstring prot_path;

	protect_path(*_descriptor_name, &prot_path);

	_internal_descriptor->sprintf("%s/%s", projname, prot_path.cast());
	_internal_descriptor->truncate(_internal_descriptor->length() - 2); /* hack off the ,v */

    } else {
	const char* vc_file_path;

	Return_if_fail(vc_file_path << rep->Rep_prepare_file(*_descriptor_name));

	_full_rep_path->assign(vc_file_path);

	_temp_path->assign(vc_file_path);
	_temp_path->truncate(_temp_path->length() - 2); /* hack off the ,v */
    }

    return NoError;
}

void FileEntry::update_project_file()
{
    if(_deleted)
	return;

    if(on_command_line()) {
	Sexp* s = _file_descriptor;

	nullify_descriptor();

	while(!s->is_empty())
	    s = s->cdr();

	Dstring* fd = s->empty_string();

	switch (file_type()) {
	case RealFile:
	    fd->append(*_internal_descriptor);
	    fd->append(' ');
	    fd->append(*_rcs_version_string);
	    fd->append(' ');
	    fd->append(*_file_mode);
	    break;
	case SymLink:
	    protect_path(*_link_name, fd);
	    break;
	case Directory:
	    /* nothing */
	    break;
	}
    }
}

PrBoolError FileEntry::check_working_file()
{
    const char* name = *working_path();

    /* see that its there and that its readable */
    if(Failure(stat()) || (file_type() == RealFile && !fs_file_readable(name))) {
	if(errno != ENOENT)
	    prcswarning << "Can't stat " << squote(name) << perror;

	return false;
    }

    if(file_type() == Directory && !S_ISDIR(stat_mode()))
	throw prcserror << "Named directory is not a directory: "
			 << squote(name) << dotendl;

    if (file_type() == SymLink) {
	const char* lname;

	if(!S_ISLNK(stat_mode()))
	    throw prcserror << "Named symlink is not a symlink: "
			     << squote(name) << dotendl;

	Return_if_fail(lname << read_sym_link(name));

	_link_name = new Dstring(lname);

	return true;
    }

    if(!_file_mode)
	_file_mode = new Dstring;

    _file_mode->sprintf("%03o", 0777 & _mode);

    /* now it should be a real file or hard link */
    if (file_type() == RealFile && !(S_ISREG(stat_mode()) || S_ISLNK(stat_mode())))
	throw prcserror << "Working file is not a regular file" << squote(name) << perror;

    return true;
}

static void delete_trailing_whitespace(Dstring* s)
{
    int index = s->length() - 1;

    while(index >= 0 && isspace(s->index(index))) { index -= 1; }

    s->truncate(index + 1);
}

void FileEntry::nullify_descriptor()
{
    Sexp *s = _file_descriptor;

    while(!s->is_empty()) {
	s->car()->key()->truncate(0);
	delete_trailing_whitespace(s->car()->prev()->key());
	s = s->cdr();
    }
}

PrVoidError FileEntry::get_repository_info(RepEntry* rep_entry)
{
    RcsVersionData* data;

    Return_if_fail(data << rep_entry->lookup_rcs_file_data(*descriptor_name(), *descriptor_version_number()));

    _mtime = data->date();
    set_user(data->author());
    set_date(intern(time_t_to_rfc822(_mtime)));

    return NoError;
}

PrVoidError FileEntry::get_file_sys_info()
{
    If_fail(stat())
	throw prcserror << "Stat failed on file "
			 << squote(working_path()->cast()) << perror;

    set_user(intern(getpwuid(getuid())->pw_name));
    set_date(intern(time_t_to_rfc822(_mtime)));

    return NoError;
}

NprVoidError FileEntry::stat()
{
    struct stat buf;

    if (file_type() == SymLink) {
	if (::lstat(*working_path(), &buf) < 0)
	    return NonFatalError;
    } else {
	if (::stat(*working_path(), &buf) < 0)
	    return NonFatalError;
    }

    _mode = buf.st_mode;
    _mtime = buf.st_mtime;
    _owner_uid = buf.st_uid;
    _file_size = buf.st_size;
    _ino = buf.st_ino;

    return NoError;
}

PrVoidError FileEntry::chmod(time_t mod_time) const
{
    mode_t new_mode = 0;

    if(!_file_mode)
	new_mode = 0644;
    else {
	const char* ptr = _file_mode->cast();
	char c;

	if(strlen(ptr) != 3)
	    ptr = "644";

	while((c = *ptr++) != 0) {
	    if(c < '0' || c > '7')
		c = '6';

	    new_mode += c - '0';
	    if(*ptr)
		new_mode <<= 3;
	}
    }

    if(!option_preserve_permissions)
	new_mode &= (0777 ^ get_umask());

    If_fail(Err_chmod(*working_path(), new_mode)) {
	prcswarning << "Chmod failed on " << squote(working_path()) << perror;
    }

    struct utimbuf ut;

    ut.modtime = mod_time;
    ut.actime = get_utc_time_t();

    if(utime(*working_path(), &ut) < 0)
       prcswarning << "Utime failed on " << squote(working_path()) << perror;

    return NoError;
}
