/* -*-Mode: C++;-*-
 * $Id: repository.h 1.9.1.18 Tue, 03 Sep 1996 16:01:28 -0700 jmacd $
 * repository.h:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#ifndef _REPOSITORY_H_
#define _REPOSITORY_H_

extern NprConstCharPtrError Rep_guess_repository_path();

PrRepEntryPtrError Rep_init_repository_entry(const char*,
					     bool writeable,
					     bool create,
					     bool require_db);

PrBoolError Rep_repository_entry_exists(const char* entry);

class RepEntry {
public:
    RepEntry() { }

    /*
     * Rep_prepare_file --
     *
     *     returns a filename if file exists, or if file.gz exists and can be
     *     uncompressed.  returns NULL if the uncompression fails or
     *     file{,gz} does not exist.  the filename returned is the named of
     *     the uncompressed file, which will be different if the file was
     *     compressed and the repository isn't writable.
     */
    PrConstCharPtrError Rep_prepare_file(const char* file);

    /*
     * Rep_compress_all_files --
     *
     *     compresses all files with a ,v extension.  needs write lock.
     */
    PrVoidError Rep_compress_all_files();

    /*
     * Rep_uncompress_all_files --
     *
     *     uncompresses all files with a .gz extension.  needs write lock.
     */
    PrVoidError Rep_uncompress_all_files();

    /* Rep_needs_recompress --
     *
     *     true if repository should be recompressed
     */
    bool Rep_needs_compress() const { return writeable && compressed; }

    /*
     * Rep_(un)set_umask --
     *
     *     sets the process umask such that files written will have the right
     *     permission for the repository.  unset returns the mask to its
     *     original value.
     */
    void Rep_set_umask() const { umask(rep_mask); }
    void Rep_unset_umask() const { umask(user_mask); }
    PrVoidError Rep_chmod(mode_t mode);
    PrVoidError Rep_chgrp(const char* group);

    PrVoidError Rep_lock(bool write_lock);

    /*
     * Rep_name_in_entry --
     *
     *     these should be used only when askng for a path to create a new
     *     file in a writeable repository.
     */
    const char* Rep_entry_path();
    const char* Rep_name_in_entry(const char* file);
    void Rep_name_in_entry(const char* file, Dstring* name);

    /*
     * Rep_set_compression --
     *
     *     this is used to change the compression status of a repsitory
     *     entry after initialization.
     */
    void Rep_set_compression(bool c) { compressed = c; }

    const char* Rep_name_of_version_file() const { return project_file.cast(); }

    /* returns in new_filename the name of a file in the repository
     * which is unique. */
    void Rep_new_filename(const char* base, bool regist, Dstring* new_filename);

    PrVoidError init(const char* name, bool write, bool create, bool require);

    /* Rep_unlock_repository --
     *
     *     unlocks the repository, if a lock has been obtained.  */
    void Rep_unlock_repository();
    PrVoidError Rep_print_access();

    ProjectVersionData* lookup_version(const char* major, const char* minor) const;
    int highest_major_version() const;
    int version_count() const { return project_data_array->length(); }
    int highest_minor_version(const char* major) const; /* 0 if none */
    ProjectVersionData* highest_minor_version_data(const char* major) const;
    bool major_version_exists(const char* major) const;
    ProjectVersionData* common_version(const char* major1,
				       int minor1,
				       const char* major2,
				       int minor2) const;
    ProjectVersionData* most_distant_ancestor(const char* from,
					      int from,
					      const char* to,
					      int to) const;
    int parent_index(int index) const;

    ProjectVersionDataPtrArray* project_summary() const { return project_data_array; }
    void project_summary(ProjectVersionDataPtrArray* pda0) { project_data_array = pda0; }

    RcsFileTable* rcs_file_summary() const { return rcs_file_table; }
    void rcs_file_summary(RcsFileTable* rft0) { rcs_file_table = rft0; }

    void add_project_data(ProjectVersionData* new_data);
    void add_rcs_file_data(const char* filename, RcsVersionData* new_data);

    PrVoidError close_repository();

    PrRcsVersionDataPtrError lookup_rcs_file_data(const char* filename, const char* vesion_num) const;

    PrProjectDescriptorPtrError checkout_prj_file(const char* fullname,
						  const char* rcs_version);

    PrProjectDescriptorPtrError checkout_create_prj_file(const char* filename,
							 const char* fullname,
							 const char* rcs_version);

    const char* Rep_owner_name() const { return owner_name; }
    const char* Rep_group_name() const { return group_name; }

private:
    /*
     * Rep_file_exists --
     *
     *     this returns true if file or file.gz is in the repository
     */
    bool Rep_file_exists(const char* name);
    PrVoidError Rep_make_tag();
    PrVoidError Rep_verify_tag(UpgradeRepository* upgrades);
    PrVoidError Rep_create_repository_entry();

    ProjectVersionDataPtrArray *project_data_array;
    RcsFileTable *rcs_file_table;
    RebuildFile *rebuild_file;

    AdvisoryLock* _repository_lock;
    Dstring base;
    Dstring project_file;
    Dstring entry_name;
    int baselen;
    mode_t rep_mask, user_mask;
    bool writeable;
    bool compressed;

    struct stat rep_stat_buf;
    const char* owner_name;
    const char* group_name;
};

#endif
