/* -*-Mode: C++;-*-
 * $Id: syscmd.h 1.5.1.1.1.8 Sun, 11 Aug 1996 16:28:57 -0700 jmacd $
 * syscmd.h:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#ifndef _SYSCMD_H_
#define _SYSCMD_H_

class PipeRec {
public:
    FILE* standard_out;
    FILE* standard_err;
};

/* Similar to STDIN_FILENO and STDOUT_FILENO for pipes, I use these to
 * keep from remember which fileno is which. */
#define PIPE_WRITE_FD 1
#define PIPE_READ_FD 0

extern char *const null_environment[];

/* the following two commands fork a child and execute a command
 * without invoking a shell.  the first assigns the FILE* as standard
 * input for the command, the second copies its first arg into a pipe
 * as standard input.  It closes fd after waiting for the child. */
extern PrExitStatusError make_pipe_file_in(int fd, const char* const*, bool);
extern PrExitStatusError make_pipe_stdout(const char* const*);

/* the following command takes an argv[] returns a pointer to a pipe
 * record which has opened FILE*s to the standard error and standard
 * output of the command which is executed.  to get a return value
 * from the command, close_pipe should be called.  PIPE_OPEN_FAILURE
 * is returned if fork() or pipe() or exec() fails */
extern PrPipeRecPtrError make_pipe_out(const char* const*, PipeRec*, bool, bool);

/* takes as argument a PipeRec which must have been created with
 * make_pipe_out(), and flushes both output streams, then waits for
 * the command to terminate, returning PIPE_CLOSE_FAILURE if the pipe
 * terminated due to some signal or the exit value of the last process
 * in the pipe.  */
extern NprExitStatusError close_pipe(PipeRec* pr);

class SystemCommand {
public:
    /* Initializes a system command but does not stat the executable
     * until the first open command is called or until the init()
     * method is called.  See comment above for naming convention of
     * this initializer. */
    SystemCommand(const char* name);

    /* Forks and execs the command using fd as the standard input and
     * closing the standard output if close is true */
    PrExitStatusError open_filein(int fd, bool close);
    /* Forks and execs the command leaving the standard output and
     * standard error open. */
    PrExitStatusError open_stdout();
    /* Forks and execs the command returning a PipeRec containing
     * fdopened FILE*s for the standard output and standard error
     * streams. */
    PrVoidError open(bool out, bool err);

    FILE* standard_out() const;
    FILE* standard_err() const;

    /* Initializes an array with argv[0] of a new argument list.
     * Arguments should be apended to this before an open command is
     * called */
    PrArgListPtrError new_arg_list();

    /* Computes the full path name of the command and then stats it to
     * see that the current use has execute permission.  path_env is the
     * name of an environment variable in which to find the executable,
     * defaults to RCS_PATH. */
    PrVoidError init();
    /* Assumes init() was successful, and returns the full pathname */
    const char* path() const;

    /* Assuming already open(), wait() for the child to exit and
     * return the exit status.  An error can arise if the child does
     * not terminate due to an exit() call, if the child is stoped
     * by a signal, or if not open(). */
    PrExitStatusError close();

private:

    ArgList argl;
    const char *name;
    Dstring fp;
    PipeRec pr;
};

extern SystemCommand rcs_command;
extern SystemCommand ci_command;
extern SystemCommand co_command;
extern SystemCommand rlog_command;
extern SystemCommand tar_command;
extern SystemCommand gdiff3_command;
extern SystemCommand gdiff_command;
extern SystemCommand ls_command;
extern SystemCommand gzip_command;

extern void abort_child(const char* argv0);

#endif
