/* -*-Mode: C++;-*-
 * $Id: info.cc 1.11.1.1.1.15.1.2 Mon, 02 Sep 1996 20:38:35 -0700 jmacd $
 * info.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */

extern "C" {
#if USE_SYS_FNMATCH
#include <fnmatch.h>
#else
#include "fnmatch.h"
#endif
}

#include "prcs.h"
#include "repository.h"
#include "dstring.h"
#include "vc.h"
#include "fileent.h"
#include "projdesc.h"
#include "misc.h"
#include "checkout.h"
#include "checkin.h"


static PrVoidError print_info(ProjectVersionData* project_data, RepEntry *rep_entry);
static bool matches(const char* pat, const char* text);

static ProjectVersionData* create_pseudo_empty_version(const char* majname)
{
    ProjectVersionData *it = new ProjectVersionData();

    it->prcs_major(majname);
    it->prcs_minor("0");
    it->date(get_utc_time_t());
    it->author(get_login());

    return it;
}

static ProjectVersionData* create_pseudo_latest_version(ProjectVersionData* thismatch)
{
    ProjectVersionData *it = new ProjectVersionData();

    it->prcs_major(thismatch->prcs_major());
    it->prcs_minor("@"); /* this is a little bogus */
    it->rcs_version(thismatch->rcs_version());
    it->date(thismatch->date());
    it->author(thismatch->author());

    return it;
}

static ProjectDescriptor* create_pseudo_empty_projdesc(const char* majnam)
{
    ProjectDescriptor* it = fake_project_descriptor();

    it->project_version_name()->assign(cmd_root_project_name);
    it->project_version_major()->assign(majnam);
    it->parent_version_major()->assign(majnam);

    return it;
}

PrPrcsExitStatusError info_command()
{
    int versions;
    bool first = true;
    RepEntry *rep_entry;
    ProjectVersionDataPtrArray* project_data_array = new ProjectVersionDataPtrArray;
    const char* major_pat = cmd_version_specifier_major;
    const char* minor_pat = cmd_version_specifier_minor;
    ProjectDescriptor* project;

    if (option_really_long_format)
	option_record_tags = 1;

    kill_prefix(prcsoutput);

    Return_if_fail(rep_entry << Rep_init_repository_entry(cmd_root_project_name,
							  false, false, true));

    project_data_array->assign(*rep_entry->project_summary());
    versions = project_data_array->length();

    if(!major_pat[0] || !minor_pat[0]) {
	/*  we need a project descriptor */
	If_fail(project << read_project_file(cmd_root_project_full_name,
					     cmd_root_project_file_path))
	    throw prcserror << "Failed reading project file, can't resolve "
	       "null version" << dotendl;

	if(!major_pat[0])
	    major_pat = *project->project_version_major();

	if(!minor_pat[0])
	    minor_pat = *project->project_version_minor();
    }

    for(int i = 0; i < versions; i += 1) {

        if(project_data_array->index(i) == NULL)
	    /* Its already been printed */
            continue;

        if(matches(major_pat, project_data_array->index(i)->prcs_major())) {

            ProjectVersionData *thismatch = project_data_array->index(i);

            /* Append the empty version for this major */
            if(strcmp(minor_pat, "0") == 0) {
                versions += 1;
                project_data_array->append(create_pseudo_empty_version(thismatch->prcs_major()));
            }

	    for(int j = i; j < versions; j += 1) {

                if(project_data_array->index(j) == NULL)
                    continue;

                if(matches(thismatch->prcs_major(), project_data_array->index(j)->prcs_major())) {
                    if(matches(minor_pat, project_data_array->index(j)->prcs_minor())) {

			if(project_data_array->index(j)->deleted())
			    continue;

			if(first) {
                            first = false;
                        } else if (option_long_format || option_really_long_format) {
                            prcsoutput << prcsendl;
                        }

			if(strcmp(project_data_array->index(j)->prcs_minor(), "@") == 0)
			    project_data_array->index(j, rep_entry->highest_minor_version_data(thismatch->prcs_major()));

                        Return_if_fail(print_info(project_data_array->index(j), rep_entry));

                    } else if(i == j) {
                        /* append the "@" minor version */
                        versions += 1;
                        project_data_array->append(create_pseudo_latest_version(thismatch));
                    }

                    project_data_array->index(j, (ProjectVersionData*)0);
                }
            }
        }
    }

    if(first) {
        prcswarning << "No versions match "
		    << (cmd_version_specifier_major[0] == '\0' ? "(null)" : major_pat)
		    << '.'
		    << (cmd_version_specifier_minor[0] == '\0' ? "(null)" : minor_pat)
		    << dotendl;
    }

    delete project_data_array;

    return ExitSuccess;
}


static bool matches(const char* pat, const char* text)
{
    return fnmatch(pat, text, FNM_NOESCAPE) == 0;
}


static PrVoidError print_info(ProjectVersionData* project_data, RepEntry *rep_entry)
{
    ProjectDescriptor *P = NULL;

    if(option_long_format || option_really_long_format) {
        if(!project_data->rcs_version()) {
            /* its an empty version */
            P = create_pseudo_empty_projdesc(project_data->prcs_major());
        } else {
	    If_fail(P << rep_entry->checkout_prj_file(cmd_root_project_full_name,
						      project_data->rcs_version()))
		return NoError;
        }
    }

    prcsoutput << cmd_root_project_name << ' '
	       << project_data->prcs_major() << '.'
	       << project_data->prcs_minor() << ' '
	       << time_t_to_rfc822(project_data->date()) << " by "
	       << project_data->author() << endl;

    if(option_long_format) {

        static Dstring *desc = NULL;
        Dstring thisdesc, thisvlog;
        const char* descstripped, *vlogstripped;

	if(!desc)
	    desc = new Dstring;

        thisdesc.assign(*P->project_description());
        thisvlog.assign(*P->version_log());

        if(thisdesc.index(0) == '\"') {
            /* kill the quotes */
            thisdesc.truncate(thisdesc.length() - 1);
            descstripped = (const char*)thisdesc + 1;
        } else {
            descstripped = thisdesc;
        }

        if(thisvlog.index(0) == '\"') {
            /* kill the quotes */
            thisvlog.truncate(thisvlog.length() - 1);
            vlogstripped = (const char*)thisvlog + 1;
        } else {
            vlogstripped = thisvlog;
        }

	prcsoutput << "Parent-Version:" << setcol(21)
		<< P->parent_version_major() << '.'
		<< P->parent_version_minor() << prcsendl;

	if(P->descends_from_major() && strcmp(*P->descends_from_major(), "-*-") != 0) {
	    prcsoutput << "Descends-From:" << setcol(21)
		    << P->descends_from_major() << '.'
		    << P->descends_from_minor() << prcsendl;
	}

        if(strchr(vlogstripped, '\n') != NULL || strlen(vlogstripped) > 60) {
            prcsoutput << "Version-Log:" << prcsendl
		    << vlogstripped << prcsendl;
        } else {
            prcsoutput << "Version-Log:" << setcol(21)
		    << (vlogstripped[0] == '\0' ? "empty" : vlogstripped)
		    << prcsendl;
        }

        if(desc->length() == 0 || strcmp(*desc, descstripped) != 0) {
            if(strchr(descstripped, '\n') != NULL || strlen(descstripped) > 60) {
                prcsoutput << "Project-Description:" << prcsendl
			<<  descstripped << prcsendl;
            } else {
                prcsoutput << "Project-Description:" << setcol(21)
			<< (descstripped[0] == '\0' ? "empty" : descstripped) << prcsendl;
            }
        } else {
            prcsoutput << "Project-Description:" << setcol(21) << "as above" << prcsendl;
        }


        desc->assign(descstripped);
    }

    if(option_really_long_format) {

        if(P->file_entries()->length() == 0) {
            prcsoutput << "No working files" << dotendl;
        } else {
            FileType type;
            int width = 0;

	    eliminate_unnamed_files(P);

	    foreach_fileent(fe_ptr, P) {
		FileEntry* fe = *fe_ptr;

		if (!fe->on_command_line())
		    continue;

                int thiswidth = fe->working_path()->length();

                if (thiswidth > width)
                    width = thiswidth;
            }

            if(width < 21)
                width = 21;

	    foreach_fileent(fe_ptr2, P) {
		FileEntry* fe = *fe_ptr2;

		if (!fe->on_command_line())
		    continue;

                type = fe->file_type();

                prcsoutput << fe->working_path()
			   << setcol(width) << ' '
			   << (fe->file_tags() ? fe->file_tags()->cast() : ":real-file");

		if (fe->file_type() == RealFile) {
		    RcsVersionData* version_data;

		    Return_if_fail(version_data << rep_entry->
				   lookup_rcs_file_data(*fe->descriptor_name(),
							*fe->descriptor_version_number()));

		    prcsoutput << " MD5 = ";

		    for(int i = 0; i < 16; i += 1) {
			int c = version_data->unkeyed_checksum()[i] & 0xff;

			prcsoutput.form("%02x", c);
		    }
		}

		prcsoutput << prcsendl;
            }
        }
    }

    if (P) delete P;

    return NoError;
}
