/* -*-Mode: C++;-*-
 * $Id: misc.cc 1.30.1.3.1.17.1.11.1.7.1.12 Tue, 03 Sep 1996 16:01:28 -0700 jmacd $
 * misc.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


extern "C" {
#include <sys/utsname.h>
#include <fcntl.h>
#include <time.h>
#include <pwd.h>
#include <stdarg.h>
#include <stdio.h>
#include <sys/wait.h>
#include <signal.h>
#include "maketime.h"
}

#include "prcs.h"
#include "prcsdir.h"
#include "misc.h"
#include "vc.h"
#include "syscmd.h"
#include "populate.h"
#include "repository.h"
#include "system.h"
#include "checkout.h"

const char  *cmd_root_project_name = NULL;
const char  *cmd_root_project_full_name;
const char  *cmd_root_project_file_path;
const char  *cmd_root_project_path;
int          cmd_root_project_path_len;

bool         cmd_prj_given_as_file = false;
char       **cmd_filenames_given;
const char **cmd_corrected_filenames_given;
int          cmd_filenames_count;
bool        *cmd_filenames_found;
char       **cmd_diff_options_given;
int          cmd_diff_options_count;
const char  *cmd_version_specifier_major = NULL;
const char  *cmd_version_specifier_minor = NULL;
int          cmd_version_specifier_minor_int = -1;
const char  *cmd_alt_version_specifier_major = NULL;
const char  *cmd_alt_version_specifier_minor = NULL;
const char  *cmd_repository_path = NULL;

int option_immediate_uncompression = 0;
int option_force_resolution = 0;
int option_long_format = 0;
int option_really_long_format = 0;
int option_report_actions = 0;
int option_version_present = 0;
int option_diff_keywords = 0;
int option_diff_new_file = 0;
int option_populate_delete = 0;
int option_package_compress = 0;
int option_be_silent = 0;
int option_preserve_permissions = 0;
int option_file_family = 0;
int option_unlink = 0;
int option_match_file = 0;
int option_not_match_file = 0;
int option_all_files = 0;
int option_preorder = 0;
int option_pipe = 0;
int option_record_tags = 0;
#ifdef PRCS_DEVEL
int option_debug = 1;
#endif

const char *option_match_file_pattern;
const char *option_not_match_file_pattern;

const int prcs_version_number[3] = { PRCS_MAJOR_VERSION_NUMBER,
				     PRCS_MINOR_VERSION_NUMBER,
				     PRCS_MICRO_VERSION_NUMBER };

const char maintainer[] = "prcs-bugs@XCF.Berkeley.EDU";

const char* temp_file_1 = NULL;
const char* temp_file_2 = NULL;
const char* temp_file_3 = NULL;
const char* temp_directory = NULL;

RepEntry* Rep = NULL;

bool is_linked(const char* path)
{
    struct stat buf;

    if (lstat(path, &buf) < 0) {
        return false;
    } else if (S_ISLNK(buf.st_mode) || buf.st_nlink > 1) {
        return true;
    } else {
        return false;
    }
}

/* this unbackslashes all backslashed characters and removes excess
 * slashes from path */
Dstring* correct_path(const char *path)
{
    ASSERT(path != NULL, "can't be null");
    Dstring *in = new Dstring;

    while(*path != '\0') {
        if(*path == '\\') {
            path += 1;
        }

        in->append(*path);
        path += 1;
    }

    return in;
}

/* backslashes all backslashes and quotation marks in a path so that
 * they can be inserted into the project file */
Dstring* protect_path(const char* path)
{
    Dstring *new_prot = new Dstring;

    protect_path(path, new_prot);

    return new_prot;
}

void protect_path(const char* path, Dstring* in)
{
    in->truncate(0);

    while(*path != '\0') {
        if(*path == '\"' || *path == '\\' || *path == ')' || *path == '(' || isspace(*path))
            in->append('\\');
        in->append(*path);
        path += 1;
    }
}


bool weird_pathname(const char* N)
{
    /* don't allow anything outside the repository */
    if (strncmp("../", N, 3) == 0)
        return true;
    else if (strncmp("./", N, 2) == 0)
	return true;

    if (N[0] == '/')
	return true;

    /* don't allow any '/../'s in the path */
    while ((N = strchr(N, '/')) != NULL) {
        if (strncmp(N, "/../", 4) == 0)
            return true;
	if (strncmp(N, "/./", 3) == 0)
	    return true;
        N += 1;
    }

    if (N) {
	int len = strlen(N);

	if (len > 0 && N[len - 1] == '/')
	    return true;
    }

    return false;
}

const char* strip_leading_path(const char* P)
{
    char* s = strrchr(P, '/');

    if ( s == '\0' )
        return P;
    else
        return s+1;
}

bool fs_is_directory(const char* P)
{
    struct stat buf;

    if(stat(P, &buf) < 0)
        return false;
    if(!S_ISDIR(buf.st_mode))
        return false;
    return true;
}

NprVoidError fs_nuke_file(const char* file)
{
    bool ret = true;
    if(fs_is_directory(file)) {
	const char* old_dir;

	If_fail(old_dir << name_in_cwd(""))
	    ret = false;

        If_fail(change_cwd(file)) {
            ret = false;
        }

	if (ret) {
	    Dir current_dir(".");

	    foreach(ent_ptr, current_dir, Dir::DirIterator) {
                If_fail(fs_nuke_file(*ent_ptr)) {
		    ret = false;
		}
	    }

            Return_if_fail(change_cwd(old_dir));

            ret &= current_dir.OK();
            ret &= rmdir(file) >= 0;
        }

    } else {
        ret &= unlink(file) >= 0;
    }

    if(ret)
	return NoError;
    else
	return FatalError;
}

mode_t get_umask()
{
    mode_t mask = umask(0);

    umask(mask);

    return mask;
}

const char* get_login()
{
    static char buf[32]; /* L_cuserid */
    static bool do_once = false;

    if(do_once) return buf;

    struct passwd *user = getpwuid(getuid());

    if(!user)
	strcpy(buf, "unknown");
    else
	strcpy(buf, user->pw_name);

    do_once = true;

    return buf;
}

/*
 * get_utc_time --
 *
 *     returns a time formatted like so: Sun, 24 Mar 1996 23:20:44 -0800
 *     which is UTC in that the offset from gmt is included, but the hour
 *     minute and date are in local time.  careful--the result is over
 *     written by the next call.
 */
const char* get_utc_time()
{
    static char buf[64];
    static bool do_once = false;
    if(do_once)
	return buf;
    strcpy(buf, time_t_to_rfc822(get_utc_time_t()));
    do_once = true;
    return buf;
}

time_t get_utc_time_t()
{
    static time_t t = 0;

    if(t)
	return t;

    time(&t);

    return t;
}

/*
 * timestr_to_time_t --
 *
 *     this calls the somewhat excessive RCS time conversion routine that
 *     converts a standard RCS time string such as 1996/01/23 07:36:14
 *     into a time_t.  I could write a less bulky conversion routine that
 *     didn't try to convert 10 different formats.
 */
time_t timestr_to_time_t(const char* rcstime)
{
    return str2time(rcstime, 0, 0);
}

/*
 * hours_away_from_gmt, minutes_away_from_gmt --
 *
 *     memoized routine that calculates in a hopefully POSIXLY CORRECT
 *     manner the offsets from UTC.  Hours will be +-12.  Minutes will
 *     be 0 < mins < 60.
 */
static bool away_calculated;
static int minutes_away, hours_away;
int hours_away_from_gmt()
{
    if(away_calculated)
        return hours_away;
    struct tm loc, gmt;
    time_t t = time(NULL);
    loc = *localtime(&t);
    gmt = *gmtime(&t);

    hours_away = loc.tm_hour - gmt.tm_hour;
    if(hours_away > 12)
        hours_away -= 24;
    minutes_away = loc.tm_min - gmt.tm_min;
    if(minutes_away < 0) {
        hours_away -= 1;
        minutes_away += 60;
    }
    return hours_away;
}

int minutes_away_from_gmt()
{
    if(away_calculated)
        return minutes_away;
    hours_away_from_gmt();
    return minutes_away;
}

/*
 * From RFC#822, Aug 13, 1982
 *
 *    5.1.  SYNTAX
 *
 *    date-time   =  [ day "," ] date time        ; dd mm yy
 *                                                ;  hh:mm:ss zzz
 *    day         =  "Mon"  / "Tue" /  "Wed"  / "Thu"
 *                /  "Fri"  / "Sat" /  "Sun"
 *    date         = 1*2DIGIT month 2*4DIGIT      ; correction in RFC#1123
 *                                                ; day month year
 *                                                ;  e.g. 20 Jun 82
 *    month       =  "Jan"  /  "Feb" /  "Mar"  /  "Apr"
 *                /  "May"  /  "Jun" /  "Jul"  /  "Aug"
 *                /  "Sep"  /  "Oct" /  "Nov"  /  "Dec"
 *    time        =  hour zone                    ; ANSI and Military
 *    hour        =  2DIGIT ":" 2DIGIT [":" 2DIGIT]
 *                                                ; 00:00:00 - 23:59:59
 *    zone        =  "UT"  / "GMT"                ; Universal Time
 *                                                ; North American : UT
 *                /  "EST" / "EDT"                ;  Eastern:  - 5/ - 4
 *                /  "CST" / "CDT"                ;  Central:  - 6/ - 5
 *                /  "MST" / "MDT"                ;  Mountain: - 7/ - 6
 *                /  "PST" / "PDT"                ;  Pacific:  - 8/ - 7
 *                /  1ALPHA                       ; Military: Z = UT;
 *                                                ;  A:-1; (J not used)
 *                                                ;  M:-12; N:+1; Y:+12
 *                / ( ("+" / "-") 4DIGIT )        ; Local differential
 *                                                ;  hours+min. (HHMM)
 */
const char* time_t_to_rfc822(time_t t)
{
    static char timebuf[64], tz[10];
    struct tm *tmbuf;

    tmbuf = localtime(&t);
    strftime(timebuf, 64, "%a, %d %h %Y %T ", tmbuf);
    sprintf(tz, "%+03d%02d", hours_away_from_gmt(), minutes_away_from_gmt());
    strcat(timebuf, tz);
    return timebuf;
}

/*
 * get_host_name --
 *
 *     This used to be here because I couldn't find a portable method
 *     of finding a hostname, but now I did.  This is used by the
 *     locking mechanism to alert users what machine is holding a lock.
 */
const char* get_host_name()
{
    static char buf[256];
    struct utsname utsbuf;
    if(uname(&utsbuf) < 0)
        strcpy(buf, "** uname() failed -- unknown host **");
    else
        strcpy(buf, utsbuf.nodename);
    return buf;
}

/*
 * name_in_cwd --
 *
 *     I don't like SunOS's getcwd(), it freaked me out when I was
 *     looking at a profiler output.  It calls popen on 'pwd' and fgets
 *     on the output!
 *
 *     returns NULL on failure.
 */
static bool changed = true;
PrConstCharPtrError name_in_cwd(const char* name)
{
    static char buf[MAXPATHLEN];
    static int buflen;

    if(changed) {
        if(
#if defined(sun) && !defined(__SVR4)
           getwd(buf)
#else
           getcwd(buf, MAXPATHLEN)
#endif
           == NULL) {
            throw prcserror << "Getcwd failed" << perror;
        } else
            changed = false;
        buflen = strlen(buf);
    }

    if(buflen + strlen(name) + 1 > MAXPATHLEN)
        throw prcserror << "Failed building full pathname to " << squote(name)
			 << ", name too long" << dotendl;

    buf[buflen] = '/';
    strcpy(buf + buflen + 1, name);

    return buf;
}

/*
 * change_cwd --
 *
 *     Change the current working directory and tell name_in_cwd its different
 *     now.
 */
NprVoidError change_cwd(const char* path)
{
    changed = true;
    if(chdir(path) == 0)
	return NoError;
    else
	return NonFatalError;
}

/*
 * guess_prj_name --
 *
 *     returns NULL if 0 or more than 1 .prj file are in the current directory.
 *     otherwise returns a heap-allocated Dstring containing the name of the
 *     file, after truncating the ".prj"
 */
PrDstringPtrError guess_prj_name(const char* dir)
{
    bool found = false;
    Dstring buf;
    Dstring *ret;

    Dir current_dir(dir);

    foreach(ent_ptr, current_dir, Dir::DirIterator) {
	const char* ent = *ent_ptr;

        int len = strlen(ent);
        if (len > 4 && strcmp(".prj", ent + len - 4) == 0) {
            if(found) {
                return (Dstring*)0;
            } else {
                buf.assign(ent);
                found = true;
            }
        }
    }

    if (! current_dir.OK() ) {
        throw prcserror << "Error reading current directory" << perror;
    }

    if(found) {
        ret = new Dstring(buf);
        ret->truncate(ret->length() - 4);

        return ret;
    } else {
        return (Dstring*)0;
    }
}

PrConstCharPtrError read_sym_link(const char* name)
{
    static char buf[MAXPATHLEN];
    int ret;

    if((ret = readlink(name, buf, MAXPATHLEN)) < 0) {
        throw prcserror << "Error reading symlink " << squote(name) << perror;
    } else {
        buf[ret] = '\0'; /* readlink() doesn't zero terminate */
    }

    return buf;
}

PrConstCharPtrError show_file_info(const char* file)
{
    static Dstring* lsout;
    ArgList *args;
    FILE* out;

    if(lsout == NULL)
        lsout = new Dstring();

    Return_if_fail(args << ls_command.new_arg_list());

    args->append("-ldgF");
    args->append(file);

    Return_if_fail(ls_command.open(true, false));

    out = ls_command.standard_out();

    lsout->truncate(0);

    If_fail(read_string(out, lsout))
	throw prcserror << "Read failure from ls output" << perror;

    Return_if_fail_if_ne(ls_command.close(), 0)
        throw prcserror << ls_command.path() << " exited abnormally" << dotendl;

    return lsout->cast();
}

PrConstCharPtrError absolute_path(const char* name)
{
    static Dstring *buf = NULL;

    if(!buf)
	buf = new Dstring;

    if(strcmp(name, "-") == 0) {
        return name;
    } else if(name[0] == '/') {
        return name;
    } else {
	const char* cwd;

	Return_if_fail(cwd << name_in_cwd(""));

        buf->assign(cwd);
        buf->append(name);

        return buf->cast();
    }
}

bool pathname_equal(const char*& a, const char*& b)
{
    const char *c(a), *d(b);
    for(;;) {

        while(c != a && *c == '/' && *(c - 1) == '/') { c += 1; }
        while(d != b && *d == '/' && *(d - 1) == '/') { d += 1; }

        if(*c != *d)
            return false;

        if(*c == '\0') {
            return true;
        }

        c += 1;
        d += 1;
    }
}

int pathname_hash(const char*& s, int M)
{
    const char *p;
    unsigned int h(0), g;
    for(p = s; *p != '\0'; p += 1) {
        if(p != s && *p == '/' && *(p-1) == '/')
            continue;
        h = ( h << 4 ) + *p;
        if ( ( g = h & 0xf0000000 ) ) {
            h = h ^ (g >> 24);
            h = h ^ g;
        }
    }

    return h % M;
}

char get_user_char()
    /* 0 is error */
{
    int ans, c;

    If_fail(ans << Err_fgetc(stdin))
	return 0;

    if(ans == EOF)
        return 0;
    else if(ans == '\n')
        return '\n';

    while(true) {
	If_fail(c << Err_fgetc(stdin))
	    return 0;

	if (c == EOF)
	    return 0;
	else if (c == '\n')
	    break;
    }

    return (char)ans;
}


const char* make_temp_file(const char* extension)
{
    static int counter = 0;
    ASSERT(temp_directory, "set the temp dir");
    Dstring *t = new Dstring(temp_directory);
    t->sprintfa("/prcs_temp.%d%s", counter, extension);
    counter += 1;
    return *t;
}

const char* format_type(FileType t)
{
    if (t == SymLink)
        return "symlink";
    else if (t == Directory)
        return "directory";
    else if (t == RealFile)
        return "regular file";
    else
        return "other";
}

PrVoidError fs_write_filename(FILE *in, const char* filename)
{
    FILE* out;

    If_fail(out << Err_fopen(filename, "w"))
	throw prcserror << "Failed opening file " << squote(filename)
		         << " for writing" << perror;

    Return_if_fail(write_file(in, out));

    If_fail(Err_fclose(out))
	throw prcserror << "Failed writing to file "
			 << squote(filename) << perror;

    return NoError;
}

PrVoidError fs_move_filename(const char* infile, const char* outfile)
{
    if (rename(infile, outfile) >= 0)
	return NoError;

    if (errno != EXDEV)
	throw prcserror << "Rename failed on " << squote(infile) << " to "
			<< squote(outfile) << perror;

    return fs_copy_filename(infile, outfile);
}

PrVoidError fs_copy_filename(const char* infile, const char* outfile)
{
    FILE *in, *out;

    If_fail(in << Err_fopen(infile, "r"))
	throw prcserror << "Failed opening file " << squote(infile)
	                << " for reading" << perror;

    If_fail(out << Err_fopen(outfile, "w"))
	throw prcserror << "Failed opening file " << squote(outfile)
		        << " for writing" << perror;

    Return_if_fail(write_file(in, out));

    fclose(in);

    If_fail(Err_fclose(out))
	throw prcserror << "Failed writing file " << squote(outfile)
			 << perror;

    return NoError;
}

static char *Buffer1 = NULL, *Buffer2 = NULL;
static int Buffer1Len = 0, Buffer2Len = 0;
static int Buffer1Alloc = 0, Buffer2Alloc = 0;

static void opt_buffer(int toread, int* buflen, int* bufalloc, char** buf)
{
    struct stat statbuf;

    *buflen = 1 << 10;  /* default */

    if(toread >= 0 && fstat(toread, &statbuf) >= 0)
        *buflen = statbuf.st_blksize;

    if(*bufalloc < *buflen) {
        if(*bufalloc == 0) {
            *buf = NEWVEC(char, *buflen);
            *bufalloc = *buflen;
        } else {
            *buf = (char*)EXPANDVEC(*buf, *bufalloc, *buflen - *bufalloc);
        }
    }
}

#define opt_buffer_1(fd) opt_buffer(fd, &Buffer1Len, &Buffer1Alloc, &Buffer1)
#define opt_buffer_2(fd) opt_buffer(fd, &Buffer2Len, &Buffer2Alloc, &Buffer2)

PrVoidError write_file(FILE* in, FILE* out)
{
    int c;

    opt_buffer_1(-1);

    while((c = fread(Buffer1, 1, Buffer1Len, in)) > 0) {
        fwrite(Buffer1, 1, c, out);
    }

    if(ferror(in))
        throw prcserror << "Read failure" << perror;

    if(ferror(out))
        throw prcserror << "Write failure" << perror;

    return NoError;
}

bool empty_file(FILE* empty)
{
    opt_buffer_1(-1);

    while(fread(Buffer1, 1, Buffer1Len, empty) > 0)
        /* do nothing */;
    return true;
}

PrVoidError fs_zero_file(const char* name)
{
    /* use open so that if FILE doesn't exist create it. */
    If_fail(Err_open(name, O_TRUNC|O_CREAT, 0644))
        throw prcserror << "Failed truncating file "<< squote(name) << perror;

    return NoError;
}

NprVoidError write_fds(int infd, int outfd)
{
    int c;

    opt_buffer_1(infd);

    do {
	If_fail(c << Err_read(infd, Buffer1, Buffer1Len))
	    return ReadFailure;

	If_fail(Err_write(outfd, Buffer1, c))
	    return WriteFailure;

    } while (c);

    return NoError;
}

bool fs_same_file_owner(const char* file)
{
    struct stat buf;

    if(stat(file, &buf) < 0)
        return false;

    return getuid() == buf.st_uid;
}

NprBoolError cmp_fds(int one, int two)
{
    int nRead1 = 1, nRead2 = 1, len;

    opt_buffer_1(one);
    opt_buffer_2(two);

    len = Buffer1Len < Buffer2Len ? Buffer1Len : Buffer2Len;

    while(nRead1 > 0 && nRead1 == nRead2) {
	If_fail(nRead1 << Err_read(one, Buffer1, len))
	    return ReadFailure;
	If_fail(nRead2 << Err_read(two, Buffer2, len))
	    return ReadFailure;
        if(memcmp(Buffer1, Buffer2, nRead1) != 0)
            return 1;
    }

    return nRead1 != nRead2;
}

NprVoidError read_string(FILE* f, Dstring* s)
{
    while(true) {
	int c;

	Return_if_fail(c << Err_fgetc(f));

	if (c == EOF)
	  return ReadFailure;

	if (c == '\n')
	    break;

	s->append((char)c);
    }

    return NoError;
}

NprCFilePtrError Err_fopen(const char* file, const char* type)
{
    FILE* f = fopen(file, type);

    if(f)
	return f;
    else
	return NonFatalError;
}

NprVoidError Err_fclose(FILE* f)
{
    if(fclose(f) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_unlink(const char* a)
{
    if(unlink(a) < 0 && errno != ENOENT)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_rename(const char* a, const char* b)
{
    if(rename(a, b) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprIntError Err_read(int fd, void* buf, size_t nbytes)
{
    int c;

    do {
	c = read(fd, buf, nbytes);
    } while (c < 0 && errno == EINTR);

    if (c < 0)
	return ReadFailure;
    else
	return c;
}

NprVoidError Err_read_expect(int fd, void* buf, size_t nbytes)
{
    int c;

    while(nbytes > 0) {
	Return_if_fail(c << Err_read(fd, buf, nbytes));

	nbytes -= c;
    }

    return NoError;
}

NprVoidError Err_write(int fd, const void* buf, size_t nbytes)
{
    int c = 0;

    do {
	nbytes -= c;

	if (nbytes <= 0)
	    break;

	c = write(fd, buf, nbytes);
     } while (c < 0 && errno == EINTR);

    if (c < 0)
	return WriteFailure;
    else
	return NoError;
}

NprIntError Err_open(const char* a, int flags, mode_t mode)
{
    int fd;
    fd = open(a, flags, mode);
    if(fd < 0)
	return NonFatalError;
    else
	return fd;
}

NprVoidError Err_close(int fd)
{
    if(close(fd) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_mkdir(const char* name, mode_t mode)
{
    if(mkdir(name, mode) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_chmod(const char* name, mode_t mode)
{
    if(chmod(name, mode) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_fstat(int fd, struct stat* buf)
{
    if(fstat(fd, buf) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_stat(const char* file, struct stat* buf)
{
    if(stat(file, buf) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprVoidError Err_symlink(const char* a, const char* b)
{
    if(symlink(a, b) < 0)
	return NonFatalError;
    else
	return NoError;
}

NprIntError Err_waitpid(int pid)
{
    int status;
    int ret;

    while (true) {
        while ((ret = waitpid(pid, &status, 0)) < 0 && errno == EINTR) { }

        if (WIFSTOPPED(status))
            kill(SIGCONT, pid);
        else
            break;
    }

    if (ret < 0)
	return FatalError;
    else
	return status;
}

NprIntError Err_waitpid_nostart(int pid)
{
    int status;
    int ret;

    while (true) {
        while ((ret = waitpid(pid, &status, 0)) < 0 && errno == EINTR) { }

        if (WIFSTOPPED(status))
            sleep(1);
        else
            break;
    }

    if (ret < 0)
	return FatalError;
    else
	return status;
}

NprIntError Err_fgetc(FILE* f)
{
    int c;

    while ((c = fgetc(f)) == EOF && errno == EINTR) { }

    if (c == EOF && ferror(f))
	return ReadFailure;
    else
	return c;
}

const char* intern(const char* name)
{
    char *s = NEWVEC(char, strlen(name)+1);

    strcpy(s, name);

    return s;
}

PrConstCharPtrError find_real_filename(const char* filename)
{
    static char buf[MAXPATHLEN];
    struct stat sbuf;

    if (lstat(filename, &sbuf) < 0)
	throw prcserror << "Lstat failed on "
			<< squote(filename)
			<< perror;

    if (S_ISLNK(sbuf.st_mode)) {
	const char* sym;

	Return_if_fail(sym << read_sym_link(filename));

	if (sym[0] == '/') {
	    return find_real_filename(sym);
	} else {
	    Dstring directory;

	    dirname(filename, &directory);

	    directory.append(sym);

	    return find_real_filename(sym);
	}

    } else {
	strncpy(buf, filename, MAXPATHLEN-1);
	buf[MAXPATHLEN-1] = 0;

	return buf;
    }
}

void dirname(const char* p, Dstring* s)
{
    const char* last_slash = strrchr(p, '/');

    if (!last_slash)
	s->assign("./");
    else
	s->append(p, last_slash - p);
}

PrVoidError WriteableFile::copy(FILE* copy_me)
{
    opt_buffer_1(-1);

    int c;

    while((c = fread(Buffer1, 1, Buffer1Len, copy_me)) > 0)
	os.write(Buffer1, c);

    if (os.bad())
	throw prcserror << "Error writing " << squote(temp_name) << perror;

    return NoError;
}

PrVoidError WriteableFile::copy(const char* copy_me)
{
    FILE* it;

    If_fail(it << Err_fopen(copy_me, "r"))
	throw prcserror << "Open failed on " << squote(copy_me) << perror;

    PrVoidError ret = copy(it);

    fclose(it);

    return ret;
}
