/* -*-Mode: C++;-*-
 * $Id: package.cc 1.15.1.2.1.8.1.8.1.5 Tue, 03 Sep 1996 16:48:15 -0700 jmacd $
 * package.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


extern "C" {
#include <sys/wait.h>
#include <signal.h>
}

#include "prcs.h"
#include "syscmd.h"
#include "misc.h"
#include "repository.h"
#include "projdesc.h"
#include "system.h"

const char gzip_magic_1 = '\037';
const char gzip_magic_2 = '\213';
const char magic_number_1 = '\200';
const char magic_number_2 = '\352';
const int header_length = 256;
const mode_t package_file_mode = 0666;

static PrVoidError package_project(const char* project_name, const char* package_file);

PrPrcsExitStatusError package_command()
{
    Return_if_fail(package_project(cmd_root_project_name, cmd_filenames_given[0]));

    return ExitSuccess;
}

static PrVoidError package_project(const char* project_name, const char* package_file)
{
    ArgList *args;
    int outfd = STDOUT_FILENO;
    char buf[header_length];
    int pipefd[2], pid = -1, status;
    const char* rep_path;

    Return_if_fail(args << tar_command.new_arg_list());

    If_fail(rep_path << Rep_guess_repository_path())
	throw prcserror << "Please set your PRCS_REPOSITORY environment variable" << dotendl;

    if(!fs_is_directory(rep_path))
	throw prcserror << "Repository " << squote(rep_path) << " does not exist" << dotendl;

    Dstring rep_entry_path(rep_path);
    rep_entry_path.append('/');
    rep_entry_path.append(project_name);

    if(!fs_is_directory(rep_entry_path))
	throw prcserror << "Repository entry " << squote(rep_entry_path) << " does not exist" << dotendl;

    args->append("cf");
    args->append("-");
    args->append("-C");
    args->append(rep_path);
    args->append(project_name);

    Return_if_fail(tar_command.open(true, false));

    if(strcmp(package_file, "-") != 0) {
	If_fail(outfd << Err_open(package_file, O_WRONLY|O_CREAT|O_TRUNC, package_file_mode))
	    throw prcserror << "Can't open " << squote(package_file)
		            << " for writing" << perror;
    }

    memset(buf, 0, header_length);
    buf[0] = magic_number_1;
    buf[1] = magic_number_2;
    sprintf(buf + 2, "%s", project_name);

    if(option_package_compress) {
	Return_if_fail(gzip_command.init());

	if(pipe(pipefd) != 0) {
	    throw prcserror << "Pipe failed, can't compress data" << perror;
	} else if((pid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if (pid == 0) {
	    dup2(pipefd[PIPE_READ_FD], STDIN_FILENO);
	    dup2(outfd, STDOUT_FILENO);
	    close(pipefd[PIPE_READ_FD]);
	    close(pipefd[PIPE_WRITE_FD]);

	    execl(gzip_command.path(), "-", "-f", NULL);
	    abort_child(gzip_command.path());
	} else {
	    dup2(pipefd[PIPE_WRITE_FD], outfd);
	    close(pipefd[PIPE_READ_FD]);
	    close(pipefd[PIPE_WRITE_FD]);
	}
    }

    If_fail(Err_write(outfd, buf, header_length))
	throw prcserror << "Write failed writing package file" << perror;

    If_fail(write_fds(fileno(tar_command.standard_out()), outfd))
	/* @@@ could use more exception info */
	throw prcserror << "Write failed writing package file" << perror;

    If_fail(Err_close(outfd))
	throw prcserror << "Write failed writing package file" << perror;

    Return_if_fail_if_ne(tar_command.close(), 0) {
	unlink(package_file);
	throw prcserror << "System command tar failed" << dotendl;
    }

    if(option_package_compress) {
	If_fail(status << Err_waitpid(pid))
	    throw prcserror << "Waitpid failed on pid " << pid << perror;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "System command gzip failed" << dotendl;
    }

    return NoError;
}

static PrVoidError invalid()
{
    throw prcserror << squote(cmd_root_project_name)
		    << " is not a valid package file" << dotendl;
}

static PrVoidError feed_pipe(const char* header_buf, int from_fd, int to_fd)
{
    If_fail(Err_write(to_fd, header_buf, header_length))
	throw prcserror << "Write failed writing to gzip pipe" << perror;

    If_fail(write_fds(from_fd, to_fd))
	throw prcserror << "Write failed writing to gzip pipe" << perror;

    return NoError;
}

PrPrcsExitStatusError unpackage_command()
{
    ArgList *args;
    int fd, zipoutpipefd[2], zipinpipefd[2], zippid = 0, writepid = 0;
    Dstring projname;
    char buf[header_length];
    const char* rep_path;

    if(strcmp(cmd_root_project_name, "-") != 0) {
	If_fail(fd << Err_open(cmd_root_project_name, O_RDONLY))
	    throw prcserror << "Open failed on package file" << perror;
    } else {
	fd = STDIN_FILENO;
    }

    If_fail(rep_path << Rep_guess_repository_path())
	throw prcserror << "Please set your PRCS_REPOSITORY environment variable" << dotendl;

    if(!fs_is_directory(rep_path))
	throw prcserror << "Repository " << squote(rep_path) << " does not exist" << dotendl;

    If_fail(Err_read_expect(fd, buf, header_length)) {
	if (errno)
	    throw prcserror << "Read error on package file" << perror;
	else
	    return invalid();
    }

    if(buf[0] == magic_number_1 && buf[1] == magic_number_2) {
	option_package_compress = false;
    } else if(buf[0] == gzip_magic_1 && buf[1] == gzip_magic_2) {
	option_package_compress = true;
    } else {
	return invalid();
    }

    if(option_package_compress) {
	Return_if_fail(gzip_command.init());

	if(pipe(zipoutpipefd) != 0 || pipe(zipinpipefd) != 0) {
	    throw prcserror << "Pipe failed, can't compress data" << perror;
	} else if((zippid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if (zippid == 0) {
	    dup2(zipoutpipefd[PIPE_WRITE_FD], STDOUT_FILENO);
	    dup2(zipinpipefd[PIPE_READ_FD], STDIN_FILENO);

	    close(zipoutpipefd[PIPE_READ_FD]);
	    close(zipoutpipefd[PIPE_WRITE_FD]);
	    close(zipinpipefd[PIPE_READ_FD]);
	    close(zipinpipefd[PIPE_WRITE_FD]);

	    execl(gzip_command.path(), "-", "-d","-f", NULL);
	    abort_child(gzip_command.path());
	} else if((writepid = fork()) < 0) {
	    throw prcserror << "Fork failed, can't compress data" << perror;
	} else if(writepid == 0) {
	    int exit_val = 0;

            close(zipinpipefd[PIPE_READ_FD]);

	    If_fail(feed_pipe(buf, fd, zipinpipefd[PIPE_WRITE_FD]))
		exit_val = 1;

	    close(zipinpipefd[PIPE_WRITE_FD]);

	    _exit(exit_val);
	} else {

	    dup2(zipoutpipefd[PIPE_READ_FD], fd);
	    close(zipoutpipefd[PIPE_READ_FD]);
	    close(zipoutpipefd[PIPE_WRITE_FD]);
	    close(zipinpipefd[PIPE_READ_FD]);
	    close(zipinpipefd[PIPE_WRITE_FD]);

	    If_fail(Err_read_expect(fd, buf, header_length))
		return invalid();

	    if(buf[0] != magic_number_1 || buf[1] != magic_number_2)
		return invalid();
	}
    }

    projname.assign(buf + 2);

    Return_if_fail(args << tar_command.new_arg_list());

    prcsinfo << "Unpackage project " << squote(projname) << dotendl;

    Dstring entry_path(rep_path);
    entry_path.append('/');
    entry_path.append(projname);

    if(fs_is_directory(entry_path)) {
	prcsquery << "Project " << squote(projname) << " already exists.  "
		  << force("Replacing")
		  << report("Replace")
		  << optfail('n')
		  << defopt('y', "Delete project in repository")
		  << query("Replace");

	Return_if_fail(prcsquery.result());

	if(option_report_actions)
	    return ExitSuccess;

	If_fail(fs_nuke_file(entry_path)) {
	    throw prcserror << "Failed removing old repository entry "
			     << squote(entry_path) << dotendl;
	}
    }

    if(option_report_actions)
	return ExitSuccess;

    args->append("xpf");
    args->append("-");
    args->append("-C");
    args->append(rep_path);

    Return_if_fail_if_ne(tar_command.open_filein(fd, false), 0) {
	throw prcserror << "System command tar failed" << dotendl;
    }

    int status;

    if(option_package_compress) {
	If_fail(status << Err_waitpid(writepid))
	    throw prcserror << "Waitpid failed on pid " << writepid << perror;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "Pipe reader exited with non-zero status" << dotendl;

	If_fail(status << Err_waitpid(zippid))
	    throw prcserror << "Waitpid failed on pid " << zippid << perror;

	if(WIFSIGNALED(status) && WTERMSIG(status) == SIGPIPE)
	    throw prcserror << "Child gzip process was terminated with SIGPIPE" << dotendl;

	if(!WIFEXITED(status) || WEXITSTATUS(status) != 0)
	    throw prcserror << "Child gzip process returns non-zero status" << dotendl;
    }

    return ExitSuccess;
}

PrVoidError save_package()
{
    Dstring new_package;
    bool compress_it = true;

    new_package.append(cmd_root_project_name);
    new_package.append(".pkg");

    prcsquery << "Enter a package name"
	      << definput(new_package)
	      << string_query("");

    const char* ans;

    Return_if_fail(ans << prcsquery.string_result());

    prcsquery << "Packaging repository.  "
	      << force("Compressing")
	      << report("Compress")
	      << option('n', "Don't compress the package")
	      << defopt('y', "Compress the package")
	      << query("Compress");

    char c;

    Return_if_fail(c << prcsquery.result());

    compress_it = (c == 'y');

    option_package_compress = compress_it;

    Return_if_fail(package_project(cmd_root_project_name, ans));

    return NoError;
}
