/* -*-Mode: C++;-*-
 * $Id: populate.cc 1.11.1.6.1.12.1.3.1.7 Mon, 02 Sep 1996 16:59:57 -0700 jmacd $
 * populate.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include "prcs.h"
#include "populate.h"
#include "prcsdir.h"
#include "projdesc.h"
#include "sexp.h"
#include "hash.h"
#include "repository.h"
#include "misc.h"
#include "fileent.h"
#include "system.h"

static int added = 0;
static int deleted = 0;

static void append_new_files(ProjectDescriptor* P,
			     FileRecordList *record_list,
			     bool delete_files);
static int hash_dir(const char*, InoTable*, PathTable*);
static int hash_file_or_dir(const char*, InoTable*, PathTable*);
static PrVoidError check_project_file_populate(ProjectDescriptor *,
					       PathTable *,
					       InoTable *);
static PrVoidError init_ignore(ProjectDescriptor* project);
static bool not_ignored(const char* name);

PrPrcsExitStatusError populate_command_filename(const char* filename, bool verbose);

static FileRecordList *new_records = NULL;
static FileRecordList *deleted_records = NULL;

static bool use_regex;
static re_pattern_buffer *populate_regex;

PrPrcsExitStatusError populate_command()
{
    return populate_command_filename(cmd_root_project_file_path, true);
}

PrPrcsExitStatusError populate_command_filename(const char* filename, bool verbose)
{
    ProjectDescriptor *P;
    PathTable table(pathname_hash, pathname_equal),
	      descr(pathname_hash, pathname_equal);
    InoTable inodes;

    Return_if_fail(P << read_project_file(cmd_root_project_full_name,
					  cmd_root_project_file_path));

    Return_if_fail(check_project_file_populate(P, &table, &inodes));

    Return_if_fail(init_ignore(P));

    if(cmd_filenames_count < 1)
	hash_dir(cmd_root_project_path[0] ? cmd_root_project_path : ".", &inodes, &table);
    else {
	/* the functionality here is the same as eliminate_unnamed_files
	 * from checkin.cc, but its not really eliminating, rather its
	 * marking in this case. */
	for(int i = 0; i < cmd_filenames_count; i += 1) {
	    if (weird_pathname(cmd_corrected_filenames_given[i] + cmd_root_project_path_len))
		throw prcserror << "Illegal file name "
				<< squote(cmd_corrected_filenames_given[i])
				<< "names may not contain "
				<< squote("../") << " or " << squote("./")
				<< "or end or begin with " << squote("/")
				<< dotendl;
	    else
		hash_file_or_dir(cmd_corrected_filenames_given[i], &inodes, &table);
	}
    }

    if(added == 0 && deleted == 0) {
	prcsinfo << "No new files" << dotendl;
	return ExitSuccess;
    }

    if(verbose && option_long_format)
	prcsinfo << "New files are: " << prcsendl;

    if(added > 0) {
	P->append_file("\n; Files added by populate at ");
	P->append_file(get_utc_time());
	P->append_file(":\n");
	append_new_files(P, new_records, false);
	P->append_file("\n");
    }

    if(deleted > 0) {
	P->append_file("\n; Files deleted by populate at ");
	P->append_file(get_utc_time());
	P->append_file(":\n");
	append_new_files(P, deleted_records, true);
	P->append_file("\n");
    }

    if(verbose) {
	if(added == 1) {
	    prcsinfo << "One file was added" << dotendl;
	} else if(added == 0) {
	    prcsinfo << "No new files" << dotendl;
	} else {
	    prcsinfo << added << " files were added" << dotendl;
	}
	if(option_populate_delete) {
	    if(deleted == 1) {
		prcsinfo << "One file was deleted" << dotendl;
	    } else if(deleted == 0) {
		prcsinfo << "No files deleted" << dotendl;
	    } else {
		prcsinfo << deleted << " files were deleted" << dotendl;
	    }
	}
    }

    if(option_report_actions)
	return ExitSuccess;

    Return_if_fail(P->write_project_file(filename));

    return ExitSuccess;
}

static int hash_file_or_dir(const char* name, InoTable* T, PathTable* P)
{
    struct stat statbuf;
    Dstring *n;
    FileRecord r;

    n = new Dstring(name);
    r.name = n;

    if ( lstat(name, &statbuf) < 0 ) {
	delete n;
	return 0;
    } else if ( S_ISREG(statbuf.st_mode) ) {
	/* first look in the table for this inode */
	FileType *lu(T->lookup(statbuf.st_ino));
	if ( lu == NULL && not_ignored(*n) ) { /* if not found then its new */
	    r.type = RealFile; /* set its type */
	    T->insert(statbuf.st_ino, RealFile); /* insert it */
	    P->insert(*n, RealFile);
	    new_records = new FileRecordList(r, new_records);
	    added += 1;
	}
    } else if ( S_ISLNK(statbuf.st_mode) ) {
	FileType *lu(P->lookup(*n));
	if ( lu == NULL && not_ignored(*n) ) {
	    r.type = SymLink;
	    P->insert(*n, SymLink);
	    new_records = new FileRecordList(r, new_records);
	    added += 1;
	}
    } else if ( S_ISDIR(statbuf.st_mode) ) {
	Dstring tmp(name);
	tmp.append('/');
	int inthisdir = hash_dir(tmp, T, P);
	FileType *lu(T->lookup(statbuf.st_ino));
	if ( inthisdir == 0 &&  lu == NULL && not_ignored(*n)) {
	    r.type = Directory;
	    T->insert(statbuf.st_ino, Directory);
	    P->insert(*n, Directory);
	    new_records = new FileRecordList(r, new_records);
	    added += 1;
	}
    } else {
	prcswarning << "Ignoring special file: " << squote(name)
		    << ", continuing" << dotendl;

	delete n;

	return 0;
    }

    return 1;
}

static int hash_dir(const char* dir, InoTable* T, PathTable *P)
{
    Dstring d = dir;
    int len, total = 0;

    if(strcmp(d, ".") == 0)
	d.truncate(0);

    len = d.length();

    Dir current_dir(dir);

    foreach(ent_ptr, current_dir, Dir::DirIterator) {
	d.truncate(len);
	d.append(*ent_ptr);
	total += hash_file_or_dir(d, T, P);
    }

    if (!current_dir.OK())
	prcserror << "Error reading directory " << squote(dir) << perror;

    return total;
}

static bool heuristic_keyword_guess(const char* path)
{
    FILE* file;
    char buffer[1024];
    int nread = 0;

    If_fail(file << Err_fopen(path, "r"))
	return false;

    nread = fread(buffer, 1, 1024, file);

    fclose(file);

    if (nread < 0)
	return false;
    else
	return memchr(buffer, 0, nread) != NULL;
}

static void append_new_files(ProjectDescriptor* P,
			     FileRecordList *record_list,
			     bool delete_files)
{
    Dstring append;
    const char *name, *path, *keys;
    FileType ft;

    while (record_list != NULL) {
	ft = record_list->head().type;
	name = *record_list->head().name + cmd_root_project_path_len;
	path = *protect_path(name);

	if(option_long_format) {
	    if(delete_files) {
		prcsoutput << "Deleted " << squote(name) << ": "
			   << format_type(ft) << prcsendl;
	    } else {
		prcsoutput << "Added " << squote(name) << ": "
			   << format_type(ft) << prcsendl;
	    }
	}

	if(delete_files) {
	    append.sprintfa("\n  ; `%s'", path);
	    P->append_file(append);
	} else if (ft == SymLink) {
	    const char* ln;
	    If_fail(ln << read_sym_link(*record_list->head().name))
		continue;

	    append.sprintfa("\n  (%s (%s) :symlink)",
			    path, protect_path(ln)->cast());
	    P->append_file(append);
	} else if (ft == Directory) {
	    append.sprintfa("\n  (%s () :directory)", path);
	    P->append_file(append);
	} else if (ft == RealFile) {
	    if(heuristic_keyword_guess(path))
		keys = " :no-keywords";
	    else
		keys = "";

	    append.sprintfa("\n  (%s ()%s)", path, keys);
	    P->append_file(append);
	}

	append.truncate(0);
	record_list = record_list->tail();
    }
}

PrVoidError check_project_file_populate(ProjectDescriptor *P,
					PathTable *table,
					InoTable *inodes)
{
    ino_t prjino, ino;
    FileType type;
    const char* name;
    struct stat buf;

    if(stat(P->project_file_path(), &buf) < 0)
	throw prcserror << "Can't stat project file" << perror;

    prjino = buf.st_ino;
    inodes->insert(prjino, RealFile);

    if(stat(P->project_aux_path(), &buf) >= 0)
	inodes->insert(buf.st_ino, RealFile);

    table->insert(P->project_file_path(), RealFile);

    foreach_fileent(fe_ptr, P) {
	FileEntry *fe = *fe_ptr;
	type = fe->file_type();
	name = *fe->working_path();
	bool file_present;

	if(Failure(file_present << fe->check_working_file()) || !file_present) {
	    if(option_populate_delete) {
		char c;
		static BangFlag bang;

		prcsquery << "File " << squote(name) << " is unavailable.  "
			  << force("Deleting")
			  << report("Delete")
			  << allow_bang(bang)
			  << option('n', "Don't delete this file")
			  << defopt('y', "Delete from project file")
			  << query("Delete");

		Return_if_fail(c << prcsquery.result());

		if(c == 'y') {
		    deleted += 1;
		    deleted_records = new FileRecordList(
			FileRecord(fe->file_type(),
				   new Dstring(*fe->working_path()))
			, deleted_records);
		    P->delete_file(fe);
		}
	    }

	    continue;
	}

	ino = fe->stat_inode();

	if(table->isdefined(name))
	    throw prcserror << "Duplicate entry in Files attribute: "
			     << squote(name) << dotendl;

	if(type == RealFile && inodes->isdefined(ino) && ino == prjino)
	    throw prcserror << "The project file is implicitly under version "
		"control, you may not enter it in the Files attribute: "
			     << squote(name) << dotendl;

	table->insert(name, type);
	inodes->insert(ino, type);
    }

    return NoError;
}

static void correct_string(const char* in, Dstring* out)
{
    char c;
    while((c = *in++) != 0) {
	if(c == '\\' && *in)
	    out->append(*in++);
	else
	    out->append(c);
    }
}

static PrVoidError init_ignore(ProjectDescriptor* project)
{
    Sexp* ignored = project->populate_ignore();

    if(ignored && ignored->length() > 0) {
	Dstring pat_buf, pat_buf_corrected;
	Dstring ind_pat;

	use_regex = true;

	while(!ignored->is_empty()) {
	    ind_pat.assign(ignored->car()->key()->cast() + 1);
	    ind_pat.truncate(ind_pat.length() - 1);

	    pat_buf.append(ind_pat);

	    if(!ignored->cdr()->is_empty())
		pat_buf.sprintfa("|");

	    ignored = ignored->cdr();
	}

	correct_string(pat_buf, &pat_buf_corrected);

	populate_regex = (re_pattern_buffer*)_malloc_(sizeof(re_pattern_buffer));

	Return_if_fail(prcs_compile_regex(pat_buf_corrected, populate_regex));

    } else {
	use_regex = false;
    }

    return NoError;
}

static bool not_ignored(const char* name)
{
    if(use_regex) {
	bool matches = prcs_regex_matches(name, populate_regex);

	if(option_long_format && matches)
	    prcsoutput << "Ignoring file " << squote(name) << dotendl;

	return !matches;
    } else {
	return true;
    }
}

PrVoidError prcs_compile_regex(const char* pat, re_pattern_buffer *r)
{
    const char* regerr;

    r->translate = 0;
    r->fastmap = 0;

    if((regerr = re_compile_pattern(pat, strlen(pat), r)) != NULL)
	throw prcserror << "Regular expression compilation failed on "
			<< pat << ": " << regerr << dotendl;

    return NoError;
}

bool prcs_regex_matches(const char* name, re_pattern_buffer* r)
{
    int len = strlen(name);

    return re_search(r, name, len, 0, len, NULL) >= 0;
}
