/* -*-Mode: C++;-*-
 * $Id: prcserror.cc 1.8.1.11.1.3.1.6 Mon, 02 Sep 1996 20:38:35 -0700 jmacd $
 * prcserror.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */

extern "C" {
#include <unistd.h>
#include <termios.h>
#include <stdarg.h>
#include <ctype.h>
#include <signal.h>

#ifndef __svr4__
#include <sys/ioctl.h>
#endif
}

#include "prcs.h"
#include "misc.h"

static const int min_tty_width = 40;
static char** argv_save = NULL;
static int argc_save;

const char* re_query_message = NULL;
int re_query_len;

#ifndef PRCS_DEVEL
ErrorToken global_error_token;
int return_if_fail_if_ne_val;
#endif

stdiobuf stdout_stream(stdout);
stdiobuf stderr_stream(stderr);
strstreambuf query_stream;

static int always_print = 0;

static PrettyStreambuf pretty_stdout_stream(&stdout_stream, &always_print);
static PrettyStreambuf pretty_stderr_stream(&stderr_stream, &always_print);
static PrettyStreambuf pretty_severable_stream(&stderr_stream, &option_be_silent);
static PrettyStreambuf pretty_query_stream(&query_stream, &always_print);
#ifdef PRCS_DEVEL
static PrettyStreambuf pretty_debug_stream(&stderr_stream, &option_debug);
#endif

PrettyOstream prcsoutput(&pretty_stdout_stream, NoError);
PrettyOstream prcsinfo(&pretty_severable_stream, NonFatalError);
PrettyOstream prcswarning(&pretty_severable_stream, NonFatalError);
PrettyOstream prcserror(&pretty_stderr_stream, NonFatalError);
#ifdef PRCS_DEVEL
PrettyOstream prcsdebug(&pretty_debug_stream, NonFatalError);
#endif

QueryOstream prcsquery(&query_stream, &pretty_query_stream,
		       &stdout_stream, &stderr_stream);

char const default_fail_query_message[] = "Fail and abort PRCS";

static inline int next_word_length(const char* s, int n)
{
    int length = 0;

    while(length < n && *s && !isspace(*s)) {
	s+=1;
	length += 1;
    }

    return length;
}

static inline void advance_buffer(const char*&s, int& n, int& col, int a)
{
    s += a;
    n -= a;
    col += a;
}

int PrettyStreambuf::overflow(int c)
{
    line_buffer.append(c);
    if(isspace(c))
	xsputn(NULL, 0);

    return 1;
}

int PrettyStreambuf::xsputn(const char* s0, int n0)
{
    if (*dont_print)
	return n0;

    int wordlen;
    const char* s = line_buffer.cast();
    int n = line_buffer.length();

    for(int i = 0; i < 2; i += 1, line_buffer.truncate(0), s = s0, n = n0) {
	if(n < 1)
	    continue;

	while(true) {
	    if(n <= 0) {
		break;
	    } else if(col == 0 && prefix.length() > 0) {
		while(n > 0 && isspace(*s)) {
		    s += 1;
		    n -= 1;
		}

		if(n == 0)
		    break;

		if(!new_line) {
		    for(int j = prefix.length(); j; j -= 1) {
			forward->xsputn(" ", 1);
			col += 1;
		    }
		} else {
		    forward->xsputn(prefix.cast(), prefix.length());
		    col += prefix.length();
		}

		while(isspace(*s) && *s != '\n' && *s != '\0') {
		    n -= 1;
		    s += 1;
		}
		if(n < 1)
		    continue;
	    }

	    if(!breakon)
		wordlen = n;
	    else
		wordlen = next_word_length(s, n);

	    if(wordlen == 0) {
		if(*s == '\n' || (col + 1 > width)) {
		    new_line = false;
		    forward->xsputn("\n", 1);
		    advance_buffer(s, n, col, 1);
		    col = 0;
		} else {
		    forward->xsputn(s, 1);
		    advance_buffer(s, n, col, 1);
		}
	    } else if(wordlen + col <= width) {
		forward->xsputn(s, wordlen);
		advance_buffer(s, n, col, wordlen);
	    } else if((wordlen + prefix.length() >= width) && (col == prefix.length())) {
		forward->xsputn(s, wordlen);
		advance_buffer(s, n, col, wordlen);
	    } else if(wordlen == 1 && col == width) {
		new_line = false;
		forward->xsputn(s, 1);
		advance_buffer(s, n, col, wordlen);
		forward->xsputn("\n", 1);
		col = 0;
	    } else {
		new_line = false;
		forward->xsputn("\n", 1);
		col = 0;
	    }
	}
    }

    return n;
}

int PrettyStreambuf::sync()
{
    xsputn(NULL, 0);
    return forward->sync();
}

static int tty_width(int fileno)
{
    struct winsize ws;
    int width;

    if (ioctl(fileno, TIOCGWINSZ, &ws) == -1 || ws.ws_col == 0)
	width = 79;
    else
	width = ws.ws_col - 1;

    return width < min_tty_width ? min_tty_width : width;
}

void re_query()
{
    stdout_stream.xsputn(re_query_message, re_query_len);
    stdout_stream.sync();
}

void continue_handler(SIGNAL_ARG_TYPE)
{
    if(re_query_message)
	re_query();
}

void adjust_streams()
{
    if(option_force_resolution || option_report_actions) {
	Dstring prefix(strip_leading_path(argv_save[0]));
	prefix.append(": ");

	pretty_query_stream.set_fill_prefix(prefix);
    }
}

void setup_streams(int argc, char** argv)
{
    argv_save = argv;
    argc_save = argc;

    Dstring prefix(strip_leading_path(argv[0]));
    prefix.append(": ");

    int stdout_width = tty_width(STDOUT_FILENO);
    int stderr_width = tty_width(STDERR_FILENO);

    pretty_stdout_stream.set_fill_width(stdout_width);
    pretty_stderr_stream.set_fill_width(stderr_width);
    pretty_query_stream.set_fill_width(stdout_width);
    pretty_severable_stream.set_fill_width(stderr_width);
    pretty_stdout_stream.set_fill_prefix(prefix);
    pretty_stderr_stream.set_fill_prefix(prefix);
    pretty_severable_stream.set_fill_prefix(prefix);
#ifdef PRCS_DEVEL
    pretty_debug_stream.set_fill_width(stderr_width);
    pretty_debug_stream.set_fill_prefix("DEBUG: ");
#endif
}

ostream& __omanip_squote(ostream& o, const char* s)
{
    PrettyStreambuf& ps = ((PrettyOstream&)o).ostreambuf();

    ps.set_fill_break(false);

    Dstring copy;
    copy.append('`');

    char c;

    while ((c = *s++) != 0) {
	if (!isgraph(c) && c != ' ')
	    copy.sprintfa("\\%03o", c);
	else
	    copy.append(c);
    }

    copy.append('\'');
    o << copy.cast();

    ps.set_fill_break(true);

    return o;
}

ostream& __omanip_setcol(ostream& o, int col)
{
    PrettyStreambuf& ps = ((PrettyOstream&)o).ostreambuf();

    ps.xsputn(NULL, 0);
    ps.set_column(col);

    return o;
}

void kill_prefix(ostream& o)
{
    PrettyStreambuf &ps = ((PrettyOstream&)o).ostreambuf();

    ps.set_fill_prefix("");
}

ostream& __omanip_fileline(ostream& o, FileLine fl)
{
    PrettyStreambuf &ps = ((PrettyOstream&)o).ostreambuf();

    Dstring old_prefix(ps.fill_prefix());

    ps.set_fill_prefix("");

    o << fl.file << ":" << fl.line << ": ";

    ps.set_fill_prefix(old_prefix);

    return o;
}

void PrettyStreambuf::set_fill_width(int width0)
{
    width = width0;
}

void PrettyStreambuf::set_fill_prefix(const char* prefix0)
{
    prefix.assign(prefix0);
}

void PrettyStreambuf::set_fill_break(bool breakon0)
{
    breakon = breakon0;
}

int PrettyStreambuf::set_column(int col0)
{
    int diff = col0 - col;

    if(diff <= 0)
 	return 0;

    Dstring spaces(' ', diff);
    set_fill_break(false);
    xsputn(spaces.cast(), diff);
    set_fill_break(true);

    return col0;
}

void PrettyStreambuf::reset_column()
{
    col = 0;
    new_line = true;
}

ostream& prcsendl(ostream& s)
{
    PrettyStreambuf &ps = ((PrettyOstream&)s).ostreambuf();

    ps.xsputn("\n", 1);

    ps.reset_column();

    return s;
}

ostream& dotendl(ostream& s)
{
    PrettyStreambuf &ps = ((PrettyOstream&)s).ostreambuf();

    ps.xsputn(".\n", 2);

    ps.reset_column();

    return s;
}

ostream& perror(ostream& s)
{
    PrettyStreambuf &ps = ((PrettyOstream&)s).ostreambuf();

    s << ": " << strerror(errno) << '\n';

    ps.reset_column();

    return s;
}

PrettyStreambuf::PrettyStreambuf(streambuf *forward0, int* dont_print0)
    :streambuf(), forward(forward0), col(0),
     width(80), breakon(true), new_line(true), dont_print(dont_print0) { }

/* QueryOstream */
QueryOstream::QueryOstream(strstreambuf* base_stream0,
			   PrettyStreambuf* query_stream0,
			   stdiobuf* stdout_stream0,
			   stdiobuf* stderr_stream0)
    :PrettyOstream(query_stream0, NoError),
     base_stream(base_stream0),
     query_stream(query_stream0),
     stdout_stream(stdout_stream0),
     stderr_stream(stderr_stream0),
     val(0), string_val((const char*) 0), default_option(0),
     force_message(NULL), report_message(NULL),
     option_count(0), bang_flag(NULL) { }

ostream& QueryOstream::option_manip(QueryOption opt)
{
    options[option_count++] = opt;
    return *this;
}

ostream& QueryOstream::default_manip(QueryOption opt)
{
    default_option = option_count;
    options[option_count++] = opt;
    return *this;
}

ostream& QueryOstream::report_manip(const char* message)
{
    report_message = message;
    return *this;
}

ostream& QueryOstream::force_manip(const char* message)
{
    force_message = message;
    return *this;
}

ostream& QueryOstream::definput_manip(const char* input)
{
    default_input = input;
    return *this;
}

ostream& QueryOstream::bang(BangFlag* flag)
{
    bang_flag = flag;
    return *this;
}

ostream& QueryOstream::string_query_manip(const char* message)
{
    if(option_force_resolution && option_be_silent) {
	/* silence */
    } else if(option_force_resolution) {
	*this << " -- " << default_input << prcsendl;
	string_val = PrConstCharPtrError(default_input);
	stderr_stream->xsputn(base_stream->str(),
			      base_stream->out_waiting());
	stderr_stream->sync();
    } else if(option_report_actions) {
	*this << " -- " << default_input << prcsendl;
	string_val = PrConstCharPtrError(default_input);
	stdout_stream->xsputn(base_stream->str(),
			      base_stream->out_waiting());
	stdout_stream->sync();
    } else {
	*this << "[" << default_input << "] " << message;
	query_stream->sync();

	re_query_message = base_stream->str();
	re_query_len = base_stream->out_waiting();

	while(true) {
	    stdout_stream->xsputn(re_query_message, re_query_len);
	    stdout_stream->sync();

	    static Dstring *ans = NULL;

	    if (!ans) ans = new Dstring;

	    If_fail(read_string(stdin, ans)) {
		cout << endl;
		string_val = PrConstCharPtrError(UserAbort);
		break;
	    }

	    if(ans->length() > 0) {
		string_val = PrConstCharPtrError(ans->cast());
		break;
	    } else {
		string_val = PrConstCharPtrError(default_input);
		break;
	    }
	}
    }

    re_query_message = NULL;

    query_stream->reset_column();
    base_stream->freeze(0);
    base_stream->seekoff(0, ios::beg);
    option_count = 0;
    bang_flag = NULL;

    return *this;
}

ostream& QueryOstream::query_manip(const char* message)
{
    if(option_force_resolution && option_be_silent) {
	/* silence */
    } else if(option_report_actions) {
	*this << report_message << dotendl;
	query_stream->sync();
	stdout_stream->xsputn(base_stream->str(),
			      base_stream->out_waiting());
	stdout_stream->sync();
	val = options[default_option].val;
    } else if(option_force_resolution) {
	*this << force_message << dotendl;
	query_stream->sync();
	stderr_stream->xsputn(base_stream->str(),
			      base_stream->out_waiting());
	stderr_stream->sync();
	val = options[default_option].val;
    } else if(bang_flag && bang_flag->flag) {
	*this << force_message << dotendl;
	query_stream->sync();
	stdout_stream->xsputn(base_stream->str(),
			      base_stream->out_waiting());
	stdout_stream->sync();
	val = options[default_option].val;
    } else {
	char query_buf[40];
	int query_buf_index = 0;

	query_buf[query_buf_index++] = '(';

	for(int i = 0; i < option_count; i += 1)
	    query_buf[query_buf_index++] = options[i].let;

	query_buf[query_buf_index++] = 'q';

	if(bang_flag)
	    query_buf[query_buf_index++] = '!';

	query_buf[query_buf_index++] = '?';
	query_buf[query_buf_index++] = ')';
	query_buf[query_buf_index++] = '[';
	query_buf[query_buf_index++] = options[default_option].let;
	query_buf[query_buf_index++] = ']';
	query_buf[query_buf_index++] = ' ';
	query_buf[query_buf_index] = 0;

	*this << message << query_buf;

	query_stream->sync();

	re_query_message = base_stream->str();
	re_query_len = base_stream->out_waiting();

	while(true) {
	    char c;
	    int found = false;

	    stdout_stream->xsputn(re_query_message, re_query_len);
	    stdout_stream->sync();

	    c = get_user_char();

	    if(c == '\n') { /* User accepts default. */
		c = options[default_option].let;
	    } else if(c == '\0') { /* EOF. */
		cout << endl;
		val = PrCharError(UserAbort);
		break;
	    } else if(c == '?') { /* User needs help. */
		prcsoutput << "Valid options are:" << prcsendl;
		for(int i = 0; i < option_count; i += 1) {
		    prcsoutput << options[i].let << " -- ";
		    if(i == default_option)
			prcsoutput << "[default] ";
		    prcsoutput << options[i].descr << dotendl;
		}
		prcsoutput << "q -- " << default_fail_query_message << dotendl;
		if(bang_flag) {
		    prcsoutput << "! -- Take default action each time this query "
			"is reached" << dotendl;
		}
		continue;
	    } else if(bang_flag && c == '!') {
		val = options[default_option].val;
		bang_flag->flag = true;
		found = true;
	    } else if(c == 'q') {
		val = PrVoidError(UserAbort);
		found = true;
	    }

	    for(int i = 0; i < option_count; i += 1) {
		if(c == options[i].let) {
		    val = options[i].val;
		    found = true;
		}
	    }

	    if(found) break;
	}
    }

    re_query_message = NULL;

    query_stream->reset_column();
    base_stream->freeze(0);
    base_stream->seekoff(0, ios::beg);
    option_count = 0;
    bang_flag = NULL;

    return *this;
}

#if EXPLICIT_TEMPLATES
#include "prcserror.tl"
#endif

#if 0
PrVoidError test_query()
{
    char c;

    prcsquery << "Subproject file " << squote("file.prj") << " contains errors.  "
	      << force("Ignoring entire subproject")
	      << report("Ignore entire subproject")
	      << option('n', "Fail and abort PRCS.", UserAbort)
	      << defopt('y', "Ignore this subproject, as if it were omitted "
			 "from the file-or-directory list on the command line.")
	      << query("Ignore and continue");

    Return_if_fail(c << prcsquery);

    return NoError;
}

int main(int /*argc*/, const char** argv)
{
    setup_streams(argv[0]);
#if 0
    if(argc != 2) {
	cerr << "NO!" << dotendl;
	exit(1);
    }


    FILE* f = fopen(argv[1], "r");

    if(!f) {
	cerr << "open failed" << dotendl;
	exit(1);
    }

    Dstring s;

    read_string(f, &s);

    s.truncate(s.length() - 1);

    prcserror << s << squote("this is quoted, don't line break") << dotendl;

    prcserror << "The path is: " << setcol(40) << squote("a file") << dotendl;

    prcserror << "The path to the repository is: " << setcol(40) << squote("another file") << dotendl;
#endif
    If_fail(test_query()) {
	prcserror << "Command failed" << dotendl;
	exit(1);
    }
#if 0
    prcserror << "This has no newline";
#endif
    exit(0);
}
#endif
