/* -*-Mode: C++;-*-
 * $Id: projdesc.cc 1.9.1.1.1.4.1.17.1.5.1.14 Tue, 03 Sep 1996 20:09:11 -0700 jmacd $
 * projdesc.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


extern "C" {
#include <stdarg.h>
}

#include "prcs.h"
#include "sexp.h"
#include "fileent.h"
#include "projdesc.h"
#include "vc.h"
#include "repository.h"
#include "misc.h"
#include "hash.h"
#include "quick.h"
#include "diff.h"
#include "system.h"
#include "checkout.h"

#include <strstream.h>

/* prj_source_name is used to print syntax error messages. */
static const char* prj_source_name;
static Sexp* prj_substrings = NULL;

static void delete_sexp(Sexp* s, bool keep_space);
static PrSexpPtrError read_sexp();
static PrSexpPtrError read_list();

static Dstring* new_substring(void);
static PrVoidError read_keyword(Dstring* DS);
static bool read_whitespace(Dstring *DS);
static bool peek_whitespace();

ProjectDescriptor::~ProjectDescriptor()
{
    if(_file_entries) {
	foreach_fileent(fe_ptr, this)
	    delete(*fe_ptr);

	delete _file_entries;
    }

    if(sexps) delete sexps;

    if(substrings) {
        Sexp* s = substrings;
        while(s) {
	    Sexp* p = s;
            s = s->next();

            delete p->key();
	    delete p;
        }
    }

    if(!substrings) {
	/* if it was a fake project */
        delete _project_version_name;
        delete _project_version_major;
        delete _project_version_minor;
	delete _parent_version_name;
	delete _parent_version_major;
	delete _parent_version_minor;
        delete _project_description;
        delete _version_log;
        delete _checkin_time;
        delete _checkin_login;
    }

    if(_merge_selected_major) delete _merge_selected_major;
    if(_merge_selected_minor) delete _merge_selected_minor;
    if(_merge_data_w_lookup_table) delete _merge_data_w_lookup_table;
    if(_merge_data_c_lookup_table) delete _merge_data_c_lookup_table;
    if(_merge_data_s_lookup_table) delete _merge_data_s_lookup_table;
}

void ProjectDescriptor::init_fake_project()
{
    _parent_version_name = NULL;
    _parent_version_major = NULL;
    _parent_version_minor = NULL;
    _created_by_major = NULL;
    _created_by_minor = NULL;
    _created_by_micro = NULL;
    _new_version_log = NULL;
    _file_entries = new FileEntryPtrArray;
    substrings = NULL;
    sexps = NULL;
    _descends_from_major = NULL;
    _descends_from_minor = NULL;
    _merge_data_w_lookup_table = NULL;
    _merge_data_c_lookup_table = NULL;
    _merge_data_s_lookup_table = NULL;

    _project_version_name = new Dstring();
    _project_version_major = new Dstring();
    _project_version_minor = new Dstring("0");
    _parent_version_name = new Dstring();
    _parent_version_major = new Dstring();
    _parent_version_minor = new Dstring("0");
    _project_description = new Dstring("empty version");
    _version_log = new Dstring("empty version");
    _checkin_time = new Dstring();
    _checkin_time->sprintf("\"%s\"", get_utc_time());
    _checkin_login = new Dstring(get_login());
}

ProjectDescriptor* fake_project_descriptor()
{
    ProjectDescriptor* it = new ProjectDescriptor();

    it->init_fake_project();

    return it;
}

PrVoidError ProjectDescriptor::bad_project_file(Sexp *s, const char* mes)
{
    throw prcserror << fileline(prj_source_name, s->line_number())
		     << "Invalid project file, " << mes << dotendl;
}

PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
					      const char* filename)
{
    ProjectDescriptor* new_project = new ProjectDescriptor();

    Return_if_fail(new_project->init_from_file(project_full_name, filename, NULL));

    return new_project;
}

PrProjectDescriptorPtrError read_project_file(const char* project_full_name,
					      const char* file_name,
					      FILE* infile)
{
    ProjectDescriptor* new_project = new ProjectDescriptor();

    Return_if_fail(new_project->init_from_file(project_full_name, file_name, infile));

    return new_project;
}

PrVoidError ProjectDescriptor::init_repository_entry(const char* file, bool lock, bool create)
{
    Return_if_fail(_repository_entry << Rep_init_repository_entry(file, lock, create, true));

    return NoError;
}

void ProjectDescriptor::set_descends_from(const char* major, int minor)
{
    if(descends_from_major() &&
       strcmp(*descends_from_major(), "-*-") != 0)
	/* In case someone has already set descends-from, return. */
	return;

    if(!descends_from_major()) {

	Dstring* descends = new Dstring("(Descends-From ");
	descends->append(project_name());
	descends->append(' ');
	descends->append(major);
	descends->sprintfa(" %d)\n", minor);

	Sexp* new_substrings = new Sexp(descends);
	new_substrings->next(substrings);
	substrings->prev(new_substrings);
	substrings = new_substrings;

    } else {

        descends_from_name()->assign(project_name());
	descends_from_major()->assign(major);
	descends_from_minor()->sprintf("%d", minor);

    }
}

PrVoidError ProjectDescriptor::init_from_file(const char* full_name,
					      const char* filename, FILE* infile)
{
    Sexp *sexplist, *firstkey, *first, **lookup;
    SexpTable table;
    const char* last_slash;

    _project_full_name.assign(full_name);
    _project_file_path.assign(full_name);
    _project_file_path.append(".prj");
    _quick_elim = NULL;

    last_slash = strrchr(full_name, '/');

    if(last_slash) {
	_project_name = last_slash + 1;
	_project_path.assign(full_name);
	_project_path.truncate(last_slash - full_name + 1);
    } else {
	_project_name = full_name;
	/*_project_path.truncate(0);*/
    }

    _project_aux_path.sprintf("%s.%s.prcs_aux", _project_path.cast(), _project_name);

    if(infile == NULL) {
        if((infile = fopen(filename, "r")) == NULL)
	    throw prcserror << "Can't open file " << squote(filename) << perror;
    }

    Return_if_fail(sexplist << parse_prj_file(infile, filename));

    substrings = prj_substrings;
    sexps = sexplist;

    for(;!sexplist->is_empty();sexplist = sexplist->cdr()) {
        /* first points at the car of the list */
        first = sexplist->car();

        if ( !first->is_pair() )
            return bad_project_file(first, "Expected parenthesized form");

        if ( first->length() == 0 )
            return bad_project_file(first, "Empty list in project file");

        /* firstkey points to a keyword such as Files */
        firstkey = first->car();

	if( firstkey->is_pair() )
	    return bad_project_file(firstkey, "Expected a keyword");

	if ( (lookup = table.lookup(*firstkey->key())) != NULL )
	    return bad_project_file(firstkey, "Duplicated entry");

	table.insert(*firstkey->key(), first);
    }

    /* emit a warning if project names don't match? */

    Return_if_fail(created_by_prj_entry_func(&table));
    Return_if_fail(project_ver_prj_entry_func(&table));
    Return_if_fail(parent_ver_prj_entry_func(&table));
    Return_if_fail(project_desc_prj_entry_func(&table));
    Return_if_fail(version_log_prj_entry_func(&table));
    Return_if_fail(new_version_log_prj_entry_func(&table));
    Return_if_fail(checkin_time_prj_entry_func(&table));
    Return_if_fail(checkin_login_prj_entry_func(&table));
    Return_if_fail(files_prj_entry_func(&table));
    Return_if_fail(merge_parents_prj_entry_func(&table));
    Return_if_fail(new_merge_parents_prj_entry_func(&table));
    Return_if_fail(descends_from_prj_entry_func(&table));
    Return_if_fail(populate_ignore_prj_entry_func(&table));

    return NoError;
}

PrVoidError ProjectDescriptor::write_project_file(const char* filename) const
{
    project_version_name()->assign(project_name());

    if (strcmp(*parent_version_major(), "-*-") != 0)
	parent_version_name()->assign(project_name());

    created_by_major()->sprintf("%d", prcs_version_number[0]);
    created_by_minor()->sprintf("%d", prcs_version_number[1]);
    created_by_micro()->sprintf("%d", prcs_version_number[2]);

    Return_if_fail(make_subdirectories(filename));

    WriteableFile outfile;

    Return_if_fail(outfile.open(filename));

    Sexp* s = substrings;

    while (!s->is_empty()) {
	outfile.stream() << s;
	s = s->cdr();
    }

    Return_if_fail(outfile.close());

    return NoError;
}

void ProjectDescriptor::append_file(const char* s)
{
    last_files_dstring->append(s);
}

void ProjectDescriptor::delete_file(FileEntry *fe)
{
    fe->set_deleted();

    delete_sexp(fe->file_entry_sexp()->car()->prev(), true);
}

/* format should be a string consisting of 'U', 'N', 'L', 'S', and
 * 'A', for user, number, label, string and any.  returns false if a
 * match fails.  arguments following format are the addresses of
 * Dstring*s to fill in. an example is:
 *
 *     sexp_scan(s, "LLL", &d1, &d2, &d3);
 *
 * where dN will get the Nth label of three.
 *
 * if the first character of format is an asterisk, then "-*-" is
 * accepted as a label.  if the first cahracter of format is an 'e',
 * then empty lists are allowed.  */
static bool sexp_scan(Sexp* list, const char* format, ...)
{
    va_list args;
    char c;
    int len = list->length();
    const char* type = NULL;
    bool accept_ast = false;
    bool accept_empty = false;
    bool loop_done = false;

    while (!loop_done) {
	switch (*format) {
	case '*':
	    accept_ast = true;
	    format += 1;
	    break;
	case 'e':
	    accept_empty = true;
	    format += 1;
	    break;
	default:
	    loop_done = true;
	    break;
	}
    }

    va_start(args, format);

    if(!list->is_pair()) goto error;

    while((c = *format++)) {
	switch(c) {
	case 'N': type = "number"; break;
	case 'S': type = "string"; break;
	case 'L': type = "label"; break;
	case 'U': type = "login"; break;
	case 'A': type = "any"; break;
	}

	/* If there are no characters left, and we are expecting one, bail */
	if(len == 0) goto error;

	bool okay_empty = false;

	if(list->car()->is_pair()) {
	    if (accept_empty && list->car()->is_empty())
		okay_empty = true;
	    else
		goto error;
	}

	/* now check the type */
	if(!okay_empty && VC_check_token_match(*list->car()->key(), type) <= 0) {
	    if(!(accept_ast && strcmp(*list->car()->key(), "-*-") == 0))
		goto error;
	}

	Dstring** dptr = va_arg(args, Dstring**);

	if (dptr) {
	    if (okay_empty)
		(*dptr) = NULL;
	    else
		(*dptr) = list->car()->key();
	}

	list = list->cdr();
	len -= 1;
    }

    va_end(args);
    return len == 0;
error:
    va_end(args);
    return false;
}

PrVoidError ProjectDescriptor::created_by_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Created-By-Prcs-Version");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LNNN", NULL, &_created_by_major,
		      &_created_by_minor, &_created_by_micro )) {
	    return bad_project_file(s, "Badly formed Created-By-Prcs-Version entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Created-By-Prcs-Version "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::project_ver_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Project-Version");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LLLN", NULL, &_project_version_name,
		      &_project_version_major, &_project_version_minor )) {
	    return bad_project_file(s, "Badly formed Project-Version entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Project-Version "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::parent_ver_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Parent-Version");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "*LLLN", NULL, &_parent_version_name,
		      &_parent_version_major, &_parent_version_minor )) {
	    return bad_project_file(s, "Badly formed Parent-Version entry");
	} else {
	    bool major_absent = strcmp(*_parent_version_name, "-*-") == 0;
	    bool minor_absent = strcmp(*_parent_version_major, "-*-") == 0;
	    bool name_absent = strcmp(*_parent_version_minor, "-*-") == 0;

	    if(major_absent ^ minor_absent ||
	       major_absent ^ name_absent ||
	       name_absent ^ minor_absent)
		return bad_project_file(s, "illegal use of -*-");
	}
    } else {
        throw prcserror << "Missing project file attribute Parent-Version "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::project_desc_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Project-Description");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LS", NULL, &_project_description)) {
	    return bad_project_file(s, "Badly formed Project-Description entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Project-Description "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::version_log_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Version-Log");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LS", NULL, &_version_log)) {
	    return bad_project_file(s, "Badly formed Version-Log entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Version-Log "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::new_version_log_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("New-Version-Log");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LS", NULL, &_new_version_log)) {
	    return bad_project_file(s, "Badly formed New-Version-Log entry");
	}
    } else {
        throw prcserror << "Missing project file attribute New-Version-Log "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::checkin_time_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Checkin-Time");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LS", NULL, &_checkin_time)) {
	    return bad_project_file(s, "Badly formed Checkin-Time entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Checkin-Time "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

PrVoidError ProjectDescriptor::checkin_login_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Checkin-Login");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "LU", NULL, &_checkin_login)) {
	    return bad_project_file(s, "Badly formed Checkin-Login entry");
	}
    } else {
        throw prcserror << "Missing project file attribute Checkin-Login "
	    "in file " << squote(prj_source_name) << dotendl;
    }
    return NoError;
}

/* merge_notify, has_been_merged --
 *
 *     these two functions deserve documentation.  merge_notify is called to
 *     update the project file immediately after a merge action is taken, in
 *     merge_finish.  the MergeAction argument identifies the type of action
 *     taken.  the goal is to save all information and avoid offering to merge
 *     a particular working file twice.  The data saved is as follows:
 *
 *     (working_name common_name selected_name action)
 *
 *     file families are not saved because for common and selected versions
 *     they are well defined.  for working files, they are not guaranteed to
 *     exist.  both names and families may change with merges and users edits
 *     so tracking merges by them is not useful, in my opinion.
 */
void ProjectDescriptor::merge_notify(FileEntry* working,
				     FileEntry* common,
				     FileEntry* selected,
				     MergeAction ma)
{
    Dstring protect;

    _merge_data_insertion_point->append(" (");

    if (working) {
	protect_path(*working->working_path(), &protect);
	_merge_data_insertion_point->append(protect);
    } else {
	_merge_data_insertion_point->append("()");
    }

    _merge_data_insertion_point->append(' ');

    if (common) {
	protect_path(*common->working_path(), &protect);
	_merge_data_insertion_point->append(protect);
    } else {
	_merge_data_insertion_point->append("()");
    }

    _merge_data_insertion_point->append(' ');

    if (selected) {
	protect_path(*selected->working_path(), &protect);
	_merge_data_insertion_point->append(protect);
    } else {
	_merge_data_insertion_point->append("()");
    }

    _merge_data_insertion_point->sprintfa(" %c)", (char)ma);
}

/* algorithm: to be true,
 * for each of w,c,s, if a lookup of filename in the correct
 * table must succeed.
 */
bool ProjectDescriptor::has_been_merged(FileEntry* working,
					FileEntry* common,
					FileEntry* selected) const
{
    return
	(working  && _merge_data_w_lookup_table->isdefined(*working ->working_path())) ||
	(common   && _merge_data_c_lookup_table->isdefined(*common  ->working_path())) ||
	(selected && _merge_data_s_lookup_table->isdefined(*selected->working_path()));
}

bool ProjectDescriptor::merge_continuing(const char* major, const char* minor)
{
    /* If _new_merge_parents_prj_entry is NULL, insert either at the beginning
     * or end of substrings (and change Descends-From if you do the end)
     * a new entry (Merge-Parents (version_name)), set _merge_data_insertion_point
     * to                                     ^ this empty string,
     * otherwise, set search the list for version_name, and set the lookup
     * table to a populated table of the file entries in that list, if the
     * entry does not exist, add it to the end of the _new_merge_parents_prj_entry
     * list */
    _merge_data_w_lookup_table = new SexpTable(pathname_hash, pathname_equal);
    _merge_data_c_lookup_table = new SexpTable(pathname_hash, pathname_equal);
    _merge_data_s_lookup_table = new SexpTable(pathname_hash, pathname_equal);
    _merge_selected_major = new Dstring(major);
    _merge_selected_minor = new Dstring(minor);

    Dstring version_name;

    version_name.append(major);
    version_name.append('.');
    version_name.append(minor);

    Dstring prot_version;

    protect_path(version_name, &prot_version);

    if(!_new_merge_parents_prj_entry) {

	_merge_data_insertion_point = new Dstring;
	_merge_data_insertion_point->sprintf("(Merge-Parents)\n(New-Merge-Parents (%s",
					     prot_version.cast());

	Sexp* new_substrings = new Sexp(_merge_data_insertion_point);
	Sexp* new_substrings2 = new Sexp(new Dstring("))\n"));

	new_substrings->next(new_substrings2);
	new_substrings2->next(substrings);

	substrings->prev(new_substrings2);
	new_substrings2->prev(new_substrings);

	substrings = new_substrings;

    } else {

	Sexp* s = _new_merge_parents_prj_entry->cdr();

	while( !s->is_empty() ) {

	    if(strcmp(prot_version, *s->caar()->key()) == 0) {
		/* Merge with this version is in progress */

		s = s->car()->cdr();

		while( !s->is_empty() ) {

		    if (!s->caar()->is_pair()) {
			Dstring* leak = correct_path(*s->caar()->key());
			_merge_data_w_lookup_table->insert(*leak, s);
		    }

		    if (!s->car()->cadr()->is_pair()) {
			Dstring *leak = correct_path(*s->car()->cadr()->key());
			_merge_data_c_lookup_table->insert(*leak, s);
		    }

		    if (!s->car()->cdr()->cadr()->is_pair()) {
			Dstring *leak = correct_path(*s->car()->cdr()->cadr()->key());
			_merge_data_s_lookup_table->insert(*leak, s);
		    }

		    s = s->cdr();
		}

		_merge_data_insertion_point = s->empty_string();

		return true;
	    }

	    s = s->cdr();
	}

	_merge_data_insertion_point = s->empty_string();

	_merge_data_insertion_point->sprintf(" (%s ", prot_version.cast());

	/* This is SEVERE data abstraction violation.  (at least I commented it). */
	Dstring* forward_close = s->car()->next()->key();

	forward_close->prepend(')');
    }

    return false;
}

void ProjectDescriptor::merge_finish()
{
    if(!_merge_complete || _merge_incomplete) {
	prcsinfo << "Merge incomplete" << dotendl;
    } else {
	prcsinfo << "Merge against version "
		 << _merge_selected_major << '.'
		 << _merge_selected_minor << " complete" << dotendl;
    }
}

PrVoidError ProjectDescriptor::strongly_verify_merge_parents(Sexp* s)
{
    while( !s->is_empty() ) {
	Sexp* parent = s->car();

	Dstring this_version_major;
	Dstring this_version_minor;

	const char* last_period = strrchr(parent->car()->key()->cast(), '.');

	if(!last_period)
	    return bad_project_file(parent, "illegal version number");

	this_version_major.assign(parent->car()->key()->cast());
	this_version_major.truncate(last_period - parent->car()->key()->cast());

	this_version_minor.assign(last_period + 1);

	if(VC_check_token_match(this_version_major, "label") <= 0 ||
	   VC_check_token_match(this_version_minor, "number") <= 0)
	    return bad_project_file(parent, "illegal version number");

	if(!repository_entry()->lookup_version(this_version_major, this_version_minor))
	    return bad_project_file(parent, "version not found in repository");

	if (s->cdr()->is_empty()) {
	    set_descends_from(*project_version_major(),
			      atoi(*project_version_minor()));

	    project_version_major()->assign(this_version_major);
	    project_version_minor()->assign(this_version_minor);

	    prcsinfo << "Merged working version, checkin as version "
		     << project_version_major() << '.'
		     << project_version_minor() << dotendl;
	}

	s = s->cdr();
    }

    return NoError;
}

PrVoidError ProjectDescriptor::adjust_merge_parents()
{
    if (_new_merge_parents_prj_entry &&
	_new_merge_parents_prj_entry->length() > 1) {
	Dstring *old_merge_parents;

	if (_merge_parents_prj_entry) {
	    delete_sexp(_merge_parents_prj_entry->car()->prev(), false);
	    old_merge_parents = _merge_parents_prj_entry->car()->prev()->key();
	} else {
	    old_merge_parents = new Dstring;

	    Sexp* new_substrings = new Sexp(old_merge_parents);

	    new_substrings->next(substrings);
	    substrings->prev(new_substrings);

	    substrings = new_substrings;
	}

	ostrstream os;

	Sexp* s = _new_merge_parents_prj_entry->cdr();

	os << s;

	old_merge_parents->assign("(Merge-Parents ");

	ASSERT(os.rdbuf()->str(), "what the hell");

	old_merge_parents->append(os.rdbuf()->str() + 1);
	old_merge_parents->append("\n");

	os.rdbuf()->freeze(0);
	os.rdbuf()->seekoff(0, ios::beg);

	Return_if_fail(strongly_verify_merge_parents(s));

	delete_sexp(_new_merge_parents_prj_entry->car()->prev(), false);

	_new_merge_parents_prj_entry->car()->key()->assign("(New-Merge-Parents)\n");
    } else if (_merge_parents_prj_entry &&
	       _merge_parents_prj_entry->length() > 1) {

	delete_sexp(_merge_parents_prj_entry->car()->prev(), false);
	_merge_parents_prj_entry->car()->prev()->key()->
	    assign("(Merge-Parents)\n");
    }

    return NoError;
}

PrVoidError ProjectDescriptor::new_merge_parents_prj_entry_func(SexpTable* table)
{
    _merge_selected_major = NULL;
    _merge_selected_minor = NULL;
    _merge_complete = false;
    _merge_incomplete = false;

    /* these four will be set by a call to merge_continuing */
    _merge_data_w_lookup_table = NULL;
    _merge_data_c_lookup_table = NULL;
    _merge_data_s_lookup_table = NULL;
    _merge_data_insertion_point = NULL;

    _new_merge_parents_prj_entry = NULL;

    Sexp** lookup;
    lookup = table->lookup("New-Merge-Parents");

    if (lookup) {
	_new_merge_parents_prj_entry = *lookup;

	return verify_merge_parents(*lookup);
    }

    return NoError;
}

PrVoidError ProjectDescriptor::verify_merge_parents(Sexp* s)
{
    if(s->length() == 1)
	/* an empty list (New-Merge-Parents) */
	return NoError;

    /* sexp_scan can't really help here */

    s = s->cdr(); /* s is the first elt of the list */

    while( !s->is_empty() ) {
	Sexp* parent = s->car();

	if(!parent->is_pair())
	    return bad_project_file(parent, "expected a list");

	if(parent->car()->is_pair())
	    return bad_project_file(parent->car(), "expected a version name");

	parent = parent->cdr(); /* parent is the first file pair */

	while ( !parent->is_empty() ) {

	    Sexp *entry = parent->car();

	    if(!entry->is_pair())
		return bad_project_file(entry, "expected a list");

	    if(!sexp_scan(parent->car(), "eAAAM", NULL, NULL, NULL, NULL))
		return bad_project_file(entry, "ill formed merge action");

	    Sexp* merge_action_sexp = entry->cdr()->cdr()->cadr();

	    if (merge_action_sexp->is_pair())
		return bad_project_file(merge_action_sexp, "ill formed merge action");

	    Dstring *merge_action_str = merge_action_sexp->key();

	    if(merge_action_str->length() != 1)
		return bad_project_file(parent->car()->cadr(), "ill formed merge action");

	    MergeAction action = (MergeAction)merge_action_str->index(0);

	    switch(action) {
	    case MergeActionMerge:
	    case MergeActionNothing:
	    case MergeActionDelete:
	    case MergeActionAdd:
	    case MergeActionReplace:
		break;
	    default:
		return bad_project_file(parent->car()->cadr(), "ill formed merge action");
	    }

	    parent = parent->cdr();
	}

	s = s->cdr();
    }

    return NoError;
}

PrVoidError ProjectDescriptor::merge_parents_prj_entry_func(SexpTable* table)
{
    _merge_parents_prj_entry = NULL;

    Sexp** lookup;
    lookup = table->lookup("Merge-Parents");

    if (lookup) {
	_merge_parents_prj_entry = *lookup;

	return verify_merge_parents(*lookup);
    }

    return NoError;
}

PrVoidError ProjectDescriptor::descends_from_prj_entry_func(SexpTable* table)
{
    _descends_from_major = NULL;
    _descends_from_minor = NULL;
    _descends_from_name = NULL;

    Sexp** lookup;
    lookup = table->lookup("Descends-From");
    if(lookup) {
	Sexp *s = *lookup;

	if(!sexp_scan(s, "*LLLN", NULL, &_descends_from_name,
		      &_descends_from_major, &_descends_from_minor )) {
	    _descends_from_major = NULL;
	    _descends_from_minor = NULL;
	    _descends_from_name = NULL;
	} else {
	    bool major_absent = strcmp(*_descends_from_major, "-*-") == 0;
	    bool minor_absent = strcmp(*_descends_from_minor, "-*-") == 0;
	    bool name_absent = strcmp(*_descends_from_name, "-*-") == 0;

	    if(major_absent ^ minor_absent ||
	       major_absent ^ name_absent ||
	       name_absent ^ minor_absent)
		return bad_project_file(s, "illegal use of -*-");
	}
    }

    return NoError;
}

PrVoidError ProjectDescriptor::individual_insert_fileent(Sexp* file, FileEntry* entry)
{
    int length;

    if(!file->is_pair())
	return bad_project_file(file, "expected a list");

    length = file->length();

    if(length < 2)
	return bad_project_file(file, "file entry too short");

    if(file->car()->is_pair())
	return bad_project_file(file, "expected a filename");

    if(!file->cadr()->is_pair())
	return bad_project_file(file, "malformed internal file identifier");

    /* _work_path is set -- memory allocated by correct_path */
    Dstring full_work_path(project_path());

    full_work_path.append(*file->car()->key());

    entry->set_working_path(correct_path(full_work_path));

    if(weird_pathname(*file->car()->key()))
	return bad_project_file(file, "pathnames may not contain `../' or `./', end "
				"in '/', or begin with '/'");

    entry->set_file_descriptor(file->cadr());

    entry->set_file_entry_sexp(file);

    file = file->cdr()->cdr();
    length -= 2;

    bool saw_real_file = false;

    while(length > 0) {
	if(file->car()->is_pair() || file->car()->key()->index(0) != ':')
	    return bad_project_file(file, "invalid file attribute");

	const char* attribute = *file->car()->key();

	Dstring prot_att;

	protect_path(attribute, &prot_att);

	entry->append_tag(prot_att); /* Append all tags, otherwise its an error */

	if(strcmp(prot_att, ":real-file") == 0) {
	    saw_real_file = true;
	} else if(strcmp(prot_att, ":no-keywords") == 0) {
	    entry->set_keyword_sub(false);
	} else if(strcmp(prot_att, ":directory") == 0) {
	    if(entry->file_type() != RealFile)
		return bad_project_file(file, "multiple file type attributes");

	    entry->set_file_type(Directory);
	} else if(strcmp(prot_att, ":symlink") == 0) {
	    if(entry->file_type() != RealFile)
		return bad_project_file(file, "multiple file type attributes");

	    entry->set_file_type(SymLink);
	} else if(strncmp(prot_att, ":tag", 4) == 0) {
	    /* its okay */
	} else {
	    return bad_project_file(file, "unrecognized attribute");
	}

	file = file->cdr();
	length -= 1;
    }

    if (saw_real_file && entry->file_type() != RealFile)
	return bad_project_file(file, "multiple file type attributes");

    if ( (entry->file_type() == Directory || entry->file_type() == SymLink ) ) {
	/* keyword_sub() defaults to true, so turn it off for these, and warn
	 * if it was already done */
	if(!entry->keyword_sub()) {
	    prcswarning << "warning: You may not replace keywords on the irregular files"
			<< squote(entry->working_path()) << dotendl;
	}

	entry->set_keyword_sub(false);
    }

    if (entry->file_type() == SymLink || entry->file_type() == Directory) {
	if(entry->file_descriptor()->length() > 1 ||
	    entry->file_descriptor()->car()->is_pair()) {
	    return bad_project_file(file, "malformed internal file descriptor");
	}
    } else {
	if(entry->file_descriptor()->length() == 0) {

	    /* its empty */
	    entry->set_descriptor_version_number(new Dstring);
	    entry->set_internal_descriptor(new Dstring);
	    entry->set_file_mode(new Dstring("644"));

	} else {
	    if(entry->file_descriptor()->length() == 2) {
		if(entry->file_descriptor()->car()->is_pair() ||
		   entry->file_descriptor()->cadr()->is_pair()) {
		    return bad_project_file(file, "malformed internal file descriptor");
		}

		/* its the old format -- no file mode */
		entry->set_internal_descriptor(new Dstring(*entry->file_descriptor()->car()->key()));
		entry->set_descriptor_version_number(correct_path(*entry->file_descriptor()->cadr()->key()));
		entry->set_file_mode(new Dstring("644"));

	    } else if(entry->file_descriptor()->length() == 3) {
		if(entry->file_descriptor()->car()->is_pair() ||
		   entry->file_descriptor()->cadr()->is_pair() ||
		   entry->file_descriptor()->cdr()->cadr()->is_pair()) {
		    return bad_project_file(file, "malformed internal file descriptor");
		}

		/* it has the file mode */
		entry->set_file_mode(correct_path(*entry->file_descriptor()->cdr()->cadr()->key()));
		entry->set_internal_descriptor(new Dstring(*entry->file_descriptor()->car()->key()));
		entry->set_descriptor_version_number(correct_path(*entry->file_descriptor()->cadr()->key()));

	    } else {
		return bad_project_file(file, "malformed internal file descriptor");
	    }

	    Dstring* corr_path = correct_path(*entry->internal_descriptor());
	    const char* first_slash = strchr(*corr_path, '/');

	    if(first_slash == NULL || strlen(first_slash) < 4) {  /* /0_a is legal (4 chars) */
		return bad_project_file(file, "malformed internal file descriptor");
	    }

	    if(strncmp(*corr_path, project_name(), first_slash - *corr_path) != 0) {
		return bad_project_file(file, "you may not use cross project file identifiers");
	    }

	    first_slash += 1;

	    Dstring *descriptor_name = new Dstring;

	    descriptor_name->assign(first_slash);
	    descriptor_name->append(",v");

	    delete corr_path;

	    entry->set_descriptor_name(descriptor_name);
	}
    }

    return NoError;
}

PrVoidError ProjectDescriptor::verify_project_file()
{
    /* This code insures:
     * no descriptor matches the project file's.
     * no filename matches the project file's.
     * no descriptor matches another.
     * no filename matches another.
     * no filename is a prefix of another unless it has :directory
     */

    Dstring project_descriptor;

    project_descriptor.sprintf("%s.prj,v", project_name(), project_name());

    /* Using the project_file_path and project_file_aux will
     * be wrong later */
    Dstring project_file_name(project_file_path());
    Dstring project_aux_name(project_aux_path());

    PathTable descriptors(pathname_hash, pathname_equal),
	      filenames  (pathname_hash, pathname_equal);

    descriptors.insert(project_descriptor.cast(), RealFile);
    filenames.insert(project_file_name.cast(), RealFile);
    filenames.insert(project_aux_name.cast(), RealFile);

    foreach_fileent(fe_ptr, this) {
	FileEntry* fe = *fe_ptr;

	const char* fe_name = *fe->working_path();

	if(filenames.isdefined(fe_name)) {
	    if(strcmp(fe_name, project_file_name) == 0)
		return bad_project_file(fe->file_descriptor(),
					"entry duplicates project file name");
	    else if (strcmp(fe_name, project_aux_name) == 0)
		return bad_project_file(fe->file_descriptor(),
					"entry duplicates project auxiliary file name");
	    else
		return bad_project_file(fe->file_descriptor(), "duplicate file name");
	}

	filenames.insert(fe_name, fe->file_type());

	if(fe->file_type() != RealFile || !fe->descriptor_name())
	    continue;

	const char* fe_desc = *fe->descriptor_name();

	if(descriptors.isdefined(fe_desc)) {
	    if(strcmp(fe_desc, project_descriptor) == 0)
		return bad_project_file(fe->file_descriptor(), "entry duplicates project file descriptor");
	    else
		return bad_project_file(fe->file_descriptor(), "duplicate file descriptor");
	}

	descriptors.insert(fe_desc, RealFile);
    }

    foreach_fileent(fe_ptr, this) {
	FileEntry* fe = *fe_ptr;

	Dstring name = *fe->working_path();

	for(int i = name.length() - 1; i > 0; i -= 1) {
	    if (name.index(i) == '/') {
		name.truncate(i);

		FileType* ft = filenames.lookup(name);

		if (ft && (*ft) != Directory) {
		    Dstring msg;

		    msg.sprintf("non-directory filename `%s' is prefix to another file", name.cast());

		    return bad_project_file(fe->file_descriptor(), msg);
		}
	    }
	}
    }

    return NoError;
}

PrVoidError ProjectDescriptor::files_prj_entry_func(SexpTable* table)
{
    Sexp** lookup;
    lookup = table->lookup("Files");
    if(lookup) {
	int i, count;
	Sexp *s = *lookup;

	count = s->length() - 1;

	_file_entries = new FileEntryPtrArray(count);

	/* advance past "Files" and begin inserting file entries */
	for(s = s->cdr(), i = 0; i < count;i += 1, s = s->cdr()) {
	    FileEntry* fe = new FileEntry;

	    Return_if_fail(individual_insert_fileent(s->car(), fe));

	    _file_entries->append(fe);
	}

	Return_if_fail(verify_project_file());

	last_files_dstring = s->empty_string();
    } else {
        throw prcserror << "Missing project file attribute Files "
	    "in file " << squote(prj_source_name) << dotendl;
    }

    return NoError;
}

PrVoidError ProjectDescriptor::populate_ignore_prj_entry_func(SexpTable* table)
{
    _populate_ignore = NULL;

    Sexp** lookup, *s, *t;
    lookup = table->lookup("Populate-Ignore");

    if(lookup) {
	s = *lookup;

	if(s->cdr()->is_empty() || !s->cadr()->is_pair())
	    return bad_project_file(s, "expected a list");

	t = s->cadr();

	while(!t->is_empty()) {
	    if(t->car()->is_pair())
		return bad_project_file(t, "expected a string");

	    Dstring *pat = t->car()->key();
	    if(pat->length() < 2 ||
	       pat->index(0) != '\"' ||
	       pat->index(pat->length() - 1) != '\"')
		return bad_project_file(t, "expected a string");

	    t = t->cdr();
	}

	_populate_ignore = s->cadr();
    }

    return NoError;
}

void ProjectDescriptor::read_quick_elim()
{
    _quick_elim = NULL;

    If_fail(_quick_elim << ::read_quick_elim(project_aux_path()))
	return;
}

void ProjectDescriptor::quick_elim_update()
{
    if(!_quick_elim)
	_quick_elim = new QuickElim;

    _quick_elim->update_file(project_aux_path());
}

void ProjectDescriptor::quick_elim_update_fe(FileEntry* fe)
{
    if(!_quick_elim)
	_quick_elim = new QuickElim;

    _quick_elim->update(fe);
}

static void delete_sexp(Sexp* s, bool keep_space)
{
    Dstring *ds;
    char c;
    int depth = 0, l;
    bool escaped = false;

    do {
        ds = s->key();
        l = ds->length();

        for(int i = 0; i < l; i += 1) {
            c = ds->index(i);
            if(escaped) {
                escaped = false;
            } else if(c == '\\') {
                escaped = true;
            } else if(c == '(') {
                depth += 1;
            } else if(c == ')') {
                depth -= 1;
            }
	    if(keep_space) {
		if(!isspace(c))
		    ds->index(i, ' ');
	    }
        }

	ds->truncate(0);

        s = s->next();
    } while(depth > 0);

}

/*
 * the parser --
 *
 *     files is treated as if it was preceded by an open paren and closed
 *     by a closed parent after the file is read.
 */
static FILE* prjis;      /* the file being read */
static bool one_close;   /* true when one closing paren has been used */
static int linecount;    /* line count in current file */

/*
 * readchar --
 *
 *     reads a character from the input stream into c, also checking for eof
 *     and incrementing the linecount.  the return value is generally used
 *     such that false return indicates an error, except when in a
 *     whitespace/comment area.
 */
static bool readchar(char &c)
{
    int ch;

    If_fail(ch << Err_fgetc(prjis))
	return false;

    if (ch == EOF)
	return false;

    if (ch == '\n')
        linecount += 1;

    c = (char)ch;

    return true;
}

/*
 * putbackchar --
 *
 *     just insures that the linecount is correct when a char is returned
 */
static void putbackchar(char c)
{
    ungetc((int)c, prjis);
    if ( c == '\n' )
        linecount -= 1;
}

/*
 * peek_whitespace --
 *
 *     optimization to not insert extra null comment strings when one
 *     won't be read.  returns true if a non-empty comment will be read.
 */
static bool peek_whitespace()
{
    char c;
    if(!readchar(c)) return false;
    putbackchar(c);
    if(isspace(c) || c == ';') return true;
    return false;
}

/*
 * new_substring --
 *
 *     supplies a new substring, inserting it into the list of substrings
 */
static Sexp* new_sexp_substring()
{
    Dstring* DS = new Dstring();
    Sexp* S = new Sexp(DS, linecount);
    S->prev(prj_substrings);
    if(prj_substrings)
	prj_substrings->next(S);
    prj_substrings = S;
    return S;
}

static Dstring* new_substring()
{
    return new_sexp_substring()->key();
}

PrSexpPtrError parse_prj_file(FILE* infile, const char* filename)
{
    prj_source_name = filename;
    prjis = infile;
    one_close = false;
    linecount = 1;

    prj_substrings = NULL;
    prj_substrings = new_sexp_substring();

    Sexp* first = prj_substrings;
    Sexp* r;

    if(peek_whitespace() && !read_whitespace(new_substring()))
	throw prcserror << fileline(prj_source_name, linecount)
			 << "Empty project file" << dotendl;

    Return_if_fail(r << read_list());

    if(r->is_empty())
        throw prcserror << fileline(prj_source_name, linecount)
			 << "Empty project file" << dotendl;

    if(!one_close)
	throw prcserror << fileline(prj_source_name, linecount)
			 << "Trailing data in project file" << dotendl;

    fclose(infile);

    prj_substrings = first;

    return r;
}

static bool read_whitespace(Dstring *DS)
    /* read all whitespace and comments into DS */
{
    char c;

    do{
        if (!readchar(c)) return false; /* false if eof, convention in
                                           this file */
        if ( isspace(c) )
            DS->append(c);
        else if ( c == ';' ){
            do {
                DS->append(c);
                if (!readchar(c)) return false;
            } while( c != '\n' );
            DS->append(c);
        }
    } while ( isspace(c) || c == ';' );

    putbackchar(c);

    return true;
}

static PrVoidError read_keyword(Dstring* DS)
    /* reads one keyword or string from the file, any char escaped by a
       '\' will be read, so this applies to a quote mark or '\' */
{
    bool in_string = false, escaped = false;
    int start = linecount; /* remember where it starts, in case there
                              are unbalanced quotes */
    char c;

    if ( !readchar(c) )
	throw prcserror << fileline(prj_source_name, linecount)
			 << "Unexpected EOF" << dotendl;

    while ( in_string || escaped || ( !isspace(c) && c != ')' && c != '(' ) ) {
        if ( c == '\"' && !escaped )
            in_string = !in_string;
        if ( c == '\\' && !escaped )
            escaped = true;
        else
            escaped = false;
        DS->append(c);
        if ( !readchar(c) ) {
            if (in_string)
                throw prcserror << fileline(prj_source_name, start)
				 << "Unterminated string constant in project file" << dotendl;

	    throw prcserror << fileline(prj_source_name, linecount)
			     << "Unexpected EOF";
        }
    }

    putbackchar(c);

    return NoError;
}

static PrSexpPtrError read_list()
    /* reads a list, assumes the opening paren has already been read, and
       reads sexps until a closing paren is read. */
{
    Sexp *s;
    Dstring *DS;
    char c;

    if (!readchar(c)) {
        if(one_close) {
            throw prcserror << fileline(prj_source_name, linecount)
			     << "Too few closing parentheses" << dotendl;
        } else {
            one_close = true;
            return new Sexp(new_sexp_substring(), NULL, linecount);
        }
    }

    if(c == ')') {
        s = new_sexp_substring();
        DS = new_substring();
        DS->append(')');
        read_whitespace(DS);
        return new Sexp(s, NULL, linecount);
    }

    putbackchar(c);

    Return_if_fail(s << read_sexp());

    if(peek_whitespace())
        read_whitespace(new_substring());

    Sexp *r;

    Return_if_fail(r << read_list());

    return new Sexp(s, r, linecount);
}

static PrSexpPtrError read_sexp()
{
    char c;
    Sexp* s;
    Dstring *DS;

    readchar(c);

    if ( c == '(' )
    {
	DS = new_substring();
        DS->append(c);
        /* if we're starting a list, its possible that there is whitespace
           between the open paren and the first sexp in the list, so read
           some more whitespace and comments */
        read_whitespace(DS);
        return read_list();
    } else {
        /* put back the character, read a keyword, then eat up some
           more whitespace and comments */
        putbackchar(c);

	s = new_sexp_substring();

        Return_if_fail(read_keyword(s->key()));

	if(peek_whitespace())
            read_whitespace(new_substring());

	return s;
    }
}

#if 0
int main(int argc, char** argv)
{
    FILE* foo;
    Sexp* s;

    Dstring *d1=NULL, *d2=NULL, *d3=NULL, *d4=NULL;

    if(argc != 3) {
	fprintf(stderr, "wrong # of args\n");
	return 1;
    }

    if((foo = fopen(argv[1], "r")) == NULL ||
       (s = parse_prj_file(foo)) == NULL) {
	fprintf(stderr, "something failed\n");
	return 1;
    }

    if(!sexp_scan(s->car(), argv[2], &d1, &d2, &d3, &d4)) {
	fprintf(stderr, "you lose\n");
	return 1;
    }

    cout << s << prcsendl;
    cout << d1 << " " << d2 << " " << d3 << " " << d4 << prcsendl;
}
#endif

#if 0
int main(int argc, char** argv)
{
    setup_streams(argv[0]);

    if(argc != 2) {
	prcserror << "two args" << dotendl;
	exit(1);
    }

    FILE* in = fopen(argv[1], "r");

    if(!in) {
	prcserror << "failed opening it" << dotendl;
	exit(1);
    }

    prj_source_name = "bogus";

    Sexp* s;

    If_fail(s << parse_prj_file(in)) {
	prcserror << "oops" << dotendl;
	exit(1);
    }

    s = prj_substrings;

    while ( s )
    {
        cout << s->key();
        s = s->next();
    }

}
#endif
