/* -*-Mode: C++;-*-
 * $Id: quick.cc 1.7.1.2 Wed, 21 Aug 1996 13:45:07 -0700 jmacd $
 * quick.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include "prcs.h"

#include "sexp.h"
#include "quick.h"
#include "projdesc.h"
#include "fileent.h"
#include "system.h"
#include "misc.h"

/* This file leaks memory bigtime, in obvious ways.  However, since
 * its always run right before prcs exits, I am not going to worry
 * about it now. */

NprQuickElimPtrError read_quick_elim(const char* filename)
{
    QuickElim* quick_elim = new QuickElim;

    Return_if_fail(quick_elim->init_from_file(filename));

    return quick_elim;
}

static long string_to_posint(const char* string)
{
    char c;
    int val = 0;

    while((c = *string++) != 0) {
	if(c < '0' || c > '9')
	    return -1;
	val *= 10;
	val += (c - '0');
    }

    return val;
}

bool QuickElim::unchanged(FileEntry* fe)
{
    Sexp **lookup, *s;
    int length;
    time_t mtime;

    if((lookup = quick_table->lookup(fe->working_path()->cast() +
				     cmd_root_project_path_len)) != NULL) {
	s = (*lookup)->cdr();

	if((length = string_to_posint(*s->car()->key())) < 0)
	    return false;

	if(length != fe->stat_length())
	    return false;

	s = s->cdr();

	if((mtime = string_to_posint(*s->car()->key())) < 0)
	    return false;

	if(mtime != fe->stat_mtime())
	    return false;

	s = s->cdr();

	if(strcmp(*s->car()->key(), *fe->internal_descriptor()) != 0)
	    return false;

	s = s->cdr();

	if(strcmp(*s->car()->key(), *fe->descriptor_version_number()) != 0)
	    return false;

	return true;
    }

    return false;
}

void QuickElim::update(FileEntry* fe)
{
    Sexp **lookup, *s;

    if(option_report_actions)
	return;

    if((lookup = quick_table->lookup(fe->working_path()->cast() +
				     cmd_root_project_path_len)) != NULL) {
	s = (*lookup)->cdr();
    } else {
	s = new_empty_list();

	for(int i = 0; i < 5; i += 1)
	    s = new Sexp(new Sexp(new Dstring), s);

	Dstring prot_path;

	protect_path(*fe->working_path() + cmd_root_project_path_len, &prot_path);

	s->car()->key()->assign(prot_path);

	new_sexps = new Sexp(s, new_sexps);

	s = s->cdr();
    }

    s->car()->key()->sprintf("%d", fe->stat_length());
    s = s->cdr();
    s->car()->key()->sprintf("%d", fe->stat_mtime());
    s = s->cdr();
    s->car()->key()->assign(*fe->internal_descriptor());
    s = s->cdr();
    s->car()->key()->assign(*fe->descriptor_version_number());
}

PrVoidError QuickElim::update_file(const char* filename)
{
    ofstream os;

    if(option_report_actions)
	return NoError;

    os.open(filename);

    os << ";; This file is automatically generated, editing may cause PRCS to do\n;; REALLY bad things.\n";

    os << "(Created-By-Prcs-Version " << prcs_version_number[0] << " "
       << prcs_version_number[1] << " " << prcs_version_number[2] << ")\n";

    while ( first_entry && !first_entry->is_empty() ) {
        os << first_entry->car() << "\n";
        first_entry = first_entry->cdr();
    }

    while ( new_sexps && !new_sexps->is_empty() ) {
	os << new_sexps->car() << "\n";
	new_sexps = new_sexps->cdr();
    }

    os.close();

    if(os.fail()) {
	throw prcserror << "Failed opening " << squote(filename) << " to write auxiliary file, ignoring" << dotendl;
    }

    return NoError;
}

NprVoidError QuickElim::init_from_file(const char* file)
{
    FILE* quick_in;

    If_fail(quick_in << Err_fopen(file, "r")) {
	if(errno != ENOENT)
	    prcswarning << "Error opening auxiliary file " << squote(file) << dotendl;

	return NonFatalError;
    }

    If_fail(quick_sexps << parse_prj_file(quick_in, file)) {
	prcswarning << "Error parsing auxiliary file " << squote(file) << ", removing" << dotendl;
	If_fail(Err_unlink(file))
	    prcswarning << "Unlink failed on auxiliary file " << squote(file) << perror;

	return NonFatalError;
    }

    Sexp* s = quick_sexps;

    if(!s->car()->is_pair() || s->car()->length() != 4) {
	goto error;
    } else {
	Sexp* sc = s->car();

	if(sc->car()->is_pair())
	    goto error;

	if(strcmp(*sc->car()->key(), "Created-By-Prcs-Version") != 0)
	    goto error;

	s = s->cdr();
    }

    first_entry = s;

    while(!s->is_empty()) {
	Sexp* sc = s->car();

	if(!sc->is_pair() || sc->length() != 5)
	    goto error;

	for(int i = 0; i < 5; i += 1) {
	    if(sc->car()->is_pair())
		goto error;
	    else
		sc = sc->cdr();
	}

	Dstring *corrected;

	/* leak */
	corrected = correct_path(s->car()->car()->key()->cast());

	quick_table->insert(corrected->cast(), s->car());

	s = s->cdr();
    }

    return NoError;

error:

    prcswarning << "Error parsing auxiliary file " << squote(file) << ", removing" << dotendl;

    If_fail(Err_unlink(file))
	prcswarning << "Unlink failed on auxiliary file " << squote(file) << perror;

    return NoError;
}
