/* -*-Mode: C++;-*-
 * $Id: rebuild.cc 1.3.1.1.1.8.1.21.1.1 Tue, 03 Sep 1996 16:48:15 -0700 jmacd $
 * rebuild.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


extern "C" {
#include "md5.h"
}

#include "prcs.h"

#include "repository.h"
#include "misc.h"
#include "vc.h"
#include "setkeys.h"
#include "checkout.h"
#include "hash.h"
#include "system.h"
#include "memseg.h"
#include "rebuild.h"
#include "prcsdir.h"
#include "fileent.h"
#include "projdesc.h"
#include "syscmd.h"
#include "checkin.h"

static const char data_file_header[] = "This file contains data summarizing the project: ";
/* Its important that the two summary_headers are the same length! */
static const char project_summary_header[]  = "Project version count: ";
static const char rcs_file_summary_header[] = "Summary for RCS file:  ";
static const char version_count_header[] = "Version count: ";
const char prcs_data_file_name[] = "prcs_data";

static MemorySegment read_buf;

static PrRcsDeltaPtrError rebuild_repository_file(const char* name);

/**********************************************************************/
/*                        Miscellaneous Debris                        */
/**********************************************************************/

/* Simply compute the md5 checksum of a memory segment */
char* md5_buffer(const char* buffer, int buflen)
{
    MD5_CTX context;
    static unsigned char digest[16];

    MD5Init(&context);
    MD5Update(&context, (unsigned char*)buffer, buflen);
    MD5Final(digest, &context);

    return (char*)digest;
}

/* Determines whether a file is garbage by looking at its extension and
 * for the "prcs_" prefix. */
static bool is_garbage_by_name(const char* pathname)
{
    const char* name = strip_leading_path(pathname);
    int len = strlen(name);

    if(strcmp(name + len - 2, ",v") == 0)
	return false;
    else if(strncmp(name, "prcs_", 5) == 0)
	return false;
    else
	return true;
}

/* Determines whether a file is used internally by PRCS, by looking at the
 * prefix. */
static bool is_prcs_file(const char* pathname, const char* project_file_name)
{
    const char* name = strip_leading_path(pathname);
    const char* projname = strip_leading_path(project_file_name);

    if(strncmp(name, "prcs_", 5) == 0)
	return true;
    else if(strcmp(name, projname) == 0)
	return true;
    else
	return false;
}

/* Writes the data file header, consisting of the string "This file contains ..."
 * and the project name */
static void write_data_file_header(ostream& os, const char* name)
{
    os << data_file_header << name << ends;
}

/* Compute the checksum of a file by checkin it out into memory,
 * and calling md5_buffer. */
static PrVoidError compute_cksum(const char* versionfile, RcsVersionData* ver)
{
    FILE* cofile;
    char* md5_buf;

    Return_if_fail(cofile << VC_checkout_stream(ver->rcs_version(), versionfile));

    If_fail(read_buf.map_file(fileno(cofile)))
	throw prcserror << "Read failed from RCS co pipe" << dotendl;

    ver->length(read_buf.length());

    Return_if_fail(VC_close_checkout_stream(cofile, ver->rcs_version(), versionfile));

    md5_buf = md5_buffer(read_buf.segment(), read_buf.length());
    ver->unkeyed_checksum(md5_buf);

    return NoError;
}

/* Create a data file with no versions in it, for initializing a repository */
PrVoidError create_empty_data(const char* project, const char* filename0)
{
    ofstream new_data;

    new_data.open(filename0);

    write_data_file_header(new_data, project);

    if(new_data.bad()) {
	throw prcserror << "Failed creating new repository data file "
			 << squote(filename0) << perror;
    }

    return NoError;
}

/**********************************************************************/
/*                          Write Methods                             */
/**********************************************************************/

/* Write out the ProjectVersionData structure */
static void write_data_file_project_info(ostream& os, ProjectVersionData* ver)
{
    os << ver->date() << ends;
    os << ver->author() << ends;
    os << ver->rcs_version() << ends;

    os << (int)ver->deleted() << ends;

    os << ver->prcs_major() << ends
       << ver->prcs_minor() << ends
       << ver->parent_major() << ends
       << ver->parent_minor() << ends;

    if(ver->has_descends_from())
	os << 1 << ends
	   << ver->descends_from_major() << ends
	   << ver->descends_from_minor() << ends;
    else
	os << 0 << ends;

}

/* Write out the RcsVersionData structure */
static void write_data_file_rcs_info(ostream& os, RcsVersionData* ver)
{
    os << ver->date() << ends;
    os << ver->length() << ends;
    os << ver->author() << ends;
    os << ver->rcs_version() << ends;

    os.write(ver->unkeyed_checksum(), 16);
}

/* Write out a ProjectVersionData for each version in the project */
static void write_data_file_project_summary(ostream& os,
					    ProjectVersionDataPtrArray *project_summary)
{
    os << project_summary_header << project_summary->length() << ends;

    foreach(proj_ptr2, project_summary, ProjectVersionDataPtrArray::ArrayIterator) {
	write_data_file_project_info(os, *proj_ptr2);
    }
}

static void write_data_file_rcs_file_summary(ostream&os, const char* name, RcsDelta* delta)
{
    os << rcs_file_summary_header << name << ends;
    os << version_count_header << delta->count() << ends;

    foreach(ver, delta, RcsDelta::DeltaIterator) {
	write_data_file_rcs_info(os, *ver);
    }
}

/**********************************************************************/
/*                       Deletion Methods                             */
/**********************************************************************/

/* Delete a single version from a single file. */
/* Whether RCS really deletes the version or not is ignored */
static PrVoidError delete_version(const char* file_name, const char* version)
{
    if(!option_report_actions) {
	bool deleted;

	Return_if_fail(deleted << VC_delete_version(version, file_name));

	if(!deleted)
	    return NoError;
    }

    if(option_long_format) {
	prcsoutput << (option_report_actions ? "Delete" : "Deleted")
		<< " unreferenced version " << version
		<< " from version file " << squote(file_name) << dotendl;
    }

    return NoError;
}

/* Delete all unreferenced versions from an RCS file */
static PrVoidError delete_unreferenced_versions_file(const char* file_name,
						     RcsDelta* delta)
{
    foreach(data_ptr, delta, RcsDelta::DeltaIterator) {
	RcsVersionData* data = *data_ptr;

	if(data->referenced() == RcsVersionData::Referenced)
	    continue;

	delta->remove(data->rcs_version());

	Return_if_fail(delete_version(file_name, data->rcs_version()));
    }

    return NoError;
}

/* Mark all versions referenced by the a single version of a project
 * in the file table. */
static PrVoidError reference_versions(const char* project_file_name,
				      ProjectVersionData* project_data,
				      RcsFileTable* file_table)
{
    FILE* cofile;
    ProjectDescriptor* project;
    Dstring name;

    Return_if_fail(cofile << VC_checkout_stream(project_data->rcs_version(),
						project_file_name));

    name.sprintf("%s:%s", project_file_name, project_data->rcs_version());

    If_fail(project << read_project_file(cmd_root_project_name,
					 strip_leading_path(name),
					 cofile)) {
	prcswarning << "Invalid version file, not referencing any versions" << dotendl;
	return NoError;
    }

    Return_if_fail(VC_close_checkout_stream(cofile,
					    project_data->rcs_version(),
					    project_file_name));

    foreach_fileent(fe_ptr, project) {
	RcsDelta **lookup, *delta;
	RcsVersionData* version_data;
	FileEntry *fe = *fe_ptr;

	if(fe->file_type() != RealFile)
	    continue;

	if((lookup = file_table->lookup(*fe->descriptor_name())) == NULL) {
	    /*delete project;*/
	    throw prcserror << "Referenced version file "
			     << squote(fe->descriptor_name())
			     << " not found, cannot clean repository" << dotendl;
	}

	delta = *lookup;

	version_data = delta->lookup(*fe->descriptor_version_number());

	if(!version_data) {
	    /*delete project;*/
	    throw prcserror << "Referenced version " << fe->descriptor_version_number()
			     << " in repository file " << squote(fe->descriptor_name())
			     << " not found, cannot clean repository" << dotendl;
	}

	version_data->reference(RcsVersionData::Referenced);
    }

    delete project;

    return NoError;
}

/* For each file, call delete_unreferenced_versions_file */
static PrVoidError
delete_unreferenced_versions(RepEntry* rep_entry)
{
    ProjectVersionDataPtrArray* project_summary = rep_entry->project_summary();
    RcsFileTable* file_table = rep_entry->rcs_file_summary();

    foreach(project_data_ptr, project_summary,
	    ProjectVersionDataPtrArray::ArrayIterator) {

	if((*project_data_ptr)->deleted())
	    continue;

	Return_if_fail(reference_versions(rep_entry->Rep_name_of_version_file(),
					  *project_data_ptr,
					  file_table));
    }

    foreach(file_pair_ptr, file_table, RcsFileTable::HashIterator) {
	Return_if_fail(delete_unreferenced_versions_file(
	    rep_entry->Rep_name_in_entry((*file_pair_ptr).x()),
	    (*file_pair_ptr).y()));

	if((*file_pair_ptr).y()->count() == 0) {
	    /* Takes care of deleteing the empty file */
	    rebuild_repository_file(rep_entry->Rep_name_in_entry((*file_pair_ptr).x()));
	}
    }

    return NoError;
}


/**********************************************************************/
/*                          Gather Methods                            */
/**********************************************************************/

/* Read one RCS file's rlog, and return an RcsDelta table */
static PrRcsDeltaPtrError rebuild_repository_file(const char* name)
{
    RcsDelta* data;

    Return_if_fail(data << VC_get_version_data(name));

    if(data->count() == 0) {
	if(option_long_format) {
	    prcsoutput << (option_report_actions ? "Delete" : "Deleting")
		    << " empty version file " << squote(strip_leading_path(name))
		    << " from repository" << dotendl;
	}

	if(!option_report_actions) {
	    If_fail(Err_unlink(name))
		throw prcserror << "Unlink failed on "
				 << squote(name) << perror;
	}

	return (RcsDelta*)0;
    }

    foreach(rev_ptr, data, RcsDelta::DeltaIterator) {
	Return_if_fail(compute_cksum(name, *rev_ptr));
    }

    return data;
}

static PrRcsFileTablePtrError rebuild_file_table(RepEntry* rep_ptr)
{
    Dir dir(rep_ptr->Rep_entry_path());
    RcsDelta* delta_ptr;
    RcsFileTable* file_table = new RcsFileTable;

    foreach(ent_ptr, dir, Dir::FullDirIterator) {
	const char* ent = *ent_ptr;

	if(is_garbage_by_name(ent)) {

	    /* Delete it! */
	    if(!option_report_actions) {
		If_fail(fs_nuke_file(ent)) {
		    prcswarning << "Failed removing garbage file "
				<< squote(strip_leading_path(ent))
				<< " from repository" << dotendl;
		    continue;
		}
	    }

	    if(option_long_format) {
		prcsoutput << (option_report_actions ? "Delete" : "Deleting")
			<< " garbage file " << squote(strip_leading_path(ent))
			<< " from repository" << dotendl;
	    }

	} else if(!is_prcs_file(ent, rep_ptr->Rep_name_of_version_file())) {

	    /* Found an RCS file, run rebuild */

	    Return_if_fail(delta_ptr << rebuild_repository_file(ent));

	    if(!delta_ptr)
		continue;

	    file_table->insert(intern(strip_leading_path(ent)), delta_ptr);
	}
    }

    return file_table;
}


/* This is the top-level call to rebuilding a repository data file. */
static PrVoidError rebuild_repository(RepEntry *rep_ptr,
				      ProjectVersionDataPtrArray* project_summary,
				      RcsFileTable* file_table)
{
    if(!file_table || !project_summary) {
	Return_if_fail(project_summary <<
		       VC_get_project_version_data(rep_ptr->Rep_name_of_version_file()));

	Return_if_fail(file_table << rebuild_file_table(rep_ptr));
    }

    rep_ptr->Rep_set_umask();

    /* Unlink here in case we're mmaping the file that's about to get truncated. */
    If_fail(Err_unlink(rep_ptr->Rep_name_in_entry(prcs_data_file_name))) {
	throw prcserror << "Unlink failed on old repository data file "
			 << squote(rep_ptr->Rep_name_in_entry(prcs_data_file_name))
			 << perror;
    }

    rep_ptr->project_summary(project_summary);
    rep_ptr->rcs_file_summary(file_table);

    Return_if_fail(delete_unreferenced_versions(rep_ptr));

    ofstream new_data(rep_ptr->Rep_name_in_entry(prcs_data_file_name));

    write_data_file_header(new_data, cmd_root_project_name);

    write_data_file_project_summary(new_data, project_summary);

    foreach(file_pair_ptr, file_table, RcsFileTable::HashIterator) {
	write_data_file_rcs_file_summary(new_data,
					 (*file_pair_ptr).x(),
					 (*file_pair_ptr).y());
    }

    new_data.close();

    if(new_data.bad()) {
	throw prcserror << "Write failed to reconstructed project data file "
			 << squote(rep_ptr->Rep_name_in_entry(prcs_data_file_name))
			 << perror;
    }

    return NoError;
}

/* Run rebuild with an opened repository */
PrVoidError admin_rebuild_command_no_open(RepEntry* rep, bool valid_rep_data)
{
    Return_if_fail(rep->Rep_uncompress_all_files());

    if(valid_rep_data)
	Return_if_fail(rebuild_repository(rep,
					  rep->project_summary(),
					  rep->rcs_file_summary()));
    else
	Return_if_fail(rebuild_repository(rep, NULL, NULL));

    Return_if_fail(rep->Rep_compress_all_files());

    return NoError;
}

/* This is it. */
PrPrcsExitStatusError admin_rebuild_command()
{
    RepEntry *rep;

    Return_if_fail(rep << Rep_init_repository_entry(cmd_root_project_name,
						    true, false, false));

    prcsinfo << "This command may run for a long time" << dotendl;

    Return_if_fail(admin_rebuild_command_no_open(rep, false));

    return ExitSuccess;
}

/* Mark a project version as deleted by changing its log information */
static PrVoidError mark_deleted(ProjectVersionData* project_data, RepEntry* rep_entry)
{
    Dstring log;

    project_data->deleted(true);

    format_version_log(project_data, log);

    Return_if_fail(VC_set_log(log.cast(),
			      project_data->rcs_version(),
			      rep_entry->Rep_name_of_version_file()));

    return NoError;
}

/* This is it, too. */
PrPrcsExitStatusError delete_command()
{
    RepEntry *rep_entry;
    ProjectVersionData* project_data;

    Return_if_fail(rep_entry << Rep_init_repository_entry(cmd_root_project_name,
							  true, false, true));

    Return_if_fail(project_data << resolve_version(cmd_version_specifier_major,
						   cmd_version_specifier_minor,
						   cmd_root_project_full_name,
						   cmd_root_project_file_path,
						   NULL,
						   rep_entry));

    if(project_data->prcs_minor_int() == 0)
	throw prcserror << "You may not delete minor version 0" << dotendl;

    if(project_data->deleted())
	throw prcserror << "That version is already deleted" << dotendl;

    prcsquery << "Delete version "
	      << project_data->prcs_major()
	      << "." << project_data->prcs_minor()
	      << force("")
	      << report("")
	      << defopt('y', "You know what you're doing, go ahead")
	      << optfail('n')
	      << query(".  Are you sure");

    Return_if_fail(prcsquery.result());

    if(option_report_actions)
	return ExitSuccess;

    Return_if_fail(mark_deleted(project_data, rep_entry));

    Return_if_fail(admin_rebuild_command_no_open(rep_entry, true));

    return ExitSuccess;
}

/* READ */

/* Called when something goes wrong reading a data file */
PrVoidError RebuildFile::bad_data_file()
{
    throw prcserror << "Invalid repository data file "
		     << squote(source_name())
		     << ".  Please run " << squote("prcs admin rebuild")
		     << " to rebuild the data file"
		     << dotendl;
}

/* Read the header and assert that the projects match */
PrVoidError RebuildFile::read_header(const char* projname)
{
    if(!get_string(data_file_header))
	return bad_data_file();

    if(!get_string(projname, true))
	return bad_data_file();

    return NoError;
}

/* Read a single project version */
PrProjectVersionDataPtrError RebuildFile::read_project_version()
{
    ProjectVersionData* ver = new ProjectVersionData;
    int get;

    if((get = get_size()) < 0)
	return bad_data_file();

    ver->date(get);
    ver->author(get_string());
    ver->rcs_version(get_string());

    if(!(ver->author() && ver->rcs_version()))
	return bad_data_file();

    if((get = get_size()) < 0)
	return bad_data_file();

    ver->deleted(get);
    ver->prcs_major(get_string());
    ver->prcs_minor(get_string());
    ver->parent_major(get_string());
    ver->parent_minor(get_string());

    if(!(ver->prcs_major() && ver->prcs_minor() &&
	 ver->parent_major() && ver->parent_minor()))
	return bad_data_file();

    if((get = get_size()) < 0)
	return bad_data_file();
    else if(get > 0) {
	ver->descends_from_major(get_string());
	ver->descends_from_minor(get_string());
    }

#ifdef TEST_DB_READ
    cout << "version date: " << time_t_to_rfc822(ver->date()) << endl;
    cout << "version length: " << ver->length() << endl;
    cout << "version author: " << ver->author() << endl;
    cout << "rcs version: " << ver->rcs_version() << endl;
    cout << "deleted: " << ver->deleted() << endl;
    cout << "major: " << ver->prcs_major() << endl;
    cout << "minor: " << ver->prcs_minor() << endl;
    cout << "parent major: " << ver->parent_major() << endl;
    cout << "parent minor: " << ver->parent_minor() << endl;
    cout << "has descends: " << ver->has_descends_from() << endl;
    if(ver->has_descends_from()) {
	cout << "major descends: " << ver->descends_from_major() << endl;
	cout << "minor descends: " << ver->descends_from_minor() << endl;
    }
#endif

    return ver;
}

/* Read a single RCS version */
PrRcsVersionDataPtrError RebuildFile::read_rcs_version()
{
    RcsVersionData* ver = new RcsVersionData;
    const char* cksum;

    ver->date(get_size());
    ver->length(get_size());

    if(ver->date() < 0 || ver->length() < 0)
	return bad_data_file();

    ver->author(get_string());
    ver->rcs_version(get_string());

    if(!(ver->author() && ver->rcs_version()))
	return bad_data_file();

    if(!(cksum = get_string(16)))
	return bad_data_file();

    ver->unkeyed_checksum(cksum);

#ifdef TEST_DB_READ
    cout << "version date: " << time_t_to_rfc822(ver->date()) << endl;
    cout << "version length: " << ver->length() << endl;
    cout << "version author: " << ver->author() << endl;
    cout << "rcs version: " << ver->rcs_version() << endl;
    cout << "checksum: ";
    cout.write(ver->unkeyed_checksum(), 16);
    cout << endl;
#endif

    return ver;
}

/* Read some number of project versions */
PrVoidError RebuildFile::read_project_summary()
{
    int version_count;
    ProjectVersionData* ver;

    if((version_count = get_size()) < 0)
	return bad_data_file();

#ifdef TEST_DB_READ
    cout << "version count: " << version_count << endl;
#endif

    for(int i = 0; i < version_count; i += 1) {
#ifdef TEST_DB_READ
	cout << endl;
#endif
 	Return_if_fail(ver << read_project_version());
    	project_data->append(ver);
    }

    return NoError;
}

/* Read some number of RCS file summaries */
PrVoidError RebuildFile::read_rcs_file_summary()
{
    const char *read_file_name;
    RcsDelta **lookup, *delta_array;
    int num;

    if((read_file_name = get_string()) == NULL)
	return bad_data_file();

    if((lookup = rcs_file_table->lookup(read_file_name)) == NULL) {
	delta_array = new RcsDelta(NULL);
	rcs_file_table->insert(read_file_name, delta_array);
    } else {
	delta_array = *lookup;
    }

    if(!get_string(version_count_header))
	return bad_data_file();

    if((num = get_size()) < 0)
	return bad_data_file();

#ifdef TEST_DB_READ
    cout << endl;
    cout << "filename: " << read_file_name << endl;
    cout << "version count: " << num << endl;
#endif

    for(int i = 0; i < num; i += 1) {

#ifdef TEST_DB_READ
	cout << endl;
#endif
	RcsVersionData *ver;

	Return_if_fail(ver << read_rcs_version());

	delta_array->insert(ver->rcs_version(), ver);

	ASSERT(delta_array->lookup(ver->rcs_version()) == ver, "hope so");
    }

    return NoError;
}

/* the following RebuildFile methods deal with reading or expecting strings
 * and positive ints from a file. */
PrRebuildFilePtrError read_repository_data(const char* project, const char* filename0, bool write)
{
    RebuildFile* file = new RebuildFile;

    Return_if_fail(file->init_from_file(project, filename0, write));

    return file;
}

PrVoidError RebuildFile::init_from_file(const char* project, const char* filename0, bool write)
{
    filename.assign(filename0);

    seg = new MemorySegment(write);

    If_fail(seg->map_file(filename0))
	throw prcserror << "Failed opening project data file "
			 << squote(filename0) << perror;

    last = seg->segment() + seg->length();
    offset = 0;

    Return_if_fail(read_header(project));

    rcs_file_table = new RcsFileTable;

    project_data = new ProjectVersionDataPtrArray;

    const char* s;

    while((s = get_string(strlen(project_summary_header))) != NULL) {
	if(strncmp(s, project_summary_header, strlen(project_summary_header)) == 0)
	    Return_if_fail(read_project_summary());
	else if(strncmp(s, rcs_file_summary_header, strlen(rcs_file_summary_header)) == 0)
	    Return_if_fail(read_rcs_file_summary());
	else
	    return bad_data_file();
    }

    if(done())
	return NoError;
    else
	return bad_data_file();
}

const char* RebuildFile::get_string()
{
    const char* val = seg->segment() + offset;

    if(seg->length() < offset)
	return NULL;

    offset += strlen(val) + 1;

    return val;
}

const char* RebuildFile::get_string(int len)
{
    if(last - (seg->segment() + offset) < len || seg->length() < offset)
	return NULL;

    const char* val = seg->segment() + offset;

    offset += len;

    return val;
}

bool RebuildFile::get_string(const char* expected, bool term)
{
    int len = strlen(expected);

    if(last - (seg->segment() + offset) < len || seg->length() < offset)
	return NULL;

    if(strncmp(expected, seg->segment() + offset, len) != 0)
	return false;

    offset += len;
    if(term)
	offset += 1;

    return true;
}

int RebuildFile::get_size()
{
    const char* p = seg->segment() + offset;
    char c = 0;
    int val = 0;

    while(p < last && (c = *p++) != 0)  {
	if(c < '0' || c > '9')
	    return -1;
	val *= 10;
	val += (c - '0');
    }

    offset = p - seg->segment();

    if(p <= last && c == 0)
	return val;

    return -1;
}

const char* RebuildFile::source_name() { return filename; }

bool RebuildFile::done() { return offset == seg->length(); }

void RebuildFile::init_stream()
{
    if(!buf) {
	buf = new filebuf(seg->fd());
	buf->seekoff(0,ios::end);
	os = new ostream(buf);
    }
}

/* The following three methods append new data to the file */
void RebuildFile::add_project_data(ProjectVersionData* data)
{
    init_stream();

    (*os) << project_summary_header << 1 << ends;

    write_data_file_project_info(*os, data);
}

void RebuildFile::add_rcs_file_data(const char* new_file_name, RcsVersionData* data)
{
    init_stream();

    (*os) << rcs_file_summary_header << new_file_name << ends;
    (*os) << version_count_header << 1 << ends;

    write_data_file_rcs_info(*os, data);
}

PrVoidError RebuildFile::update_project_data()
{
    os->flush();

    if(os->fail())
	goto error;

    delete os;
    delete buf;

    os = NULL;
    buf = NULL;

    If_fail(seg->unmap())
	goto error;

    delete seg;
    seg = NULL;

    return NoError;

error:
    throw prcserror << "Failed updating repository data file"
		     << squote(source_name()) << perror;
}

RebuildFile::~RebuildFile()
{
    delete seg;

    foreach(data_ptr, project_data, ProjectVersionDataPtrArray::ArrayIterator)
	delete (*data_ptr);

    delete project_data;

    foreach(file_ptr, rcs_file_table, RcsFileTable::HashIterator) {
	delete (*file_ptr).y();
    }

    delete rcs_file_table;
}

#ifdef TEST_DB_READ
int main(int /*argc*/, const char** argv)
{
    RebuildFile rebuild_file;
    ProjectVersionDataPtrArray* project_data_array;
    RcsFileTable* rcs_file_table;

    setup_streams(argv[0]);

    If_fail(rebuild_file.open(argv[1]))
	exit(1);

    If_fail(read_header("prcs", rebuild_file))
	exit(1);

    If_fail(project_data_array << read_project_summary(rebuild_file))
	exit(1);

    If_fail(rcs_file_table << read_rcs_file_summary(rebuild_file))
	exit(1);

    return 0;
}
#endif
