/* -*-Mode: C++;-*-
 * $Id: repository.cc 1.22.1.11.1.28.1.7 Tue, 03 Sep 1996 16:01:28 -0700 jmacd $
 * repository.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */

extern "C" {
#include <sys/types.h>
#include <pwd.h>
#include <grp.h>
}

#include "prcs.h"

#include "hash.h"
#include "prcsdir.h"
#include "repository.h"
#include "dstring.h"
#include "vc.h"
#include "checkout.h"
#include "misc.h"
#include "lock.h"
#include "syscmd.h"
#include "system.h"
#include "convert.h"
#include "memseg.h"
#include "rebuild.h"
#include "projdesc.h"

static char const prcs_tagname[] = "prcs_tag";
static char const prcs_compression_tag[] = "prcs_compress";
static char const prcs_lock_dir[] = ".locks";

static int rep_root_path_len;
static Dstring *rep_root_path = NULL;

static char const prcs_created_by[] = "created by PRCS version ";
static char const prcs_usable_by[] = "usable down to PRCS version ";

extern PrVoidError save_package();

static PrVoidError Rep_init_repository();
static PrVoidError Rep_init_command(const char* path);
static const char* Rep_name_in_repository(const char *s);
static bool version_less_than(const int* a, const int* b);

static PrVoidError repository_cleanup_handler(void* data, bool);

/* The lowest version which may use a repository last touched by this
 * version of PRCS */
static int const prcs_usable_number[3] = { 0, 13, 5 };

/* The lowest version which this version of PRCS may use without being
 * converted.  */
static int const prcs_may_use_number[3] = { 0, 13, 5 };

/* Rep_guess_repository_path --
 *
 *     returns the calculated repository path from environment
 *     variables or command line options, without actually
 *     initilaizing anything.  */
NprConstCharPtrError Rep_guess_repository_path()
{
    if(rep_root_path) {
	return Rep_name_in_repository("");
    }

    const char* rp;
    rep_root_path = new Dstring;

    if(cmd_repository_path)
	rp = cmd_repository_path;
    else
	rp = getenv("PRCS_REPOSITORY");

    if ( rp == NULL ) {
	rp = getenv("HOME");
	if (rp == NULL)
	    return NonFatalError;

	rep_root_path->assign(rp);
	rep_root_path->append("/PRCS");
    } else {
	rep_root_path->assign(rp);
    }

    while(rep_root_path->index(rep_root_path->length() - 1) == '/')
	rep_root_path->truncate(rep_root_path->length() - 1);

    rep_root_path_len = rep_root_path->length();

    return Rep_name_in_repository("");
}

/* Rep_init_repository --
 *
 *      Must be called before any of the below commands are used.
 *      This places a read or write lock in the repository, reads the
 *      repository tag, insures that this version of PRCS is
 *      compatible with the repository format being opened.  It will
 *      prompt the user to upgrade repository versions if the version
 *      is outdated and return NoError only when the repository is
 *      locked and ready to use by the current version of PRCS.  */
static PrVoidError Rep_init_repository()
{
    static bool do_once = true;

    if(do_once) {
	do_once = false;
    } else {
	return NoError;
    }

    If_fail(Rep_guess_repository_path())
	throw prcserror << "Please set your PRCS_REPOSITORY environment variable" << dotendl;

    Return_if_fail(Rep_init_command(rep_root_path->cast()));

    return NoError;
}

static PrVoidError Rep_init_command(const char* path)
{
    if (!fs_file_exists(path) ) {
	if (option_report_actions)
	    throw prcsinfo << "Create repository " << squote(path) << dotendl;

	If_fail( Err_mkdir(path, 0777 | S_ISGID | 01000))
	    throw prcserror << "Cannot create repository " << squote(path) << perror;
	else
	    prcswarning << "Created repository " << squote(path)
			<< ", you may wish to run " << squote("prcs admin access")
			<< " to set its permissions" << dotendl;
    }

    const char* rep_lock_name = Rep_name_in_repository(prcs_lock_dir);

    if(!fs_is_directory(rep_lock_name)) {
        if (option_report_actions)
	    throw prcserror << "Create repository lock directory "
			     << squote(rep_lock_name) << perror;

	If_fail( Err_mkdir(rep_lock_name, 0777) )
	    throw prcserror << "Cannot create repository lock directory "
			     << squote(rep_lock_name) << perror;
    }

    if(!fs_file_writeable(rep_lock_name))
       throw prcserror << "You must have write permission on the lock directory "
			<< squote(rep_lock_name) << perror;

    return NoError;
}

static const char* Rep_name_in_repository(const char *s)
{
    rep_root_path->truncate(rep_root_path_len);
    rep_root_path->append('/');
    rep_root_path->append(s);

    return rep_root_path->cast();
}

PrVoidError RepEntry::Rep_lock(bool write_lock)
{
    int lockval;
    Dstring lock_name;

    lock_name.assign(Rep_name_in_repository(prcs_lock_dir));
    lock_name.append('/');
    lock_name.append(entry_name);

    _repository_lock = new AdvisoryLock(lock_name);

    if(write_lock)
	lockval = _repository_lock->write_lock_nowait();
    else
	lockval = _repository_lock->read_lock_nowait();

    while(true) {
	if(lockval == LOCK_FAIL)
	    throw prcserror << "Failed obtaining a repository lock" << dotendl;

	if(lockval == LOCK_SUCCESS)
	    return NoError;

	char buf[256];
	int lock_count = 0;
	FILE* other_locks;

	other_locks = _repository_lock->who_is_locking();

	if(!other_locks)
	    throw prcserror << "Could not obtain a list of lockers, unknown "
		"locking error" << dotendl;

	prcsinfo << "Could not obtain a lock because of the following locks:" << prcsendl;

	while(fgets(buf, 256, other_locks) == buf) {
	    if(!isspace(buf[0])) {
		cout << buf;
		lock_count += 1;
	    }
	}

	fclose(other_locks);

	if(lock_count == 0)
	    throw prcserror << "Inconsistant lock information, cannot continue" << dotendl;

	if(lockval == LOCK_TRY_STEAL) {
	    prcsquery << "Lock age exceeds timeout age.  "
		      << force("Stealing")
		      << report("Steal")
		      << optfail('n')
		      << defopt('y', "Steal the lock and continue")
		      << query("Steal the lock");

	    Return_if_fail(prcsquery.result());

	    if(!_repository_lock->steal_lock())
		return PrVoidError(FatalError);

	} else {
	    prcsquery << force("Waiting for the lock")
		      << report("Waiting for the lock")
		      << optfail('n')
		      << defopt('y', "Wait for the lock")
		      << query("Wait for the lock");

	    Return_if_fail(prcsquery.result());
	}

	if(write_lock)
	    lockval = _repository_lock->write_lock_wait();
	else
	    lockval = _repository_lock->read_lock_wait();
    }
}

void RepEntry::Rep_unlock_repository()
{
    if(rebuild_file)
	delete rebuild_file;

    rebuild_file = NULL;

    if(_repository_lock) {
	_repository_lock->unlock();

	delete _repository_lock;

	_repository_lock = NULL;
    }
}

static bool version_less_than(const int* a, const int* b)
{
    for(int i = 0; i < 3;  i += 1) {
	if(a[i] < b[i])
	    return true;
	if(b[i] < a[i])
	    return false;
    }

    return false;
}

PrVoidError RepEntry::Rep_make_tag()
{
    const char* file_name = Rep_name_in_entry(prcs_tagname);

    unlink(file_name);

    ofstream tag_file(file_name, ios::out);

    tag_file << prcs_created_by
	     << prcs_version_number[0] << ' '
	     << prcs_version_number[1] << ' '
	     << prcs_version_number[2]
	     << endl
	     << prcs_usable_by
	     << prcs_usable_number[0] << ' '
	     << prcs_usable_number[1] << ' '
	     << prcs_usable_number[2]
	     << endl;

    tag_file.close();

    if(tag_file.bad()) {
	throw prcserror << "Failed writing repository tagfile "
			 << squote(file_name) << perror;
    }

    If_fail(Err_chmod(file_name, 0444)) {
 	throw prcserror << "Chmod failed on new repository tagfile "
			 << squote(file_name) << perror;
    }

    return NoError;
}

PrVoidError RepEntry::Rep_verify_tag(UpgradeRepository* upgrades)
{
    const char* file_name = Rep_name_in_entry(prcs_tagname);
    char buf[256];
    int created[3];
    int usable[3];

    if(!fs_file_exists(file_name)) {
	throw prcserror << "Repository is not tagged, the file "
			 << squote(file_name) << " does not exist" << dotendl;
    }

    if(!fs_file_readable(file_name)) {
	throw prcserror << "Can't read repository tag " << squote(file_name) << dotendl;
    }

    ifstream tag_file(file_name);

    if(tag_file.bad()) {
	throw prcserror << "Can't open repository tag file " << squote(file_name) << dotendl;
    }

    tag_file.get(buf, sizeof(prcs_created_by));
    tag_file >> created[0];
    tag_file >> created[1];
    tag_file >> created[2];

    if(strncmp(buf, prcs_created_by, strlen(prcs_created_by)) != 0 || tag_file.bad()) {
	throw prcserror << "Repository tag file "
			 << squote(file_name) << " is invalid" << dotendl;
    }

    tag_file.get();
    tag_file.get(buf, sizeof(prcs_usable_by));
    tag_file >> usable[0];
    tag_file >> usable[1];
    tag_file >> usable[2];

    if(strncmp(buf, prcs_usable_by, strlen(prcs_usable_by)) != 0 || tag_file.bad()) {
	throw prcserror << "Repository tag file " << squote(file_name) << " is invalid" << dotendl;
    }

    bool confirm_upgrade = false;
    while(upgrades && upgrades->upgradeFunc) {
	if(version_less_than(created, upgrades->version)) {
	    if(!writeable) {
		throw prcserror << "This version of PRCS needs to modify your repository before "
		    "continuing.  You do not currently have a write lock, required to do so.  "
		    "You can run " << squote("prcs admin rebuild") << " to upgrade your repository" << dotendl;
	    }

	    if(!confirm_upgrade) {
		prcsquery << "This version of PRCS needs to modify your repository.  "
			  << force("Upgrading")
			  << report("Upgrade")
			  << defopt('s', "Upgrade the repository, after saving a package in case of trouble")
			  << optfail('n')
			  << option('y', "Upgrade the repository, don't save a package")
			  << query("Upgrade");

		char c;

		Return_if_fail(c << prcsquery.result());

		confirm_upgrade = true;

		if(c == 's') {
		    Return_if_fail(save_package());
		}
	    }

	    prcsinfo << "Upgrading repository to version format "
		    << upgrades->version[0] << '.'
		    << upgrades->version[1] << '.'
		    << upgrades->version[2] << dotendl;
	    prcsinfo << upgrades->description << dotendl;

	    created[0] = usable[0] = upgrades->version[0];
	    created[1] = usable[1] = upgrades->version[1];
	    created[2] = usable[2] = upgrades->version[2];

	    bool should_exit;

	    Return_if_fail(should_exit << upgrades->upgradeFunc(this));
	    Return_if_fail(Rep_make_tag());

	    prcsinfo << "Finished upgrade to version format "
 		    << upgrades->version[0] << '.'
		    << upgrades->version[1] << '.'
		    << upgrades->version[2] << dotendl;

	    if(should_exit) {
		return PrVoidError(UserAbort);
	    }
	}

	upgrades += 1;
    }

    /* If the tag says an older version which this version is unable
     * to use, and the upgrade did not help, tell the user the
     * repository is too old to use.  */
    if(version_less_than(created, prcs_may_use_number)) {
	throw prcserror << "This version of PRCS outdates the version which created your repository, "
			 << created[0] << '.'
			 << created[1] << '.'
			 << created[2]
			 << ".  Please convert your repository" << dotendl;
    }

    /* If the tag says a newer version which this version is unable to
     * use, print a message saying to upgrade the binary to the newer
     * PRCS.  */
    if(version_less_than(prcs_version_number, usable)) {
	throw prcserror << "This version of PRCS is outdated by version "
			 << created[0] << '.'
			 << created[1] << '.'
			 << created[2] << ".  You must upgrade before using this repository"
			 << dotendl;
    }

    /* If the tag has a newer version number than this version which
     * is still usable, notify the user that a newer version has
     * touched the repository and continue.  */
    if(version_less_than(prcs_version_number, created))
	prcswarning << "This version of PRCS is outdated by version "
		    << created[0] << '.'
		    << created[1] << '.'
		    << created[2]
		    << ".  It is still compatible, but you should consider upgrading" << dotendl;

    /* If we have a write lock, modify the tag with the up to date
     * versions */
    if(version_less_than(created, prcs_version_number) && writeable) {
	Return_if_fail(Rep_make_tag());
    }

    return NoError;
}

PrVoidError RepEntry::Rep_create_repository_entry()
{
    const char* file_name = Rep_name_in_entry("");

    if ( option_report_actions ) {
	throw prcserror << "Create repository entry " << squote(entry_name);
    } else {
	If_fail( Err_mkdir(file_name, 0777) ) {
	    throw prcserror << "Cannot create repository entry " << squote(entry_name) << perror;
	}

	Return_if_fail(Rep_make_tag());

	VC_register(project_file);

	If_fail(VC_register(NULL)) {
	    fs_nuke_file(file_name);
	    throw prcserror << "Failed registering project version file"
			    << squote(project_file) << dotendl;
	}

	Return_if_fail(create_empty_data(entry_name, Rep_name_in_entry("prcs_data")));

	Return_if_fail(Rep_chmod(0777 ^ get_umask()));

	prcsinfo << "Created repository entry " << squote(entry_name) << dotendl;
    }

    return NoError;
}

PrBoolError Rep_repository_entry_exists(const char* entry)
{
    Return_if_fail(Rep_init_repository());

    return fs_is_directory(Rep_name_in_repository(entry));
}

/* The repsository entry class */

/* Rep_init_repository_entry --
 *
 *     assures that the named entry exists, is executable, and if
 *     writeable is true, also writeable.  Also checks the repository
 *     tag created by Rep_create_repository_entry to see that this
 *     entry has not been used by some future version of PRCS making
 *     it unusable by the version running */
PrRepEntryPtrError Rep_init_repository_entry(const char* entry, bool writeable,
					     bool create, bool require_db)
{
    RepEntry* ret = new RepEntry();

    Return_if_fail(ret->init(entry, writeable, create, require_db));

    return ret;
}

PrVoidError RepEntry::init(const char* entry, bool writeable0, bool create, bool require_db)
{
    Return_if_fail(Rep_init_repository());

    project_file.sprintf("%s/%s/%s.prj,v", Rep_name_in_repository(""), entry, entry);

    entry_name.assign(entry);

    writeable = writeable0;

    base.assign(Rep_name_in_repository(entry));

    baselen = base.length();

    bool exists;

    Return_if_fail(exists << Rep_repository_entry_exists(entry));

    compressed = false;
    rebuild_file = NULL;

    if(!exists) {
	if(create) {
	    Return_if_fail(Rep_create_repository_entry());
	} else {
	    throw prcserror << "Repository entry " << squote(entry)
			     << " does not exist, you can create an entry with "
			     << squote("prcs checkin")
			     << dotendl;
	}
    }

    If_fail(Err_stat(Rep_entry_path(), &rep_stat_buf))
	throw prcserror << "Stat failed on repository entry "
			 << squote(Rep_entry_path()) << perror;

    Return_if_fail(Rep_lock(writeable0));

    install_cleanup_handler(repository_cleanup_handler, this);

    if(!fs_file_exists(project_file)) {
	throw prcserror << "Repository entry " << squote(entry)
			 << " is missing its project version file" << dotendl;
    }

    if(writeable && !(fs_file_rwx(Rep_entry_path()))) {
	throw prcserror << "You need read, write, and execute permission on the repository entry "
			 << squote(Rep_entry_path()) << dotendl;
    } else if(!fs_file_executable(Rep_entry_path())) {
	throw prcserror << "You need execute permission on the repository entry "
			 << squote(Rep_entry_path()) << dotendl;
    }

    if(require_db) {
	if(!fs_file_exists(Rep_name_in_entry(prcs_data_file_name))) {
	    throw prcserror << "Repository entry " << squote(entry)
			    << " is missing its data file, please run "
			    << squote("prcs admin rebuild")
			    << " to generate this file" << dotendl;
	}

	Return_if_fail(rebuild_file <<
		       read_repository_data(entry, Rep_name_in_entry (prcs_data_file_name),
					    writeable0));

	project_data_array = rebuild_file->get_project_summary();

	rcs_file_table = rebuild_file->get_rcs_file_summary();
    }

    rep_mask = rep_stat_buf.st_mode ^ 0777;
    user_mask = get_umask();

    compressed = fs_file_exists(Rep_name_in_entry(prcs_compression_tag));

    Return_if_fail(Rep_verify_tag(entry_upgrades));

    return NoError;
}

PrVoidError RepEntry::Rep_uncompress_all_files()
{
    ArgList* args;

    ASSERT(writeable, "you can't uncompress without a write lock");

    Rep_set_umask();

    bool once = true;

    Dir current_dir(Rep_entry_path());

    foreach(ent_ptr, current_dir, Dir::DirIterator) {
	const char* ent = *ent_ptr;

	int len = strlen(ent);
	if(len > 4 && strcmp(".gz", ent + len - 3) == 0) {
	    if(once) {
		prcsinfo << "Uncompressing repository" << dotendl;
		once = false;
	    }

	    Return_if_fail(args << gzip_command.new_arg_list());

	    args->append("-d");
	    args->append("-f");
	    args->append(Rep_name_in_entry(ent));

	    Return_if_fail_if_ne(gzip_command.open_stdout(), 0) {
		throw prcserror << "Gzip returned non-zero status, aborting" << dotendl;
	    }
	}
    }

    if (! current_dir.OK() ) {
	throw prcserror << "Error reading repository directory" << perror;
    }

    Rep_unset_umask();

    return NoError;
}

PrVoidError RepEntry::Rep_compress_all_files()
{
    ArgList* args;

    if(!Rep_needs_compress()) return NoError;

    ASSERT(writeable, "you need a write lock");

    Rep_set_umask();

    bool once = true;
    Dir current_dir(Rep_entry_path());

    const char* project_file_base = strip_leading_path (project_file);

    foreach(ent_ptr, current_dir, Dir::DirIterator) {
	const char* ent = *ent_ptr;

	int len = strlen(ent);
	if(len > 3 && strcmp(",v", ent + len - 2) == 0 && strcmp(ent, project_file_base) != 0) {

	    if(once) {
		prcsinfo << "Compressing repository" << dotendl;
		once = false;
	    }

	    Return_if_fail(args << gzip_command.new_arg_list());

	    args->append("-f");
	    args->append(Rep_name_in_entry(ent));

	    Return_if_fail_if_ne(gzip_command.open_stdout(), 0) {
		throw prcserror << "Gzip returned non-zero status, aborting" << dotendl;
	    }
	}
    }

    if (! current_dir.OK() ) {
	throw prcserror << "Error reading repository directory" << perror;
    }

    Rep_unset_umask();

    return NoError;
}

/*
 * Rep_prepare_file --
 *
 *     this procedure has to copy the file to a temp location in order
 *     to avoid writing the repository if the file is compressed and
 *     the repository is only read-only.  */
PrConstCharPtrError RepEntry::Rep_prepare_file(const char* file)
{
    static Dstring ret;
    Dstring fullname, zipname;
    ArgList *args;

    ASSERT(strlen(file) > 0, "no empty files");

    Rep_name_in_entry(file, &fullname);
    Rep_name_in_entry(file, &zipname);
    zipname.append(".gz");

    if(fs_file_exists(zipname)) {
	Return_if_fail(args << gzip_command.new_arg_list());

	if(writeable) {
	    args->append("-f");
	    args->append("-d");
	    args->append(zipname);

	    Return_if_fail_if_ne(gzip_command.open_stdout(), 0)
		throw prcserror << "Gzip returned non-zero status, aborting" << dotendl;
	} else {
	    const char* temp = make_temp_file(",v");

	    args->append("-f");
	    args->append("-d");
	    args->append("-c");
	    args->append(zipname);

	    Return_if_fail(gzip_command.open(true, false));

	    Return_if_fail(fs_write_filename(gzip_command.standard_out(), temp));

	    Return_if_fail_if_ne(gzip_command.close(), 0)
		throw prcserror << "Gzip returned non-zero status, aborting" << dotendl;

	    ret.assign(temp);

	    return ret.cast();
	}
    }

    if(!fs_file_exists(fullname))
	throw prcserror << "Repository file " << squote(fullname)
			 << " not found--try running " << squote("prcs admin rebuild")
			 << " for a complete diagnosis" << dotendl;

    ret.assign(fullname);

    return ret.cast();
}

bool RepEntry::Rep_file_exists(const char* name)
{
    Dstring d;
    Rep_name_in_entry(name, &d);
    if(fs_file_exists(d)) return true;
    d.append(".gz");
    if(fs_file_exists(d)) return true;
    return false;
}

const char* RepEntry::Rep_name_in_entry(const char* name)
{
    base.truncate(baselen);
    base.append('/');
    base.append(name);
    return base;
}

void RepEntry::Rep_name_in_entry(const char* name, Dstring * buf)
{
    base.truncate(baselen);
    buf->assign(base);
    buf->append('/');
    buf->append(name);
}

static PrVoidError copy_change_ownership(const char* name)
{
    int len = strlen(name);
    char* temp = new char[len + 8];

    strcpy(temp, name);
    strcat(temp, ".XXXXXX");

    if(!mktemp(temp))
	throw prcserror << "Mktemp failed on " << squote(name) << perror;

    Return_if_fail(fs_copy_filename(name, temp));

    If_fail(Err_rename(temp, name)) {
	throw prcserror << "Rename " << squote(temp) << " to " << squote(name)
			 << " failed while changing group -- correct this manually!" << perror;
    }

    delete [] temp;

    return NoError;
}

static PrVoidError set_grid(const char* name, gid_t gid, struct stat *buf)
{
    uid_t uid = getuid();

    if(buf->st_uid != uid) {
	Return_if_fail(copy_change_ownership(name));
    } else if(chown(name, uid, gid) < 0) {
	throw prcserror << "Chown failed on " << squote(name) << perror;
    }

    return NoError;
}

PrVoidError RepEntry::Rep_chgrp(const char* name)
{
    /* This is sort of weird, to change the group ownership of a file
     * that's not your own, you have to copy it and remove the old
     * file.  Thus, for each file in the repository thats not your own
     * copy it and unlink the old one, then set the new group, if
     * it needs to be set.  We assume of course, that we have write
     * permission on the directory.  This can only be called by the
     * owner of the directory, too. */
    struct group* grp;

    grp = getgrnam(name);

    if (!grp)
        throw prcserror << "Cannot lookup group " << squote(name) << perror;

    gid_t gid = grp->gr_gid;

    Dir current_dir(Rep_entry_path());
    uid_t uid = getuid();

    if( chown(Rep_name_in_repository(entry_name), uid, gid) < 0) {
	throw prcserror << "Chown failed on " << squote(Rep_name_in_repository(entry_name))
			 << perror;
    }

    foreach(ent_ptr, current_dir, Dir::DirIterator) {
	const char* name = Rep_name_in_entry(*ent_ptr);
	int len = strlen(name);
	struct stat buf;

	if(!(strcmp(name + len - 2, ",v") == 0 ||
	     strcmp(name + len - 5, ",v.gz") == 0 ||
	     strncmp(strip_leading_path(name), "prcs_", 5) == 0))
	    continue;

	If_fail(Err_stat(name, &buf)) {
	    throw prcserror << "Stat failed on " << squote(name) << perror;
	}

	if(buf.st_uid != uid || buf.st_gid != gid)
	    Return_if_fail(set_grid(name, gid, &buf));
    }

    if ( !current_dir.OK() ) {
	throw prcserror << "Error reading repository directory" << perror;
    }

    return NoError;
}

PrVoidError RepEntry::Rep_chmod(mode_t mode)
{
#define DO_CHMOD_REP(f, m) do { If_fail(Err_chmod(f, m)) { throw prcserror << "Chmod failed on repository entry " << squote(f) << perror; } } while (false)

    DO_CHMOD_REP(Rep_name_in_entry(prcs_tagname), 0444 & mode);

    if(compressed)
	DO_CHMOD_REP(Rep_name_in_entry(prcs_compression_tag), 0444 & mode);

    DO_CHMOD_REP(Rep_name_in_entry(prcs_data_file_name), 0666 & mode);

    DO_CHMOD_REP(Rep_name_in_repository(entry_name), mode | S_ISGID);

    Dir current_dir(Rep_entry_path());

    foreach(ent_ptr, current_dir, Dir::DirIterator) {

	const char* name = Rep_name_in_entry(*ent_ptr);
	int len = strlen(name);

	if(!(strcmp(name + len - 2, ",v") == 0 ||
	     strcmp(name + len - 5, ",v.gz") == 0))
	    continue;

	DO_CHMOD_REP(name, 0444 & mode);
    }

    if ( !current_dir.OK() )
	throw prcserror << "Error reading repository directory" << perror;

#undef DO_CHMOD_REP

    return NoError;
}

void RepEntry::Rep_new_filename(const char* filename, bool regist, Dstring* buf)
{
    static PathTable *table = NULL;

    if (!filename) {
	if (table) {
	    foreach(copy_ptr, table, PathTable::HashIterator)
		delete (*copy_ptr).x();

	    delete table;
	}

	return;
    }

    int i = 0;

    Dstring safe_file_name(filename);

    if(safe_file_name.length() > 10)
	safe_file_name.truncate(10);
    else
	while(safe_file_name.length() < 2)
	    safe_file_name.append('x');

    if(safe_file_name.length() > 1 &&
       strcmp(filename + safe_file_name.length() - 2, ",v") == 0)
	safe_file_name.append('x');

    if(safe_file_name.length() > 2 &&
       strcmp(filename + safe_file_name.length() - 3, ".gz") == 0)
	safe_file_name.append('x');

    if (!table) {
	table = new PathTable(pathname_hash, pathname_equal);

	Dir rep_dir(Rep_entry_path());

	foreach(ent_ptr, rep_dir, Dir::DirIterator)
	    table->insert(intern(*ent_ptr), RealFile);

	/* Warning: failure reading the directory will be ignored. */
    }

    do {
	buf->sprintf("%d_%s,v", i++, safe_file_name.cast());
    } while (table->isdefined(buf->cast()));

    /* These and the above leak. */
    table->insert(intern(buf->cast()), RealFile);

    if(regist)
	VC_register(Rep_name_in_entry(*buf));
}

ProjectVersionData* RepEntry::lookup_version(const char* major, const char* minor) const
{
    int minorint = atoi(minor);

    for(int i = 0; i < version_count(); i += 1) {
	if (minorint == project_data_array->index(i)->prcs_minor_int() &&
	    strcmp(project_data_array->index(i)->prcs_major(), major) == 0) {
	    return project_data_array->index(i);
	}
    }

    return NULL;
}

int RepEntry::highest_major_version() const /* -1 if none */
{
    int max = -1;

    for(int i = 0; i < version_count(); i += 1) {
	if (VC_check_token_match(project_data_array->index(i)->prcs_major(), "number")) {
	    int val = atoi(project_data_array->index(i)->prcs_major());

	    if(val > max)
		max = val;
	}
    }

    return max;
}

ProjectVersionData* RepEntry::highest_minor_version_data(const char* major) const
{
    int max = 0;
    ProjectVersionData* data = NULL;

    for(int i = 0; i < version_count(); i += 1) {
	if (strcmp(project_data_array->index(i)->prcs_major(), major) == 0) {
	    int val = project_data_array->index(i)->prcs_minor_int();

	    if(val > max) {
		data = project_data_array->index(i);
		max = val;
	    }
	}
    }

    return data;
}


int RepEntry::highest_minor_version(const char* major) const /* 0 if none */
{
    ProjectVersionData* data = highest_minor_version_data(major);

    if(!data)
	return 0;

    return data->prcs_minor_int();
}

bool RepEntry::major_version_exists(const char* major) const
{
    for(int i = 0; i < version_count(); i += 1) {
	if (strcmp(project_data_array->index(i)->prcs_major(), major) == 0) {
	    return true;
	}
    }

    return false;
}

ProjectVersionData* RepEntry::most_distant_ancestor(const char* major1,
						    int minor1,
						    const char* major2,
						    int minor2) const
    /* This is sort of inefficent, as is common_version(), with
     * which is shares enough code to be worth merging, but
     * its not much of a hit and only occurs in checkin, which
     * is overwhealmed by read() costs, so its not very important
     * to optimized. */
{
    /* Versions are ordered in decreasing date order */

    int length = project_data_array->length();
    int index1 = -1, index2 = -1, i, first_index1;
    ProjectVersionData *common_version = NULL;

    for(i = 0; i < length; i += 1) {
	if(minor1 == project_data_array->index(i)->prcs_minor_int() &&
	   strcmp(major1, project_data_array->index(i)->prcs_major()) == 0) {
	    index1 = i;
	    break;
	}
    }

    first_index1 = index1;

    for(i = 0; i < length; i += 1) {
	if(minor2 ==  project_data_array->index(i)->prcs_minor_int() &&
	   strcmp(major2, project_data_array->index(i)->prcs_major()) == 0) {
	    index2 = i;
	    break;
	}
    }

    if(index1 < 0 || index2 < 0)
	/**** This should be an error ****/
	return NULL;

    while(index1 >= 0 && index2 >= 0 && common_version == NULL) {
	if(index1 == index2) {
	    common_version = project_data_array->index(index1);
	} else if(index1 > index2) {
	    index1 = parent_index(index1);
	} else if(index2 > 0) {
	    index2 = parent_index(index2);
	} else {
	    break;
	}
    }

    if(index1 < 0 || index2 < 0)
	/**** This should be an error ****/
	return NULL;

    while(first_index1 != index1) {
	int next = parent_index(first_index1);

	if (strcmp(project_data_array->index(next)->prcs_major(),
		   project_data_array->index(first_index1)->prcs_major()) != 0)
	    return project_data_array->index(first_index1);

	first_index1 = next;
    }

    return NULL;
}


ProjectVersionData* RepEntry::common_version(const char* major1,
					     int minor1,
					     const char* major2,
					     int minor2) const
{
    /* Versions are ordered in decreasing date order */

    int length = project_data_array->length();
    int index1 = -1, index2 = -1, i;
    ProjectVersionData *common_version = NULL;

    for(i = 0; i < length; i += 1) {
	if(minor1 == project_data_array->index(i)->prcs_minor_int() &&
	   strcmp(major1, project_data_array->index(i)->prcs_major()) == 0) {
	    index1 = i;
	    break;
	}
    }

    for(i = 0; i < length; i += 1) {
	if(minor2 ==  project_data_array->index(i)->prcs_minor_int() &&
	   strcmp(major2, project_data_array->index(i)->prcs_major()) == 0) {
	    index2 = i;
	    break;
	}
    }

    if(index1 < 0 || index2 < 0)
	/**** This should be an error ****/
	return NULL;

    while(index1 >= 0 && index2 >= 0 && common_version == NULL) {
	if(index1 == index2) {
	    common_version = project_data_array->index(index1);
	} else if(index1 > index2) {
	    index1 = parent_index(index1);
	} else if(index2 > 0) {
	    index2 = parent_index(index2);
	} else {
	    break;
	}
    }

    if(index1 < 0 || index2 < 0)
	/**** This should be an error ****/
	return NULL;

    return common_version;
}

int RepEntry::parent_index(int index0) const
{
    /* Versions are ordered in increasing date order */

    const char *major_version = project_data_array->index(index0)->parent_major();
    int minor_version = project_data_array->index(index0)->parent_minor_int();
    int i = project_data_array->length() - 1;

    while(i >= 0) {
	if(minor_version == project_data_array->index(i)->prcs_minor_int() &&
	   strcmp(major_version, project_data_array->index(i)->prcs_major()) == 0) {
	    break;
	}
	i -= 1;
    }

    return i;
}

PrRcsVersionDataPtrError RepEntry::lookup_rcs_file_data(const char* filename, const char* version_num) const
{
    RcsDelta **lookup;

    if((lookup = rcs_file_table->lookup(filename)) != NULL) {
	RcsVersionData* data = (*lookup)->lookup(version_num);

	if(data)
	    return data;
    }

    throw prcserror << "Repository inconsistency detected in repository file "
		     << squote(filename) << " at file version " << version_num
		     << ", please run " << squote("prcs admin rebuild")
		     << " to diagnose this error" << dotendl;
}

void RepEntry::add_project_data(ProjectVersionData* new_data)
{
    if(!option_report_actions)
	rebuild_file->add_project_data(new_data);
}

void RepEntry::add_rcs_file_data(const char* filename, RcsVersionData* new_data)
{
    if(!option_report_actions)
	rebuild_file->add_rcs_file_data(filename, new_data);
}

PrVoidError RepEntry::close_repository()
{
    if(!option_report_actions) {
	Return_if_fail(rebuild_file->update_project_data());

	delete rebuild_file;

	rebuild_file = NULL;
    }

    return NoError;
}

PrProjectDescriptorPtrError RepEntry::checkout_prj_file(const char* fullname,
							const char* version)
{
    FILE* cofile;
    ProjectDescriptor* project;
    Dstring filename;

    filename.sprintf("%s:%s", Rep_name_of_version_file(), version);

    Return_if_fail(cofile << VC_checkout_stream(version, Rep_name_of_version_file()));

    Return_if_fail(project << read_project_file(fullname, filename, cofile));

    Return_if_fail(VC_close_checkout_stream(cofile, version, Rep_name_of_version_file()));

    project->repository_entry(this);

    return project;
}

PrProjectDescriptorPtrError RepEntry::checkout_create_prj_file (char const *filename,
								char const *fullname,
								char const *version)
{
    ProjectDescriptor *project;

    Return_if_fail(project << checkout_prj_file(fullname, version));

    Return_if_fail(project->write_project_file(filename));

    return project;
}

static void print_user_access(int mode0)
{
    int mode = mode0;

    if(mode == 0) {
	prcsoutput << "no permission" << prcsendl;
	return;
    }

    while(mode) {
	mode0 = mode;
	mode >>= 3;
    }

    Dstring perm;

    if(mode0 == 7) {
	/* write perm implies RWX */
	perm.append("read/write");
    } else if(mode0 & 1) {
	/* execute is all that's needed to read */
	perm.append("read only");
    } else {
	perm.append("neither read nor write -- fix me");
    }

    prcsoutput << perm << prcsendl;
}

static void print_access(struct stat *buf, struct passwd *pwd, struct group *grp)
{
    prcsoutput << "Repository entry owner:  " << pwd->pw_name << prcsendl;
    prcsoutput << "Repository entry group:  " << grp->gr_name << prcsendl;
    prcsoutput << "Repository entry access: " << prcsendl;
    prcsoutput << "Owner:             ";
    print_user_access(buf->st_mode & S_IRWXU);
    prcsoutput << "Group:             ";
    print_user_access(buf->st_mode & S_IRWXG);
    prcsoutput << "Other:             ";
    print_user_access(buf->st_mode & S_IRWXO);
}

PrVoidError RepEntry::Rep_print_access()
{
    struct passwd *pwd;
    struct group *grp;

    pwd = getpwuid(rep_stat_buf.st_uid);
    grp = getgrgid(rep_stat_buf.st_gid);

    if(!pwd)
	throw prcserror << "Can't lookup repository owner's name" << perror;

    if(!grp)
	throw prcserror << "Can't lookup repository group owner's name" << perror;

    owner_name = intern(pwd->pw_name);
    group_name = intern(grp->gr_name);

    print_access(&rep_stat_buf, pwd, grp);

    return NoError;
}

static PrVoidError can_set_group(const char* group_name, const char *current_group_name)
{
    struct group *grp;

    if(strcmp(group_name, current_group_name) != 0) {

	if((grp = getgrnam(group_name)) == NULL)
	    throw prcserror << "Can't lookup group " << squote(group_name) << dotendl;

	bool found = false;

	char** grp_members = grp->gr_mem;

	while(*grp_members) {
	    if(strcmp(get_login(), *grp_members++) == 0) {
		found = true;
		break;
	    }
	}

	if(!found)
	    throw prcserror << "You are not a member of that group" << dotendl;
    }

    return NoError;
}

static PrIntError query_mask()
{
    int mask = 0;

    mask |= 0077; /* clear group/other read/write */

    prcsquery << "Group members' access, "
	      << force("read only")
	      << report("read only")
	      << defopt('r', "Allow group read access")
	      << option('w', "Allow group write access")
	      << option('n', "Disallow group access")
	      << query("select");

    char c;

    Return_if_fail(c << prcsquery.result());

    if(c == 'r')
	mask &= 0727;
    else if (c == 'w')
	mask &= 0707;

    prcsquery << "Other users' access, "
	      << force("read only")
	      << report("read only")
	      << defopt('r', "Allow other users read access")
	      << option('w', "Allow other users write access")
	      << option('n', "Disallow other users access")
	      << query("select");

    Return_if_fail(c << prcsquery.result());

    if(c == 'r')
	mask &= 0772;
    else if(c == 'w')
	mask &= 0770;

    return mask;
}

const char* RepEntry::Rep_entry_path()
{
    return Rep_name_in_repository(entry_name);
}


/*
 * Compression is implemented by lazily uncompressing repository files
 * when they are needed an compressing all uncompressed files upon
 * exiting.  Therefore, the compress_command and uncompress_command
 * need only touch and remove the file COMPRESSION_TAG in the repository
 * to do this.  The code that compresses the directory is called from
 * clean_up() in prcs.cc.
 */
PrPrcsExitStatusError admin_compress_command()
{
    RepEntry* rep_entry;

    Return_if_fail(rep_entry << Rep_init_repository_entry(cmd_root_project_name, true, false, true));

    if(!rep_entry->Rep_needs_compress()) {

	if(creat(rep_entry->Rep_name_in_entry(prcs_compression_tag), 0444) < 0)
	    throw prcserror << "Failed writing repository tagfile" << perror;

	rep_entry->Rep_set_compression(true);
    }

    return ExitSuccess;
}

PrPrcsExitStatusError admin_uncompress_command()
{
    RepEntry* rep_entry;

    Return_if_fail(rep_entry << Rep_init_repository_entry(cmd_root_project_name, true, false, true));

    if(!rep_entry->Rep_needs_compress()) {
	if(option_immediate_uncompression) {
	    Return_if_fail(rep_entry->Rep_uncompress_all_files());
	} else {
	    prcsinfo << "Repository entry " << cmd_root_project_name
		     << "is already uncompressed.  Use -i to immediately decompress "
		"the entire repository" << dotendl;
	}

	return ExitSuccess;
    }

    If_fail(Err_unlink(rep_entry->Rep_name_in_entry(prcs_compression_tag))) {
	throw prcserror << "Failed unlinking repository tag" << perror;
    }

    if(option_immediate_uncompression) {
	Return_if_fail(rep_entry->Rep_uncompress_all_files());
    } else {
	prcsinfo << "Repository lazily uncompressed.  Actual decompression takes "
	    "place the next time you access each file of the project" << dotendl;
    }

    rep_entry->Rep_set_compression(false);

    return ExitSuccess;
}

PrPrcsExitStatusError admin_access_command()
{
    int mask = 0;
    const char* group_name;

    if(cmd_root_project_name) {
	RepEntry* rep_entry;

	prcsinfo << "Setting access permissions for project " << squote(cmd_root_project_name)
		 << dotendl;

	Return_if_fail(rep_entry <<
		       Rep_init_repository_entry(cmd_root_project_name, true, false, true));

	Return_if_fail(rep_entry->Rep_print_access());

	if(!fs_same_file_owner(rep_entry->Rep_entry_path()) && getuid() != 0) {
	    throw prcserror <<"Only the owner of this repository entry may modify its "
		"access permissions" << dotendl;
	}

	prcsquery << "Enter a group"
		  << definput(rep_entry->Rep_group_name())
		  << string_query("");

	Return_if_fail(group_name << prcsquery.string_result());

	Return_if_fail(can_set_group(group_name, rep_entry->Rep_group_name()));

	Return_if_fail(mask << query_mask());

	Return_if_fail(rep_entry->Rep_chgrp(group_name));

	Return_if_fail(rep_entry->Rep_chmod(0777 ^ mask));

    } else {
	/* Set permissions for repository */
	Return_if_fail(Rep_init_repository());

	const char* rep_name = Rep_name_in_repository("");

	prcsinfo << "Setting access permissions for repository "
		 << rep_name << dotendl;

	struct passwd *pwd;
	struct group *grp;
	const char* group_name, *owner_name, *new_group_name;
	struct stat stat_buf;

	If_fail(Err_stat(Rep_name_in_repository(""), &stat_buf))
	    throw prcserror << "Stat failed on " << squote(rep_name) << perror;

	pwd = getpwuid(stat_buf.st_uid);
	grp = getgrgid(stat_buf.st_gid);

	if(!pwd)
	    throw prcserror << "Can't lookup repository owner's name" << perror;

	if(!grp)
	    throw prcserror << "Can't lookup repository group owner's name" << perror;

	owner_name = intern(pwd->pw_name);
	group_name = intern(grp->gr_name);

	print_access(&stat_buf, pwd, grp);

	if(!fs_same_file_owner(rep_name) && getuid() != 0) {
	    throw prcserror <<"Only the owner of this repository may modify its "
		"access permissions" << dotendl;
	}

	prcsquery << "Enter a group"
		  << definput(group_name)
		  << string_query("");

	Return_if_fail(new_group_name << prcsquery.string_result());

	Return_if_fail(can_set_group(new_group_name, group_name));

	Return_if_fail(mask << query_mask());

	if(option_report_actions)
	    return ExitSuccess;

	If_fail(Err_chmod(rep_name, (0777 ^ mask) | S_ISGID | 01000))
	    throw prcserror << "Chmod failed on " << squote(rep_name) << perror;

	const char* rep_lock_name = Rep_name_in_repository(prcs_lock_dir);

	If_fail(Err_chmod(rep_lock_name, 0777 ^ mask))
	    throw prcserror << "Chmod failed on " << squote(rep_lock_name) << perror;

    }

    return ExitSuccess;
}

PrPrcsExitStatusError admin_init_command()
{
    RepEntry* rep_entry;

    Return_if_fail(rep_entry <<
		   Rep_init_repository_entry(cmd_root_project_name, true, true, true));

    rep_entry->Rep_print_access();

    return ExitSuccess;
}

static PrVoidError repository_cleanup_handler(void* data, bool signaled)
{
    RepEntry* rep_entry;

    rep_entry = (RepEntry*) data;

    if(signaled) {
	if(rep_entry->Rep_needs_compress()) {
	    prcswarning << "Repository entry " << rep_entry->Rep_entry_path()
			<< " not recompressed" << dotendl;
	}
    } else {
	rep_entry->Rep_compress_all_files();
    }

    rep_entry->Rep_unlock_repository();

    delete rep_entry;

    return NoError;
}
