/* -*-Mode: C++;-*-
 * $Id: setkeys.cc 1.4.1.5.1.8.1.2 Mon, 02 Sep 1996 16:59:57 -0700 jmacd $
 * setkeys.c:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Authors: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 *          Paul N. Hilfinger <Hilfinger@CS.Berkeley.EDU>
 */

extern "C" {
#include <unistd.h>
}

#include "prcs.h"
#include "setkeys.h"
#include "memseg.h"
#include "system.h"

static ostream* output_file; /* non-null if outputting to a file. */
static MemorySegment * output_segment; /* non-null if outputting to memory. */

static int deletekeys; /* true iff unsetting keys. */
static int keyword_number; /* number of keys. */
static const char* const* keyword_pair_list; /* key name-value pairs. */

static const char* unchanged_segment;
static int unchanged_len;
static bool output_changed;



static int find_keyword(const char* s, int N, const char* const* aList)
     /* If there exists i with 0 <= i < N such that S begins with	*/
     /* ALIST[2i] followed by a delimiter, return 2i.  Otherwise	*/
     /* return -1.  Delimiters are non-alphanumeric characters.		*/
{
    int i;

    for (i = 0; i < N; i += 1) {
	int len = strlen(aList[i+i]);
	if (strncmp(s, aList[i+i], len) == 0 && ! isalnum(s[len]))
	    return 2*i;
    }

    return -1;
}

static PrVoidError checked_write(char const* s, int len)
{
    if(output_file) {
	output_file->write(s, len);

	if (output_file->bad())
	    throw prcserror << "Write failed while setting keys" << perror;
    }

    if (output_segment)
	output_segment->append_segment(s, len);

    return NoError;
}

static PrVoidError output_text_segment(char const* s, long len)
{
    if (s == NULL) {
	/* nothing */
    } else if (len == 0) {
	return NoError;
    } else if(unchanged_segment+unchanged_len == s ||
	      memcmp(unchanged_segment+unchanged_len, s, len) == 0) {
	unchanged_len += len;
	return NoError;
    } else {
	output_changed = true;
    }

    if (unchanged_len > 0) {
	Return_if_fail(checked_write(unchanged_segment, unchanged_len));
	unchanged_segment += unchanged_len;
	unchanged_len = 0;
    }

    if (s != NULL)
	Return_if_fail(checked_write(s, len));

    return NoError;
}


/* Assuming that inp is positioned just after a '$' that might begin */
/* a keyword phrase other than Format, convert the phrase and send to */
/* output_text_segment, returning a pointer to just after the closing */
/* '$' in the input.  If this is not the start of a keyword phrase, */
/* return NULL.   If SIMPLE, then just output the keyword's value, */
/* without any surrounding keyword:...$ text. */
static PrConstCharPtrError process_normal_key(const char* inp, long len, int simple)
{
    const char* end_phrase;
    const char* after_key;
    int new_len;
    int n;

    end_phrase = (const char*)memchr(inp, '$', len);

    if (end_phrase == NULL || memchr(inp, '\n', end_phrase - inp) != NULL)
	return (const char*)0;

    end_phrase += 1;

    n = find_keyword(inp, keyword_number, keyword_pair_list);

    if (n >= 0) {
	int key_len = strlen(keyword_pair_list[n]);

	after_key = inp + key_len;
	new_len = strlen(keyword_pair_list[n+1]);

	if ( deletekeys ) {
	    Return_if_fail(output_text_segment(keyword_pair_list[n], key_len));

	    Return_if_fail(output_text_segment("$", 1));

	    return end_phrase;
	} else {
	    if (! simple) {
		Return_if_fail(output_text_segment(keyword_pair_list[n], key_len));

		Return_if_fail(output_text_segment(": ", 2));
	    }

	    Return_if_fail(output_text_segment(keyword_pair_list[n+1], new_len));

	    if (! simple)
		Return_if_fail(output_text_segment(" $", 2));

	}

	return end_phrase;
    } else {
	return (const char*)0;
    }
}

/* Assuming that inp is positioned just after a '$' before the string */
/* "Format:", convert the phrase (with the subsequent line) and send to */
/* output_text_segment, returning a pointer to just after the replaced */
/* line in the input.  If this is not the start of a proper Format phrase, */
/* return NULL. */
static PrConstCharPtrError process_format(const char* inp0, long len0)
{
    const char* inp;
    const char* format;
    const char* end_format_line;
    int len, format_len;

    inp = inp0; len = len0;

    inp += 7; len -= 7;

    while (isspace(*inp) && len > 0)
	inp += 1, len -= 1;

    if (*inp != '\"')
	return (const char*)0;

    format = inp += 1;
    len -= 1;

    while(true) {
	if (len == 0 || *inp == '\n')
	    return (const char*)0;
	else if (*inp == '\"')
	    break;
	else if (*inp == '\\' && len > 1)
	    inp += 2, len -= 2;
	else
	    inp += 1, len -= 1;
    }

    format_len = inp - format;

    do {
	inp += 1, len -= 1;
    } while (len > 0 && *inp != '$' && *inp != '\n');

    if (len == 0 || *inp != '$')
	return (const char*)0;

    do {
	inp += 1; len -= 1;
    } while (len > 0 && *inp != '\n');

    if (len <= 0)
	return (const char*)0;

    end_format_line = inp;

    do {
	inp += 1; len -= 1;
    } while (len > 0 && *inp != '\n');

    Return_if_fail(output_text_segment(inp0, end_format_line - inp0 + 1));

    if ( deletekeys )
	return inp;

    while (format_len > 0) {
	switch (*format) {
	case '\\':
	    switch (format[1]) {
	    case 't':
		Return_if_fail(output_text_segment("\t", 1));
		break;
	    default:
		Return_if_fail(output_text_segment(format+1, 1));
		break;
	    }
	    format += 2; format_len -= 2;
	    break;
	case '$': {
	    const char* end_phrase;

	    Return_if_fail(end_phrase << process_normal_key(format+1, format_len-1, 1));

	    if (end_phrase == NULL) {
		Return_if_fail(output_text_segment(format, 1));
		format += 1; format_len -= 1;
	    } else {
		format_len -= end_phrase - format;
		format = end_phrase;
	    }
	    break;
	}
	default:
	    Return_if_fail(output_text_segment(format, 1));
	    format += 1; format_len -= 1;
	    break;
	}
    }

    return inp;
}

/* Convert input[0 .. inputLen-1] according to keyword_number and		*/
/* keyword_pair_list as indicated in the comments at the beginning of this	*/
/* file, sending the results to output_text_segment.  */
static PrVoidError convert(const char* input_buffer, int input_buffer_len)
{
    output_changed = false;
    unchanged_len = 0;
    unchanged_segment = input_buffer;

    while(true) {
	const char* end_phrase;
	const char* next_possible_key;

	if (output_changed && !output_file && !output_segment)
	    return NoError;

	next_possible_key = (const char*)memchr(input_buffer, '$', input_buffer_len);

	if (next_possible_key == NULL) {
	    Return_if_fail(output_text_segment(input_buffer, input_buffer_len));
	    break;
	}

	next_possible_key += 1;

	Return_if_fail(output_text_segment(input_buffer, next_possible_key - input_buffer));

	input_buffer_len -= next_possible_key - input_buffer;
	input_buffer = next_possible_key;

	if (strncmp(input_buffer, "Format:", 7) == 0)
	    Return_if_fail(end_phrase << process_format(input_buffer, input_buffer_len));
	else
	    Return_if_fail(end_phrase << process_normal_key(input_buffer, input_buffer_len, 0));

	if (end_phrase == NULL)
	    continue;

	input_buffer_len -= end_phrase - input_buffer;
	input_buffer = end_phrase;
    }

    return output_text_segment(NULL, 0);
}

/*****/

static MemorySegment buffer;

PrBoolError setkeys_internal(const char* input_buffer0,
			     int input_buffer_len0,
			     ostream* os,
			     MemorySegment* seg,
			     const SetkeysData data)
{
    keyword_number = data.numkeys;
    keyword_pair_list = data.keypairs;
    deletekeys = !data.setkeys;

    output_segment = seg;
    output_file = os;

    if (seg) seg->clear_segment();

    Return_if_fail(convert(input_buffer0, input_buffer_len0));

    return output_changed;
}

PrBoolError setkeys_infile(int fd, const char* file1, const SetkeysData data)
{
    If_fail(buffer.map_file(fd))
	throw prcserror << "Read error while setting keywords" << perror;

    ofstream os(file1);

    Return_if_fail(setkeys_inputbuf(buffer.segment(), buffer.length(), &os, data));

    os.close();

    if (os.bad())
	throw prcserror << "Write failed on file " << squote(file1) << perror;

    return output_changed;
}

PrBoolError setkeys(const char* file0, const char* file1, const SetkeysData data)
{
    int fd;

    If_fail(fd << Err_open(file0, O_RDONLY))
	throw prcserror << "Error opening file " << squote(file0) << perror;

    PrBoolError ret = setkeys_infile(fd, file1, data);

    close(fd);

    return ret;
}

PrBoolError setkeys_inputbuf(const char* input_buffer0,
			     int input_buffer_len0,
			     ostream* os,
			     const SetkeysData data)
{
    return setkeys_internal(input_buffer0, input_buffer_len0, os, NULL, data);
}

PrBoolError setkeys_inoutbuf(const char* inputbuf,
			     int inputbuflen,
			     MemorySegment* seg,
			     const SetkeysData data)
{
    return setkeys_internal(inputbuf, inputbuflen, NULL, seg, data);
}

PrBoolError setkeys_outbuf(const char* file,
			   MemorySegment* seg,
			   const SetkeysData data)
{
    int fd;

    If_fail(fd << Err_open(file, O_RDONLY))
	throw prcserror << "Error opening file " << squote(file) << perror;

    If_fail(buffer.map_file(fd))
	throw prcserror << "Read error while setting keywords" << perror;

    PrBoolError ret = setkeys_internal(buffer.segment(), buffer.length(), NULL, seg, data);

    close(fd);

    return ret;
}
