/* -*-Mode: C++;-*-
 * $Id: syscmd.cc 1.5.1.19.1.4 Fri, 16 Aug 1996 01:24:25 -0700 jmacd $
 * syscmd.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */

extern "C" {
#include <sys/wait.h>
}

#include "prcs.h"
#include "dstring.h"
#include "syscmd.h"
#include "misc.h"
#include "system.h"

#ifndef SYS_RCS_COMMAND_PATH
#error Please define SYS_RCS_COMMAND_PATH
#endif

SystemCommand rcs_command(SYS_RCS_COMMAND_PATH);
SystemCommand ci_command(SYS_CI_COMMAND_PATH);
SystemCommand co_command(SYS_CO_COMMAND_PATH);
SystemCommand rlog_command(SYS_RLOG_COMMAND_PATH);
SystemCommand tar_command(SYS_TAR_COMMAND_PATH);
SystemCommand gdiff3_command(SYS_GDIFF3_COMMAND_PATH);
SystemCommand gdiff_command(SYS_GDIFF_COMMAND_PATH);
SystemCommand ls_command(SYS_LS_COMMAND_PATH);
SystemCommand gzip_command(SYS_GZIP_COMMAND_PATH);

char *const null_environment[] = { NULL };

SystemCommand::SystemCommand(const char* name0)
    :name(name0) {}

PrVoidError SystemCommand::init()
{
    static char const path_env[] = "RCS_PATH";

    if (fp.length() == 0) {

	char* user_path0 = getenv(path_env);

	if( user_path0 ) {

	    char* user_path = user_path0;

	    while(user_path && user_path[0]) {
		char* first_colon = strchr(user_path, ':');
		char* this_path;
		int this_path_len;

		if(first_colon) {
		    this_path = user_path;
		    this_path_len = first_colon - this_path;
		    user_path = first_colon + 1;
		} else {
		    this_path = user_path;
		    this_path_len = strlen(user_path);
		    user_path = NULL;
		}

		fp.append(this_path, this_path_len);

		if(this_path[this_path_len - 1] != '/')
		    fp.append('/');

		fp.append(strip_leading_path(name));

		if(access(fp, F_OK | X_OK) >= 0)
		    return NoError;

		if(errno != ENOENT)
		    prcswarning << "Can't execute " << squote(fp) << perror;

		fp.truncate(0);
	    }
	}

	fp.assign(name);

	if(access(name, F_OK | X_OK) >= 0)
	    return NoError;

	if(errno != ENOENT)
	    throw prcserror << "Can't execute system command " << squote(name) << perror;
	else if(user_path0)
	    throw prcserror << "System command "
			    << squote(strip_leading_path(name)) << " not in $"
			    << path_env << " nor the default location " << squote(name) << dotendl;
	else
	    throw prcserror << "System command " << squote(name) << " does not exist" << dotendl;

    }

    return NoError;
}

PrVoidError SystemCommand::open(bool out, bool err)
{
    Return_if_fail(make_pipe_out(argl, &pr, out, err));

    return NoError;
}

PrArgListPtrError SystemCommand::new_arg_list()
{
    Return_if_fail(init());

    close();

    argl.append(fp);

    return &argl;
}

FILE* SystemCommand::standard_out() const { return pr.standard_out; }
FILE* SystemCommand::standard_err() const { return pr.standard_err; }

PrExitStatusError SystemCommand::open_stdout()
{
    return make_pipe_stdout(argl);
}

PrExitStatusError SystemCommand::open_filein(int fd, bool cl)
{
    return make_pipe_file_in(fd, argl, cl);
}

const char* SystemCommand::path() const { return fp; }

/* returns PIPE_CLOSE_FAILURE if process did not exit properly due to
 * a call to exit(). otherwise it returns the exit value of the
 * process that was running. */
PrExitStatusError SystemCommand::close()
{
    int ret;

    if ( pr.standard_err != NULL || pr.standard_out != NULL )
	If_fail( ret << close_pipe(&pr) )
	    throw prcserror << "Child process " << squote(name)
		             << " terminated abnormally" << dotendl;

    pr.standard_err = NULL;
    pr.standard_out = NULL;
    argl.truncate(0);

    return ret;
}

static pid_t *pids = NULL;
static int stdinpipe[2];
static int maxfd;

void abort_child(const char* argv0)
{
    FILE* user_tty = fopen(ctermid(NULL), "w");
    if(user_tty)
	fprintf(user_tty, "prcs: Exec failed: %s: %s\n", argv0, strerror(errno));

    _exit(127);
}

PrPipeRecPtrError make_pipe_out(const char* const* argv, PipeRec* pr,
				bool pipeout, bool pipeerr)
{
    int pid, stdpipe[2], errpipe[2];

    if(pids == NULL) {
	maxfd = sysconf(_SC_OPEN_MAX);

	if (maxfd < 0)
	    throw prcserror << "sysconf(_SC_OPEN_MAX) failed--can't determine file "
		"descriptor table size" << perror;

	pids = NEWVEC0(pid_t, maxfd);

	if(pipe(stdinpipe) < 0)
	    throw prcserror << "pipe failed" << perror;

	close(stdinpipe[PIPE_WRITE_FD]);
    }

    if(pipeout && pipe(stdpipe) < 0)
	throw prcserror << "pipe failed" << perror;

    if(pipeerr && pipe(errpipe) < 0)
	throw prcserror << "pipe failed" << perror;

    pid = fork();
    if(pid < 0)
	throw prcserror << "fork failed" << perror;

    if (pid == 0) {
	close(STDIN_FILENO);
	// This had to be added because of a bug in FreeBSD/NetBSD's
	// freopen treatment of ferror..., or a GNU diff3 bug where by
	// it freopens stdin without noticing its stdin is closed.
	dup2(stdinpipe[PIPE_READ_FD], STDIN_FILENO);

	if(pipeout) {
	    dup2(stdpipe[PIPE_WRITE_FD], STDOUT_FILENO);
	    close(stdpipe[PIPE_WRITE_FD]);
	    close(stdpipe[PIPE_READ_FD]);
	}

	if(pipeerr) {
	    dup2(errpipe[PIPE_WRITE_FD], STDERR_FILENO);
	    close(errpipe[PIPE_WRITE_FD]);
	    close(errpipe[PIPE_READ_FD]);
	}

	execve(argv[0], (char* const*)argv, null_environment);
	abort_child(argv[0]);
    }

    if(pipeout) {
	close(stdpipe[PIPE_WRITE_FD]);
	pr->standard_out = fdopen(stdpipe[PIPE_READ_FD], "r");
	pids[stdpipe[PIPE_READ_FD]] = pid;
    }

    if(pipeerr) {
	close(errpipe[PIPE_WRITE_FD]);
	pr->standard_err = fdopen(errpipe[PIPE_READ_FD], "r");
	pids[errpipe[PIPE_READ_FD]] = pid;
    }

    return pr;
}

NprExitStatusError close_pipe(PipeRec *pr)
{
    int fd=0, status;

    ASSERT(pr, "nothing to close");

    if(pr->standard_out != NULL) {
	fd = fileno(pr->standard_out);
	empty_file(pr->standard_out);
	fclose( pr->standard_out );
    }

    if(pr->standard_err != NULL) {
	fd = fileno(pr->standard_err);
	empty_file(pr->standard_err);
	fclose( pr->standard_err );
    }

    If_fail(status << Err_waitpid(pids[0]))
	return FatalError;

    if(!WIFEXITED(status))
	return NonFatalError;
    else
	return WEXITSTATUS(status);
}

PrExitStatusError make_pipe_file_in(int fd, const char* const* argv, bool close_stdout)
{
    int status, pid;

    if((pid = fork()) < 0)
	throw prcserror << "fork failed" << perror;

    if (pid == 0) {

	if(close_stdout) {
	    close(STDOUT_FILENO);
	    close(STDERR_FILENO);
	}

	dup2(fd, STDIN_FILENO);

	execve(argv[0], (char * const*)argv, null_environment);
	abort_child(argv[0]);
    }

    If_fail(status << Err_waitpid(pid))
	throw prcserror << "Waitpid failed on pid " << pid << perror;

    close(fd);

    if(!WIFEXITED(status))
	if (WIFSIGNALED(status)) {
	    throw prcserror << "Process " << squote(argv[0]) << " terminated with signal "
			    << WTERMSIG(status) << dotendl;
	} else {
	    throw prcserror << "Process " << squote(argv[0]) << " was stopped with signal "
			    << WSTOPSIG(status) << dotendl;
	}
    else
	return WEXITSTATUS(status);
}

PrExitStatusError make_pipe_stdout(const char* const* argv)
{
    int pid, status;

    if((pid = fork()) < 0) {
	throw prcserror << "fork failed" << perror;
    } else if (pid == 0) {
	close(STDIN_FILENO);

	execve(argv[0], (char * const*)argv, null_environment);
	abort_child(argv[0]);
    }

    If_fail(status << Err_waitpid(pid))
	throw prcserror << "Waitpid failed on pid " << pid << perror;

    if(!WIFEXITED(status))
	if (WIFSIGNALED(status)) {
	    throw prcserror << "Process " << squote(argv[0]) << " terminated with signal "
			    << WTERMSIG(status) << dotendl;
	} else {
	    throw prcserror << "Process " << squote(argv[0]) << " was stopped with signal "
			    << WSTOPSIG(status) << dotendl;
	}
    else
	return WEXITSTATUS(status);
}
