/* -*-Mode: C++;-*-
 * $Id: vc.cc 1.13.1.1.1.11.1.12.1.4 Tue, 03 Sep 1996 16:48:15 -0700 jmacd $
 * vc.cc:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */


#include "prcs.h"

#include "fileent.h"
#include "hash.h"
#include "dstring.h"
#include "vc.h"
#include "syscmd.h"
#include "system.h"
#include "misc.h"

EXTERN const char* rcs_text;       /* declared in vc.l */
EXTERN int VC_get_token(FILE* is); /* declared in vc.l */

static bool VC_expect_token(FILE* stream, int type, Dstring* result);

/*
 * VC_checkin --
 *
 *     Buffers checkins of working files into RCS versionfiles.  It
 *     first calls upon RCS to lock the parent version.  Currently,
 *     one fork to RCS is required for each file.  This could possibly
 *     be replaced in the future by either some analysis of the time
 *     stamp on the file (and assume nothing has been checked in since
 *     the last was checked in) or by parsing the RCS file according
 *     to rcsfile(5).  This is the simplest and most transparent way
 *     of doing this, however, it only relies on RCS's external
 *     interface.
 *
 *     when VC_checkin is called with all arguments except the log
 *     NULL, the actual checkin takes place.  Also, if the argument
 *     size approaches sysconf(_SC_ARG_MAX), the buffer will be
 *     flushed.  The checkin unlocks the RCS file.
 * Results:
 *     true if everything is going okay.  false if an error has
 *     occured.  the new RCS version assigned to newfile is returned
 *     in the Dstring *parentversion.  If parent version was empty,
 *     then checkin will fill RCS version 1.1.
 * Side Effects:
 *     if sucessful, the RCS files will be modified.
 * Parameters:
 *     newfile -- path of the working file.
 *     parentversion -- return buffer for new version number
 *     versionfile -- path of the RCS version file.
 *     log -- a log message
 */
PrVoidError VC_checkin(const char* newfile,
		       Dstring* parentversion,
		       const char* versionfile,
		       const char* log)
{
    static List<Dstring*> *accumulate(NULL);
    static ArgList* args;
    static int arg_list_size;
    static int max_arg_size = sysconf(_SC_ARG_MAX) - (4 * MAXPATHLEN);
    static Dstring date;

    if(accumulate == NULL && newfile == NULL)
        return NoError;

    if(date.length() == 0) {
        Return_if_fail(args << ci_command.new_arg_list());

        arg_list_size = 0;
        date.append("-d");
        date.append(get_utc_time());
        args->append((char*)0); /*  args[1] == log message */
        args->append(date);
        args->append("-u");
        args->append("-T");
	args->append("-f");
    }

    /*
     * if the file being checked in has a parent version, then
     * instruct RCS to pick a good new version number by locking the
     * parent version
     */
    if( newfile != NULL &&
        parentversion != NULL &&
        parentversion->length() != 0 ) {
        Dstring ver;
        ArgList *argl;

        Return_if_fail(argl << rcs_command.new_arg_list());

        ver.append("-l");
        ver.append(*parentversion);

        argl->append("-q");
        argl->append("-T"); /* preserve time stamp */
	argl->append("-u");
	argl->append("-M");
        argl->append(ver.cast());
        argl->append(versionfile);

        Return_if_fail_if_ne(rcs_command.open_stdout(), 0 )
            throw prcserror << "rcs exited abnormally on file "
			    << squote(versionfile) << dotendl;
    } else if (newfile != NULL &&
               parentversion != NULL &&
               parentversion->length() == 0) {
        /*
         * if the version is empty, assume the file will be the *
         * initial version.  We still must try to lock the version
         * in the case the file is being checked in by a different
         * owner and no locks are set (eg. the project file, which
         * doesn't track RCS versions).   To prove this:
         * user1% rcs -i -U foo,v
         * # creates initial version without strict locking
         * user2% co foo
         * user2% chmod +w foo
         * # essentially what PRCS does, since it checks out on stdout
         * # modify foo
         * user2% ci foo
         * # no locks set by user2, fail
         * user1% ci foo
         * # since user1 owns it, it succeeds, I guess.
         *
         * The point is, you won't see why this is neccesary until
         * you try using two users.
         */
        ArgList *argl;

        Return_if_fail(argl << rcs_command.new_arg_list());

        argl->append("-q");
        argl->append("-T"); /* preserve time stamp */
	argl->append("-u");
	argl->append("-M");
        argl->append("-l");
        argl->append(versionfile);

        Return_if_fail_if_ne(rcs_command.open_stdout(), 0 )
            throw prcserror << "rcs exited abnormally on file "
			     << squote(versionfile) << dotendl;
    }

    if (newfile == NULL || arg_list_size > max_arg_size) {
	List<Dstring*> *l;
        Dstring logmes;
        int vclex;

        l = accumulate = accumulate->nreverse();

        logmes.append("-m");
        logmes.append(!log ? "*** no log ***" : log);
        args->index(1, logmes.cast());

        Return_if_fail(ci_command.open(true, true));

        vclex = VC_get_token(ci_command.standard_err());
        while (l != NULL) {

            if(vclex != NewRevision &&
               vclex != PrevRevision &&
               vclex != InitialRevision)
                throw prcserror << "Error scanning RCS ci output" << dotendl;

            if (l->head() != NULL)
                l->head()->assign(rcs_text);

            l = l->tail();
            vclex = VC_get_token(NULL);
        }

	Return_if_fail_if_ne(ci_command.close(), 0)
	    throw prcserror << "ci exited abnormally on batch" << dotendl;

	accumulate->deleteList();
        accumulate = NULL;
        arg_list_size = 0;

	Return_if_fail(args << ci_command.new_arg_list());

        args->append((char*)0); /*  args[1] == log message */
        args->append(date);
        args->append("-u");
        args->append("-T");

    }

    if(newfile != NULL) {
        accumulate = new List<Dstring*>(parentversion, accumulate);
        args->append(newfile);
        args->append(versionfile);
        arg_list_size += 2;
        arg_list_size += strlen(newfile);
        arg_list_size += strlen(versionfile);
    }

    return NoError;
}

/*
 * VC_register --
 *
 *     Used to create an initial RCS versionfile with no versions.  This
 *     avoids a condition in VC_checkin where RCS would be confused by
 *     a missing file.
 * Results:
 *     If sucessful, file has been registered.  Files are not actually
 *     created via a fork/exec to rcs until VC_register is called with
 *     NULL as its filename.  It will not register the same file twice
 *     even if the buffered files have not been created.  That is, two
 *     sucessive calls to VC_register with the same name will fail until
 *     VC_register(NULL) has been called.  After that, it will succeed
 *     again.  It is someone elses duty to check that the file doesn't
 *     already exist.
 *
 *     Should the buffer size approach sysconf(_SC_ARG_MAX), it will
 *     automatically flush the buffer.
 * Side Effects:
 *     when VC_register(NULL) gets called, files will be created.
 * Parameters:
 *     file names a file, or flushes the buffer if NULL.
 */
PrVoidError VC_register(const char* file)
{
    static DstringPtrArray *accumulate = NULL;
    static int arg_list_size;
    /*
     * max arg size is left at 2*MAXPATHLEN less than the real max for safety.
     */
    static int max_arg_size = sysconf(_SC_ARG_MAX) - (2 * MAXPATHLEN);
    ArgList *argl;

    if (accumulate == NULL && file == NULL)
        return NoError;

    if (accumulate == NULL) {
        arg_list_size = 0;
        accumulate = new DstringPtrArray;
    }

    if (file == NULL || arg_list_size > max_arg_size) {

        Return_if_fail(argl << rcs_command.new_arg_list());

        argl->append("-i");          /* RCS init */
        argl->append("-t/dev/null"); /* No initial description message. */
        argl->append("-U");          /* unset strict locking */
        argl->append("-q");          /* be quiet */
	for(int i = 0; i < accumulate->length(); i += 1)
	    argl->append(accumulate->index(i)->cast());

        Return_if_fail_if_ne(rcs_command.open_stdout(), 0)
            throw prcserror << "RCS exited abnormally on batch init" << dotendl;

	foreach (copy_ptr, accumulate, DstringPtrArray::ArrayIterator)
	    delete *copy_ptr;

        delete accumulate;

        arg_list_size = 0;

	if (file) {
	    accumulate = new DstringPtrArray;
	} else {
	    accumulate = NULL;
	}
    }

    if(file != NULL) {
	Dstring *copy = new Dstring(file);
	accumulate->append(copy);
	arg_list_size += copy->length() + 1;
    }

    return NoError;
}

/*
 * VC_expect_token --
 *
 *     Used when a certain token type is expected from the token stream.
 * Results:
 *     Returns true when the expected token type is found, false otherwise.
 *     the external variable rcs_text will also contain the token text.
 * Side Effects:
 *     One token is read from the stream via a call to VC_get_token.
 * Parameters:
 *     stream is the argument passed to VC_get_token, possibly NULL to
 *     continue reading the last stream passed. type is one of the
 *     #defined types in include/vc.h.  result gets the value of the
 *     token text if succesful, if non-null.
 */
static bool VC_expect_token(FILE* stream, int type, Dstring* result)
{
    if(type != VC_get_token(stream))
        return false;
    else if(result)
        result->assign(rcs_text);
    return true;
}

/*
 * VC_checkout_stream --
 *
 *     Retrieves version from revisionfile, into a stream.
 * Results:
 *     returns a FILE* opened for reading or NULL upon failure.
 * Side Effects:
 *     a file descriptor is open.  VC_close_checkout_stream should
 *     be called.
 * Parameters:
 *     versionfile names the pathname of an RCS file.  version is
 *     retrieved.
 */
PrCFilePtrError VC_checkout_stream(const char* version,
				   const char* revisionfile)
{
    ArgList *argl;
    Dstring ver("-p");

    Return_if_fail(argl << co_command.new_arg_list());

    ver.append(version);

    argl->append("-ko");
    argl->append(ver);
    argl->append(revisionfile);

    If_fail(co_command.open(true, true))
        throw prcserror << "Failed checking out version " << version
			 << " from file " << squote(revisionfile) << dotendl;

    return co_command.standard_out();
}

/*
 * VC_close_checkout_stream --
 *
 *     cleans up after a VC_checkout_stream command and checks that RCS
 *     did not report any errors.
 * Results:
 *     true if everything went okay.
 * Side effects:
 *     FILE *f will be closed.
 * Parameters:
 *     RCS versionfile file and version.
 */
#ifdef NDEBUG
#define UNUSED(x)
#else
#define UNUSED(x) x
#endif
PrVoidError VC_close_checkout_stream(FILE* UNUSED(f), const char* version, const char* file)
{
    /* the following shouldnt be neccesary ... */
    empty_file(co_command.standard_out());

    if(!VC_expect_token(co_command.standard_err(), RlogVersion, NULL) ||
       strcmp(rcs_text, version) != 0)
        throw prcserror << "Consistency check failed checking out version "
			 << version << " from file " << squote(file) << dotendl;

    ASSERT(f == co_command.standard_out(), "no way, man!");

    Return_if_fail_if_ne(co_command.close(), 0)
	throw prcserror << "co exited with non-zero status" << dotendl;

    return NoError;
}

/*
 * VC_checkout_file --
 *
 *     checks out a file and places it in newfile.
 * Results:
 *     true if everything went okay.
 * Side effects:
 *     None.
 * Parameters:
 *     RCS versionfile file and version.  File is placed in newfile.
 */
PrVoidError VC_checkout_file(const char* newfile,
			     const char* version,
			     const char* revisionfile)
{
    FILE *infile, *outfile;

    Return_if_fail(infile << VC_checkout_stream(version, revisionfile));

    If_fail(outfile << Err_fopen(newfile, "w"))
        throw prcserror << "Can't open file " << squote(newfile)
		         << " for writing" << perror;

    Return_if_fail(write_file(infile, outfile));
    Return_if_fail(VC_close_checkout_stream(infile, version, revisionfile));

    If_fail(Err_fclose(outfile))
	throw prcserror << "Write failed on file " << squote(newfile) << perror;

    return NoError;
}

/*
 * VC_get_project_version_data --
 *
 *     Retrieves all the log information for a PRCS project file. (which
 *     is used to map PRCS versions to RCS versions) from an RCS versionfile.
 * Results:
 *     If return is successful, return value is a heap allocated array of
 *     ProjectVersionData.
 * Side Effects:
 *     Memory is allocated for the return value.
 * Parameters:
 *     versionfile names the pathname of an RCS file.
 */
PrProjectVersionDataPtrArrayPtrError
VC_get_project_version_data(const char* versionfile)
{
    ArgList *argl;
    int rev_count, i;

    Return_if_fail(argl << rlog_command.new_arg_list());

    argl->append(versionfile);

    Return_if_fail(rlog_command.open(true, false));

    if(VC_get_token(rlog_command.standard_out()) == RlogTotalRevisions)
	rev_count = atoi(rcs_text);
    else
	throw prcserror << "Unrecognized output from RCS rlog command on RCS file "
			 << squote(versionfile) << dotendl;

    ProjectVersionDataPtrArray* pvda = new ProjectVersionDataPtrArray;
    ProjectVersionDataPtrArray* r_pvda = new ProjectVersionDataPtrArray;

    if(rev_count == 0)
	return pvda;

    if(VC_get_token(NULL) != RlogVersion)
	goto error;

    for(i = 0; i < rev_count; i += 1) {
	RcsToken vclex;
	ProjectVersionData* pvd = new ProjectVersionData;

	pvd->rcs_version(intern(rcs_text));

	pvda->append(pvd);

	while(true) {
	    vclex = (RcsToken)VC_get_token(NULL);
	    if(vclex == RlogVersion || vclex == NoTokenMatch)
		break;
	    switch(vclex) {
	    case RlogDate:
		pvd->date(timestr_to_time_t(rcs_text));
		break;
	    case RlogAuthor:
		pvd->author(intern(rcs_text));
		break;
	    case PrcsMajorVersion:
		pvd->prcs_major(intern(rcs_text));
		break;
	    case PrcsMinorVersion:
		pvd->prcs_minor(intern(rcs_text));
		break;
	    case PrcsParentMajorVersion:
		pvd->parent_major(intern(rcs_text));
		break;
	    case PrcsParentMinorVersion:
		pvd->parent_minor(intern(rcs_text));
		break;
	    case PrcsVersionDeleted:
		pvd->deleted(true);
		break;
	    case PrcsMergedParentMajor:
	    case PrcsMergedParentMinor:
	    case PrcsDescendsFromMajor:
	    case PrcsDescendsFromMinor:
	    case NoTokenMatch:
		break;
	    case RlogTotalRevisions:
	    case RlogVersion:
	    case PrevRevision:
	    case NewRevision:
	    case InitialRevision:
	    case RcsAbort:
		goto error;
	    }
	}

	if(!pvd->OK())
	    throw prcserror << "Incomplete version log information in RCS file "
			     << squote(versionfile) << dotendl;
    }

    Return_if_fail_if_ne(rlog_command.close(), 0)
	throw prcserror << "RCS rlog command returned non-zero status on RCS file "
			 << squote(versionfile) << dotendl;

    /* Reverse the list */

    for(i = pvda->length() - 1; i >= 0; i -= 1)
	r_pvda->append(pvda->index(i));

    delete pvda;

    return r_pvda;

error:
    throw prcserror << "Error scanning RCS rlog output on RCS file "
		     << squote(versionfile) << dotendl;

}

/*
 * VC_get_version_data --
 *
 *     Retrieves all the log information from an RCS file.
 * Results:
 *     If return is successful, return value is a heap allocated array of
 *     RcsVersionData.
 * Side Effects:
 *     Memory is allocated for the return value.
 * Parameters:
 *     versionfile names the pathname of an RCS file.
 */
PrRcsDeltaPtrError
VC_get_version_data(const char* versionfile)
{
    ArgList *argl;
    int rev_count, i;

    Return_if_fail(argl << rlog_command.new_arg_list());

    argl->append(versionfile);

    Return_if_fail(rlog_command.open(true, false));

    if(VC_get_token(rlog_command.standard_out()) == RlogTotalRevisions)
	rev_count = atoi(rcs_text);
    else
	throw prcserror << "Unrecognized output from RCS rlog command on RCS file "
			 << squote(versionfile) << dotendl;

    RcsDelta* pvda = new RcsDelta(NULL);

    if(rev_count == 0)
	return pvda;

    if(VC_get_token(NULL) != RlogVersion)
	goto error;

    for(i = 0; i < rev_count; i += 1) {
	RcsToken vclex;
	RcsVersionData* pvd = new RcsVersionData;

	pvd->rcs_version(intern(rcs_text));

	pvda->insert(pvd->rcs_version(), pvd);

	while(true) {
	    vclex = (RcsToken)VC_get_token(NULL);
	    if(vclex == RlogVersion || vclex == NoTokenMatch)
		break;
	    switch(vclex) {
	    case RlogDate:
		pvd->date(timestr_to_time_t(rcs_text));
		break;
	    case RlogAuthor:
		pvd->author(intern(rcs_text));
		break;
	    case PrcsMajorVersion:
	    case PrcsMinorVersion:
	    case PrcsParentMajorVersion:
	    case PrcsParentMinorVersion:
	    case PrcsVersionDeleted:
	    case PrcsMergedParentMajor:
	    case PrcsMergedParentMinor:
	    case PrcsDescendsFromMajor:
	    case PrcsDescendsFromMinor:
	    case NoTokenMatch:
		break;
	    case RlogTotalRevisions:
	    case RlogVersion:
	    case PrevRevision:
	    case NewRevision:
	    case InitialRevision:
	    case RcsAbort:
		goto error;
	    }
	}

	if(!pvd->OK())
	    throw prcserror << "Incomplete version log information in RCS file "
			     << squote(versionfile) << dotendl;
    }

    Return_if_fail_if_ne(rlog_command.close(), 0)
	throw prcserror << "RCS rlog command returned non-zero status on RCS file "
			 << squote(versionfile) << dotendl;

    return pvda;

error:
    throw prcserror << "Error scanning RCS rlog output on RCS file "
		     << squote(versionfile) << dotendl;

}

PrBoolError VC_delete_version(const char* version,
			      const char* revisionfile)
{
    ArgList* args;
    Dstring outdate;

    Return_if_fail(args << rcs_command.new_arg_list());

    outdate.append("-o");
    outdate.append(version);

    args->append("-q");
    args->append(outdate);
    args->append(revisionfile);

    Return_if_fail(rcs_command.open(true, true));

    Return_if_fail_if_ne(rcs_command.close(), 0)
	return false;

    return true;
}

PrVoidError VC_set_log(const char* new_log_text,
		       const char* version,
		       const char* revisionfile)
{
    Dstring log;
    ArgList* args;

    log.append("-m");
    log.append(version);
    log.append(":");
    log.append(new_log_text);

    Return_if_fail(args << rcs_command.new_arg_list());

    args->append(log);
    args->append(revisionfile);

    Return_if_fail(rcs_command.open(true, true));

    Return_if_fail_if_ne(rcs_command.close(), 0) {
	throw prcserror << "RCS exited abnormally while updating log" << dotendl;
    }

    return NoError;
}

/* RcsDelta */

void RcsDelta::insert(const char* index, RcsVersionData* data)
{
    insertion_count += 1;

    if(index[0] == '\0') {
	rcs_data = data;
	return;
    }

    char* end_ptr;
    int delta_num = (int)strtol(index, &end_ptr, 10);

    if(!delta_array)
	delta_array = new RcsDeltaPtrArray;

    delta_array->expand(delta_num + 1, (RcsDelta*)0);

    if(delta_array->index(delta_num) == NULL)
	delta_array->index(delta_num, new RcsDelta(this));

    if(*end_ptr)
	delta_array->index(delta_num)->insert(end_ptr + 1, data);
    else
	delta_array->index(delta_num)->insert(end_ptr, data);
}

RcsVersionData* RcsDelta::lookup(const char* index)
{
    char* end_ptr;
    int delta_num = (int)strtol(index, &end_ptr, 10);

    if(!index[0])
	return rcs_data;
    else if(!delta_array || delta_array->length() <= delta_num)
       return NULL;
    else if(delta_array->index(delta_num)) {
	if(*end_ptr)
	    return delta_array->index(delta_num)->lookup(end_ptr + 1);
	else
	    return delta_array->index(delta_num)->lookup(end_ptr);

    } else
	return NULL;
}

void RcsDelta::remove(const char* index)
{
    insertion_count -= 1;

    char* end_ptr;
    int delta_num = (int)strtol(index, &end_ptr, 10);

    if(!index[0])
	rcs_data = NULL;
    else if(!delta_array || delta_array->length() <= delta_num)
       /* its not here? */
       return;
    else if(delta_array->index(delta_num)) {
	if(*end_ptr)
	    delta_array->index(delta_num)->remove(end_ptr + 1);
	else
	    delta_array->index(delta_num)->remove(end_ptr);

    }
    /* else its not here whats up with that */
}

/* This does an iterative table walk, you dont have to read it */
void RcsDelta::DeltaIterator::next()
{
    static int array_length; /* this doesnt need to eat stack space */

    _index += 1;

    if(!_current->delta_array) {
	array_length = 0;
    } else {
	array_length = _current->delta_array->length();
    }

    /* Its a leaf, 0 is the length of the non-existant array,
     * the data of a particular node is returned after the
     * children have been passed over. */
    if(_index == array_length) {
	_current_data = _current->rcs_data;
	/* _current_data may be NULL if this version was just deleted,
	 * if so, call next() again. */
	if(!_current_data)
	    next();
	return;
    } else if(_index < array_length) {
	if(_current->delta_array->index(_index)) {
	    _current->my_iter_index = _index;
	    _current = _current->delta_array->index(_index);
	    _index = -1;
	}

	next();
	return;
    } else {
	if(!_current->parent) {
	    /* this is the root, so we have covered all of them. */
	    _finished = true;
	    return;
	} else {
	    /* Restore the parents state */
	    _current = _current->parent;
	    _index = _current->my_iter_index;
	    next();
	    return;
	}
    }
}

RcsDelta::~RcsDelta()
{
    if(delta_array) {
	foreach(child, delta_array, RcsDeltaPtrArray::ArrayIterator) {
	    if(*child) {
		delete *child;
	    }
	}

	delete delta_array;
    }

    if(rcs_data)
	delete rcs_data;
}
