/* -*-Mode: C++;-*-
 * $Id: vclex.l 1.5.1.1.1.12 Fri, 09 Aug 1996 20:16:52 -0700 jmacd $
 * vc.l:
 *
 * Copyright (C) 1996, Regents of the University of California.
 * All Rights Reserved.
 *
 * This file is part of PRCS, the Project Revision Control System.
 *
 * Author: Josh MacDonald <jmacd@CS.Berkeley.EDU>
 */

%option nounput

%{
/* use this to allow yyout to be NULL and avoid echoing */
#define ECHO { if(yyout) (void)fwrite( yytext, yyleng, 1, yyout ); }
#include "utils.h"
#include "vc.h"
#include <ctype.h>
#include <string.h>
const char* rcs_text;
static char* killnl(char*);
static char* stop_at_ws(char*);
%}

dig     [0-9]
num     {dig}+
dotnum  "."{num}
rcnum   {num}{dotnum}+
date    {num}"/"{num}"/"{num}
time    {num}":"{num}":"{num}
tad	{date}" "{time}
login	[a-zA-Z][a-zA-Z0-9\-_]*
label	[A-Za-z0-9#^_+,][A-Za-z0-9#^\-_+=,.]*

%%

   /* The rlog information that I am interested in */
^"PRCS major version: "{label}\n { rcs_text = killnl(yytext + 20);
				   return PrcsMajorVersion; }
^"PRCS minor version: "{label}\n { rcs_text = killnl(yytext + 20);
				   return PrcsMinorVersion; }
^"PRCS parent major version: "{label}\n { rcs_text = killnl(yytext + 27);
					  return PrcsParentMajorVersion; }
^"PRCS parent minor version: "{label}\n { rcs_text = killnl(yytext + 27);
					  return PrcsParentMinorVersion; }
^"PRCS descends from major: "{label}\n { rcs_text = killnl(yytext + 26);
                                         return PrcsDescendsFromMajor; }
^"PRCS descends from minor: "{label}\n { rcs_text = killnl(yytext + 26);
                                         return PrcsDescendsFromMinor; }
^"PRCS version deleted"\n { return PrcsVersionDeleted; }

   /* this also serves as the co revision name */
^"revision "{rcnum}.*\n { rcs_text = stop_at_ws(yytext + 9);
		          return RlogVersion; }
^"date: "{tad} { rcs_text = yytext + 6;
		return RlogDate; }
";  author: "{login} { rcs_text = yytext + 11;
		       return RlogAuthor; }
";  state: ".*\n { }

   /* checkin stuff */
^"new revision: "{rcnum} { rcs_text = yytext + 14;
			  return NewRevision; }
^"initial revision: "{rcnum}\n { rcs_text = killnl(yytext + 18);
				 return InitialRevision; }
^"file is unchanged; reverting to previous revision "{rcnum}\n {
                               rcs_text = killnl(yytext + 50);
			       return PrevRevision; }
"; previous revision: "{rcnum}\n { }

^"total revisions: "{num}.*\n { rcs_text = yytext + 17;
                                return RlogTotalRevisions; }

^"done"\n { }
^.*",v  <--  ".*\n { }
^.*",v  -->  ".*\n { }

^"ci aborted"\n { return RcsAbort; }
^"rcs aborted"\n { return RcsAbort; }
^"rlog aborted"\n { return RcsAbort; }
^"co aborted"\n { return RcsAbort; }

   /* Stuff I'm not interested in. Mostly the rlog header. */
^-*\n { }
^=*\n { }
^"RCS file:".*\n { }
^"Working file:".*\n { }
^\t{login}": "{rcnum}\n { }
   /* \t"locked by: "{login}";"\n { } */
^"head:".*\n { }
^"branch".*\n { }
^"locks:".*\n { }
^"access list:".*\n { }
^"symbolic names:".*\n { }
^"keyword substitution:".*\n { }
^"description:".*\n { }
^"checked in by PRCS version".*\n { }
^"*** empty log message ***"\n { }

%%

/*
 * VC_get_token --
 *
 *     from the lex parser, a function which matches one of the
 *     tokens named in include/vc.h.  This is used to determine what
 *     type of output some rcs command has and the interesting information.
 *     if its argument is NULL, it continues matching from the end
 *     of the last match.  text of the token match is left in rcs_text.
 * Results:
 *     one of the token types.
 * Side effects:
 *     the stream is advanced.
 * Parameters:
 *     new stream or NULL.
 */
#if FLEX_STOPS_LEAKING
int VC_get_token(FILE* is)
{
    yyout = stderr;

    if(is) {
	yy_delete_buffer(YY_CURRENT_BUFFER);
	yyin = is;
	yy_switch_to_buffer(yy_create_buffer(yyin, YY_BUF_SIZE));
    }

    return yylex();
}
#else
/* This seems to help flex leak less memory, but it still leaks. */
int VC_get_token(FILE* is)
{
    static YY_BUFFER_STATE buffer = NULL;

    yyout = stderr;

    if(is) {
        if(buffer)
            yy_delete_buffer(buffer);
        yyin = is;
        buffer = yy_create_buffer(yyin, YY_BUF_SIZE);
        yy_switch_to_buffer(buffer);
    }

    return yylex();
}
#endif
/* check_token_match --
 *
 *     used in checking that a given string will match a type of token
 *     so that bad tokens are not inserted into the rcs log
 *     information.
 * Results:
 *     return value is 1 if it matches, 0 if it doesn't, -1 if
 *     type is invalid.
 * Side effects:
 *     None.
 * Parameters:
 *     TYPE must be one of "number", which checks that the string
 *     contains only digits, "label" which checks that the string *
 *     satisfies the definition of label in prcs.texi ([A-Za-z0- 9#^\-
 *     *_+=,.]+), and "login", which checks that the number satisfies
 *     ([a- * zA-Z][a-zA-Z0-9\-_]*).  This uses the exact same
 *     patterns * VC_get_token does, so if this fails it means using
 *     that token will * cause output to be misconstrued at some later
 *     date.
 */
int VC_check_token_match(const char* token, const char* type)
{
    int ret, len;
    yyout = NULL;
    if(strcmp(type, "login") == 0) {
	char* test = NEWVEC(char, 32 + strlen(token));
	strcpy(test, ";  author: ");
	strcat(test, token);
	len = strlen(test);
	test[len] = test[len + 1] = YY_END_OF_BUFFER_CHAR;
	yy_scan_string(test);
	ret = yylex();
	free(test);
	if (ret == RlogAuthor && yyleng == len) {
	    return 1;
	} else {
	    return 0;
	}
    } else if(strcmp(type, "label") == 0){
	char* test = NEWVEC(char, 32 + strlen(token));
	strcpy(test, "PRCS major version: ");
	strcat(test, token);
	strcat(test, "\n");
	len = strlen(test);
	test[len] = test[len + 1] = YY_END_OF_BUFFER_CHAR;
	yy_scan_string(test);
	ret = yylex();
	free(test);
	if (ret == PrcsMajorVersion && yyleng == len) {
	    return 1;
	} else {
	    return 0;
	}
    } else if(strcmp(type, "number") == 0) {
	const char *p = token;

	if(token[0] == '0' && token[1] != '\0') {
	    return 0;
	}

	while(*p != '\0') {
	    if(!isdigit(*p)) {
		return 0;
	    }
	    p += 1;
	}
	return 1;
    } else if(strcmp(type, "string") == 0) {
	if(token[0] == '\"' && token[strlen(token) - 1] == '\"')
	    return 1;
	else
	    return 0;
    } else if(strcmp(type, "any") == 0) {
	return 1;
    }

    return -1;
}

/*
 * killnl --
 *
 *     Kills a newlines at the end of patterns.
 * Results:
 *     Actually, it will kill the last character of any string.
 * Side Effects:
 *     Modifies s.
 * Parameters:
 *     Don't let s be "" or NULL.
 */
static char* killnl(char* s)
{
    s[strlen(s) - 1] = '\0';
    return s;
}

static char* stop_at_ws(char* s)
{
    char* s0 = s;
    while(!isspace(*s)) { s += 1; }
    *s = 0;
    return s0;
}

#if 0
int main()
{
    int it;
    while(1) {
	it = VC_get_token(stdin);
	fprintf(stdout, "got token %d\n", it);
    }
}
#endif
